#! /usr/bin/env python
#
# Copyright INRA-URGI 2009-2010
# 
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
# 
# As a counterpart to the access to the source code and rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty and the software's author, the holder of the
# economic rights, and the successive licensors have only limited
# liability.
# 
# In this respect, the user's attention is drawn to the risks associated
# with loading, using, modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean that it is complicated to manipulate, and that also
# therefore means that it is reserved for developers and experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and, more generally, to use and operate it in the
# same conditions as regards security.
# 
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
#
from commons.core.utils.RepetOptionParser import RepetOptionParser
"""
Cluster the data into regions (defined by size and overlap with next region) and keep only highest peaks.
"""

import os, os.path, re
from optparse import OptionParser
from SMART.Java.Python.structure.Transcript import Transcript
from SMART.Java.Python.misc.Progress import Progress
from SMART.Java.Python.misc.UnlimitedProgress import UnlimitedProgress
from commons.core.parsing.ParserChooser import ParserChooser
from commons.core.writer.WriterChooser import WriterChooser
from commons.core.writer.Gff3Writer import Gff3Writer

class ClusterizeBySlidingWindows(object):

	def __init__(self, strands =(0, ), inputFileNames='',inputFormat='', outputFileName= '', defaultValue=None, tag=None, valuesPerStrand= {}, outputTagNames=None, operation=None, size = None, overlap=None, verbosity = 0):
		self.strands		= strands
		self.inputFileNames = inputFileNames.split(',') if inputFileNames is not None else []
		self.inputFormat = inputFormat
		self.outputFileName = outputFileName
		self.defaultValue	= defaultValue
		self.tag 			= tag
		self.valuesPerStrand = valuesPerStrand
		self.outputTagNames = outputTagNames.split(",") if outputTagNames is not None else []
		self.operation = operation
		self.size = size
		self.overlap = overlap
		self.verbosity	    = verbosity

	def setAttributesFromCmdLine(self):
		description = "Clusterize by Sliding Windows v1.0.2: Produces a GFF3 file that clusters a list of transcripts using a sliding window. [Category: Sliding Windows]"
		epilog = "\n"
		parser = RepetOptionParser(description = description, epilog = epilog)
		parser.add_option("-i", "--input",        dest="inputFileName",  action="store",					 type="string", help="input files, separated by commas [compulsory] [format: string]")
		parser.add_option("-f", "--inputFormat",  dest="inputFormat",    action="store",					 type="string", help="format of the input file [compulsory] [format: transcript file format]")
		parser.add_option("-o", "--output",	      dest="outputFileName", action="store",					 type="string", help="output file [compulsory] [format: output file in transcript format given by -u]")
		parser.add_option("-u", "--outputFormat", dest="outputFormat",   action="store",      default="gff", type="string", help="format of the output file [format: transcript file format]")
		parser.add_option("-z", "--outputStyle" , dest="outputStyle",    action="store",      default=None,  type="string", help="additional formatting of the output file [triplet: title, feature, featurePart")

		parser.add_option("-s", "--size",         dest="size",	         action="store",					 type="int",	help="size of the regions [compulsory] [format: int]")
		parser.add_option("-e", "--overlap",	  dest="overlap",        action="store",					 type="int",	help="overlap between two consecutive regions [compulsory] [format: int]")
		parser.add_option("-g", "--tag",          dest="tag",            action="store",      default=None,  type="string", help="use a given tag as input (instead of summing number of features) [format: string]")	
		parser.add_option("-r", "--operation",    dest="operation",      action="store",      default=None,  type="string", help="combine tag value with given operation [format: choice (sum, avg, med, min, max)]")
		parser.add_option("-d", "--defaultValue", dest="defaultValue",   action="store",					 type="float",  help="default value for input tag [format: float]")
		parser.add_option("-w", "--write",	      dest="writeTag",       action="store",      default=None,  type="string", help="print the result in the given tags, separated by commas (default usually is 'nbElements') [format: string]")	
		parser.add_option("-2", "--strands",      dest="strands",		 action="store_true", default=False,                help="consider the two strands separately [format: bool] [default: false]")
		parser.add_option("-v", "--verbosity",    dest="verbosity",	     action="store",      default=1,	 type="int",	help="trace level [format: int] [default: 1]")
		(options, args) = parser.parse_args()
		self._setAttributesFromOptions(options)
		
	def _setAttributesFromOptions(self, options):
		self.verbosity = options.verbosity
		self.operation = options.operation
		
		self.tag = options.tag
		self.setOutputTag(None if options.writeTag is None else options.writeTag.split(","))
		self.inputFormat = options.inputFormat
		self.inputFileNames = options.inputFileName.split(",")
		
		[title,feature,featurePart] = options.outputStyle.split(",") if options.outputStyle is not None else ["S-MART","transcript","exon"]
		self.setOutputFileName(options.outputFileName, options.outputFormat, title, feature, featurePart)
		
		self.setWindowSize(options.size)
		self.setWindowOverlap(options.overlap)
		self.setDefaultValue(options.defaultValue)
		self.setBothStrands(options.strands)


	def setInputFile(self, fileNames, format):
		parserChooser = ParserChooser(self.verbosity)
		parserChooser.findFormat(format)
		if not isinstance(fileNames, list):
			fileNames = [fileNames]
		if not self.outputTagNames:
			if len(fileNames) == 1:
				self.outputTagNames = ["nbElements"]
			else:
				operation = "nbElements" if self.operation is None else self.operation.lower()
				self.outputTagNames = ["%s%s" % (operation, os.path.splitext(os.path.basename(fileName))[0].title()) for fileName in fileNames]
		
		self.parsers = dict(zip(self.outputTagNames, [parserChooser.getParser(fileName) for fileName in fileNames]))

	def setOutputFileName(self, fileName, format="gff", title="S-MART", feature="transcript", featurePart="exon"):
		writerChooser = WriterChooser(self.verbosity)
		writerChooser.findFormat(format)
		self.writer = writerChooser.getWriter(fileName)
		self.writer.setTitle(title)
		self.writer.setFeature(feature)
		self.writer.setFeaturePart(featurePart)

	def setWindowSize(self, size):
		self.size = size

	def setWindowOverlap(self, overlap):
		self.overlap = overlap

	def setTag(self, tag):
		self.tag = tag

	def setOperation(self, operation):
		self.operation = operation

	def setBothStrands(self, bothStrands):
		if bothStrands:
			self.strands = (-1, 1)

	def setOutputTag(self, tags):
		self.outputTagNames = tags
		
	def setDefaultValue(self, defaultValue):
		self.defaultValue = defaultValue

	def checkOptions(self):
		if self.operation != None and self.operation not in ("sum", "avg", "med", "min", "max"):
			raise Exception("Do not understand tag '%s'! Aborting..." % (self.operation))
		if self.size is None:
			raise Exception("Please mention a windows size! Aborting...")
		if self.overlap is None:
			raise Exception("Please mention a windows overlap size! Aborting...")

	def getBinsFromPos(self, pos):
		bin = (pos - 1) / (self.size - self.overlap)
		if bin >= 1 and pos <= bin * (self.size - self.overlap) + self.overlap:
			return (bin - 1, bin)
		return (bin, )

	def getPosFromBin(self, bin):
		return (bin * (self.size - self.overlap) + 1, bin * (self.size - self.overlap) + self.size)

	def getNbElements(self, transcript):
		nbOccurrences = 1 if "nbOccurrences" not in transcript.getTagNames() else transcript.getTagValue("nbOccurrences")
		nbElements	= 1 if "nbElements"	not in transcript.getTagNames() else transcript.getTagValue("nbElements")
		nbOccurrences = float(nbOccurrences)
		nbElements	= float(nbElements)
		nbElements   /= float(nbOccurrences)
		return nbElements

	def setBins(self, name, parser):
		progress = UnlimitedProgress(10000, "Setting bins of file %s" % (parser.fileName), self.verbosity)
		for transcript in parser.getIterator():
			nbElements = self.getNbElements(transcript)
			strand	   = transcript.getDirection() if len(self.strands) == 2 else 0
			chromosome = transcript.getChromosome()
			if strand not in self.valuesPerStrand:
				self.valuesPerStrand[strand] = {}
			if chromosome not in self.valuesPerStrand[strand]:
				self.valuesPerStrand[strand][chromosome] = {}
			for bin in self.getBinsFromPos(transcript.getStart()):
				if self.tag is None:
					value = nbElements
				else:
					if self.tag in transcript.getTagNames():
						value = float(transcript.getTagValue(self.tag))
					else:
						if self.defaultValue is None:
							raise Exception("Tag %s undefined in transcript %s" % (self.tag, transcript))
						value = self.defaultValue
				if bin not in self.valuesPerStrand[strand][chromosome]:
					self.valuesPerStrand[strand][chromosome][bin] = {}
				if name not in self.valuesPerStrand[strand][chromosome][bin]:
					self.valuesPerStrand[strand][chromosome][bin][name] = []
				self.valuesPerStrand[strand][transcript.getChromosome()][bin][name].append(value)
			progress.inc()
		progress.done()

	def aggregateData(self, values):
		if self.operation is None or self.operation == "sum":
			return sum(values)
		if self.operation == "avg":
			return sum(values) / float(len(values))
		if self.operation == "med":
			size		 = len(values)
			sortedValues = sorted(values)
			if size % 2 == 1:
				return sorted[(size - 1) / 2]
			return (sortedValues[size / 2 - 1] + sortedValues[size / 2]) / 2.0
		if self.operation == "min":
			return min(values)
		if self.operation == "max":
			return max(values)
		if self.operation == "GCpercent":
			subSequence = self.valuesPerStrand[strand][chromosome][bin]
			NPercent = 100 * (values.countNt("N") / float(values.getSize()))
			if NPercent >= 50:
				return "NA"
			return subSequence.getGCpercentageInSequenceWithoutCountNInLength()

	def printRegions(self):
		cpt		   = 1
		for strand in self.strands:
			for chromosome in sorted(self.valuesPerStrand[strand]):
				for bin in sorted(self.valuesPerStrand[strand][chromosome]):
					transcript = Transcript()
					transcript.setName("region%d" % cpt)
					transcript.setChromosome(chromosome)
					transcript.setStart(self.getPosFromBin(bin)[0])
					transcript.setEnd(self.getPosFromBin(bin)[1])
					transcript.setDirection(1 if strand == 0 else strand)
					for name in self.valuesPerStrand[strand][chromosome][bin]:
						transcript.setTagValue(name, self.aggregateData(self.valuesPerStrand[strand][chromosome][bin][name]))
						
					#transcript.setTagValue(self.outputTagNames[0], str(self.valuesPerStrand[strand][chromosome][bin]))
					#transcript.setTagValue("%s%s" % (self.operation, self.outputTagNames), self.aggregateData(self.valuesPerStrand[strand][chromosome][bin][name]))
					self.writer.addTranscript(transcript)
					cpt += 1
		self.writer.close()

	def run(self):
		self.checkOptions()
		self.setInputFile(self.inputFileNames,self.inputFormat)
		for key, parser in self.parsers.iteritems():
			self.setBins(key, parser)
		self.printRegions()

if __name__ == "__main__":
	iClusterizeBySlidingWindows = ClusterizeBySlidingWindows()
	iClusterizeBySlidingWindows.setAttributesFromCmdLine()
	iClusterizeBySlidingWindows.run()

	

	
