#! /usr/bin/env python
#
# Copyright INRA-URGI 2009-2010
# 
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
# 
# As a counterpart to the access to the source code and rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty and the software's author, the holder of the
# economic rights, and the successive licensors have only limited
# liability.
# 
# In this respect, the user's attention is drawn to the risks associated
# with loading, using, modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean that it is complicated to manipulate, and that also
# therefore means that it is reserved for developers and experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and, more generally, to use and operate it in the
# same conditions as regards security.
# 
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.
#
"""Get the repartition of some elements in a chromosomes"""

import os
from optparse import OptionParser
from SMART.Java.Python.structure.TranscriptContainer import TranscriptContainer
from SMART.Java.Python.structure.Transcript import Transcript
from commons.core.writer.Gff3Writer import Gff3Writer
from SMART.Java.Python.misc.RPlotter import RPlotter
from SMART.Java.Python.misc.Progress import Progress
from math import *

def divideKeyDict(dictionary, ratio):
	return dict([(key / ratio, dictionary[key]) for key in dictionary])


def setTranscript(chromosome, direction, start, end, sample, name, value):
	transcript = Transcript()
	transcript.setChromosome(chromosome)
	transcript.setDirection(direction)
	transcript.setStart(start)
	transcript.setEnd(end)
	transcript.setName(name)
	transcript.setTagValue("nbElements", value)
	transcript.setTagValue("sample", sample)
	return transcript




class GetDistribution(object):

	def __init__(self, verbosity = 0):
		self.verbosity         = verbosity
		self.referenceFileName = None
		self.sliceSize         = None
		self.nbBins            = None
		self.bothStrands       = False
		self.rawData           = False
		self.chromosome        = None
		self.normalization     = None
		self.start             = None
		self.end               = None
		self.yMin              = None
		self.yMax              = None
		self.width             = None
		self.height            = None
		self.gffFileName       = None
		self.csvFileName       = None

	def setInputFiles(self, fileNames, format):
		self.format         = format
		self.inputFileNames = fileNames

	def setOutputFile(self, fileName):
		self.outputFileName = fileName

	def setNames(self, names):
		names          = names.split(",") if names != None else ["%d" % (i+1) for i in range(len(self.inputFileNames))]
		self.fileNames = dict(zip(names, self.inputFileNames))

	def setReferenceFileName(self, fileName):
		self.referenceFileName = fileName

	def setBinSize(self, binSize):
		self.sliceSize = binSize

	def setNbBins(self, nbBins):
		self.nbBins = nbBins

	def setBothStrands(self, bothStrands):
		self.bothStrands = bothStrands

	def setRawData(self, rawData):
		self.rawData = rawData

	def setNormalization(self, factors):
		if factors == None:
			self.normalization = dict([name, 1.0] for name in self.fileNames)
		else:
			self.normalization = dict(zip(self.fileNames.keys(), [float(factor) for factor in factors.split(",")]))

	def setRegion(self, chromosome, start, end):
		self.chromosome = chromosome
		self.start      = None if start == None else int(start)
		self.end        = None if end == None else int(end)

	def setYWidth(self, yMin, yMax):
		self.yMin = None if yMin == None else int(yMin)
		self.yMax = None if yMax == None else int(yMax)

	def setImageSize(self, width, height):
		self.width  = None if width == None else int(width)
		self.height = None if height == None else int(height)

	def addGff(self, fileName):
		self.gffFileName = fileName

	def addCsv(self, fileName):
		self.csvFileName = fileName

	def _getSizesFromReferenceFile(self):
		referenceHandle = open(self.referenceFileName)
		name            = None
		size            = 0
		for line in referenceHandle:
			line = line.strip()
			if line == "": continue
			if line[0] == ">":
				if name != None:
					if self.verbosity > 10:
						print name
					sizes[name] = size
					maxSize	    = max(maxSize, size)
					size        = 0
				name = line[1:]
			else:
				size += len(line)
		sizes[name]  = size
		self.maxSize = max(maxSize, size)
		if self.verbosity > 10:
			print "done"
		self.start = 0
		self.end   = self.maxSize

	def _getSizesFromData(self):
		self.start = 0
		for fileName in self.fileNames.values():
			parser   = TranscriptContainer(fileName, self.format, self.verbosity)
			progress = Progress(parser.getNbTranscripts(), "First scan of %s" % (fileName), self.verbosity)
			for transcript in parser.getIterator():
				if transcript.__class__.__name__ == "Mapping":
					transcript = transcript.getTranscript()
				chromosome = transcript.getChromosome()
				end	       = transcript.getEnd()
				if chromosome not in self.sizes:
					self.sizes[chromosome] = end
				else:
					self.sizes[chromosome] = max(self.sizes[chromosome], end)
				self.maxSize = max(self.sizes[chromosome], self.maxSize)
				progress.inc()
			progress.done()

	def _findSizes(self):
		self.sizes   = {}
		self.maxSize = 0
		if self.chromosome != None or self.start != None or self.end != None:
			self.maxSize	            = self.end
			self.sizes[self.chromosome] = self.end
		elif self.referenceFileName != None:
			self._getSizesFromReferenceFile()
		else:
			self._getSizesFromData()
		if self.sliceSize == None:
			tmp1           = int(self.maxSize / float(self.nbBins))
			tmp2           = 10 ** (len("%d" % (tmp1))-2)
			self.sliceSize = int((tmp1 / tmp2) * tmp2)

	def _initializeBins(self):
		self.bins = {}
		if self.bothStrands:
			self.binsPlus  = {}
			self.binsMinus = {}
		for sample in self.fileNames:
			self.bins[sample] = {}
			if self.bothStrands:
				self.binsPlus[sample]  = {}
				self.binsMinus[sample] = {}
			for chromosome in self.sizes:
				self.bins[sample][chromosome] = dict([(i * self.sliceSize + 1, 0) for i in range(self.start / self.sliceSize, self.sizes[chromosome] / self.sliceSize + 1)])
				if self.bothStrands:
					self.binsPlus[sample][chromosome] = dict([(i * self.sliceSize + 1, 0) for i in range(self.start / self.sliceSize, self.sizes[chromosome] / self.sliceSize + 1)])
					self.binsMinus[sample][chromosome] = dict([(i * self.sliceSize + 1, 0) for i in range(self.start / self.sliceSize, self.sizes[chromosome] / self.sliceSize + 1)])
		if not self.rawData:
			self.density = {}
			if self.bothStrands:
				self.densityPlus  = {}
				self.densityMinus = {}
			for sample in self.fileNames:
				self.density[sample] = {}
				if self.bothStrands:
					self.densityPlus[sample]  = {}
					self.densityMinus[sample] = {}
				for chromosome in self.sizes:
					self.density[sample][chromosome] = dict([(i * self.sliceSize + 1, 0) for i in range(self.start / self.sliceSize, self.sizes[chromosome] / self.sliceSize + 1)])
					if self.bothStrands:
						self.densityPlus[sample][chromosome] = dict([(i * self.sliceSize + 1, 0) for i in range(self.start / self.sliceSize, self.sizes[chromosome] / self.sliceSize + 1)])
						self.densityMinus[sample][chromosome] = dict([(i * self.sliceSize + 1, 0) for i in range(self.start / self.sliceSize, self.sizes[chromosome] / self.sliceSize + 1)])

	def _populateBins(self):
		self.maxSlice = 0
		for sample, fileName in self.fileNames.iteritems():
			parser   = TranscriptContainer(fileName, self.format, self.verbosity)
			progress = Progress(parser.getNbTranscripts(), "Parsing %s" % (fileName), self.verbosity)
			for transcript in parser.getIterator():
				if transcript.__class__.__name__ == "Mapping":
					transcript = transcript.getTranscript()
				if self.chromosome == None or (transcript.getChromosome() == self.chromosome and transcript.getStart() >= self.start and transcript.getStart() <= self.end):
					bin = (transcript.getStart() / self.sliceSize) * self.sliceSize + 1
					if transcript.getDirection() == 1:
						self.binsPlus[sample][transcript.getChromosome()][bin] += self.normalization[sample]
					else:
						self.binsMinus[sample][transcript.getChromosome()][bin] -= self.normalization[sample]
					self.bins[sample][transcript.getChromosome()][bin] += self.normalization[sample]
					self.maxSlice = max(self.maxSlice, transcript.getStart() / self.sliceSize)
				progress.inc()
			progress.done()

	def _populateDensities(self):
		for sample in self.fileNames:
			for chromosome in bins[sample]:
				self.densityPlus[sample][chromosome] = dict([(bin, float(value) / self.sliceSize * self.sliceSize) for bin, value in self.binsPlus[sample][chromosome].iteritems()])
				if self.start % self.sliceSize != 0:
					self.densityPlus[sample][chromosome][(self.start / self.sliceSize) * self.sliceSize + 1] = float(self.binsPlus[sample][chromosome][(self.start / self.sliceSize) * self.sliceSize + 1]) / (self.sliceSize - (self.start % self.sliceSize)) * self.sliceSize
				if self.sizes[chromosome] % self.sliceSize != 0:
					self.densityPlus[sample][chromosome][(self.sizes[chromosome] / self.sliceSize) * self.sliceSize + 1] = float(self.binsPlus[sample][chromosome][(self.sizes[chromosome] / self.sliceSize) * self.sliceSize + 1]) / (self.sizes[chromosome] % self.sliceSize) * self.sliceSize
				self.densityMinus[sample][chromosome] = dict([(bin, float(value) / sliceSize * sliceSize) for bin, value in binsMinus[sample][chromosome].iteritems()])
				if self.start % self.sliceSize != 0:
					self.densityMinus[sample][chromosome][(start / sliceSize) * sliceSize + 1] = float(binsMinus[sample][chromosome][(start / sliceSize) * sliceSize + 1]) / (sliceSize - (start % sliceSize)) * sliceSize
				if self.sizes[chromosome] % self.sliceSize != 0:
					self.densityMinus[sample][chromosome][(self.sizes[chromosome] / self.sliceSize) * self.sliceSize + 1] = float(self.binsMinus[sample][chromosome][(self.sizes[chromosome] / self.sliceSize) * self.sliceSize + 1]) / (self.sizes[chromosome] % self.sliceSize) * self.sliceSize
				self.density[sample][chromosome] = dict([(bin, self.densityPlus[sample][chromosome][bin] - self.densityMinus[sample][chromosome][bin]) for bin in self.bins[sample][chromosome]])

	def _preparePlotter(self, chromosome, maxX):
		if maxX <= 1000:
			unit  = "nt."
			ratio = 1.0
		elif maxX <= 1000000:
			unit  = "kb"
			ratio = 1000.0
		else:
			unit  = "Mb"
			ratio = 1000000.0
		outputFileName = "%s_%s" % (self.outputFileName, chromosome)
		if self.start != None and self.end != None:
			outputFileName += ":%d-%d" % (self.start, self.end)
		outputFileName += ".png"
		plotter = RPlotter(outputFileName, self.verbosity)
		plotter.setXLabel("Position on %s (in %s)" % (chromosome.replace("_", " "), unit))
		plotter.setYLabel("# reads")
		if len(self.fileNames.keys()) > 1:
			plotter.setLegend(True, True)
		if self.bothStrands:
			plotter.setImageSize(1000, 300)
		else:
			plotter.setImageSize(1000, 200)
		if self.height != None:
			plotter.setHeight(self.height)
		if self.width != None:
			plotter.setWidth(self.width)
		if self.yMax != None:
			plotter.setMinimumY(self.yMin)
		if self.yMax != None:
			plotter.setMaximumY(self.yMax)
		return (plotter, ratio)

	def _plot(self):
		progress = Progress(len(self.sizes.keys()), "Plotting data", self.verbosity)
		for chromosome in self.sizes:
			maxX           = max([max(self.bins[sample][chromosome].keys()) for sample in self.fileNames])
			plotter, ratio = self._preparePlotter(chromosome, maxX)
			for sample in self.fileNames:
				if sample in self.bins and chromosome in self.bins[sample]:
					if self.rawData:
						if self.bothStrands :
							lines = [self.binsPlus[sample][chromosome], self.binsMinus[sample][chromosome]]
						else:
							lines = [self.bins[sample][chromosome]]
					else:
						if self.bothStrands :
							lines = [self.densityPlus[sample][chromosome], self.densityMinus[sample][chromosome]]
						else:
							lines = [self.density[sample][chromosome]]
					for line in lines:
						plotter.addLine(divideKeyDict(line, ratio), sample)
			plotter.plot()
			progress.inc()
		progress.done()

	def _printCsv(self):
		csvHandle = open(self.csvFileName, "w")
		for slice in range(self.start / self.sliceSize, self.maxSlice + 1):
			csvHandle.write(";%d-%d" % (slice * self.sliceSize + 1, (slice+1) * self.sliceSize))
		csvHandle.write("\n")
		for sample in self.fileNames:
			csvHandle.write("%s\n" % (sample))
			if self.bothStrands:
				for chromosome in self.sizes:
					if len(self.densityPlus[sample][chromosome]) > 0:
						csvHandle.write("%s [+]" % (chromosome))
						for slice in sorted(self.densityPlus[sample][chromosome].keys()):
							csvHandle.write(";%.2f" % (self.densityPlus[sample][chromosome][slice]))
						csvHandle.write("\n")			
					if len(self.densityMinus[sample][chromosome]) > 0:
						csvHandle.write("%s [-]" % (chromosome))
						for slice in sorted(self.densityPlus[sample][chromosome].keys()):
							csvHandle.write(";%.2f" % (-self.densityMinus[sample][chromosome][slice]))
						csvHandle.write("\n")			
			else:
				for chromosome in self.density[sample]:
					if len(self.density[sample][chromosome]) > 0:
						csvHandle.write(chromosome)
						for slice in sorted(self.density[sample][chromosome].keys()):
							csvHandle.write(";%.2f" % (self.density[sample][chromosome][slice]))
						csvHandle.write("\n")
		csvHandle.close()

	def _printGff(self):
			writer = Gff3Writer(self.gffFileName, self.verbosity)
			cpt    = 1
			if self.rawData:
				valuesPlus  = self.binsPlus
				valuesMinus = self.binsMinus
				values      = self.bins
			else:
				valuesPlus  = self.densityPlus
				valuesMinus = self.densityMinus
				values      = self.density
			for sample in self.fileNames:
				if self.bothStrands:
					for chromosome in values[sample]:
						for slice in valuesPlus[sample][chromosome]:
							writer.addTranscript(setTranscript(chromosome, 1, slice, slice + self.sliceSize, sample, "region%d" % (cpt), valuesPlus[chromosome][slice]))
							cpt += 1
						for slice in valuesMinus[chromosome]:
							writer.addTranscript(setTranscript(chromosome, -1, slice, slice + self.sliceSize, sample, "region%d" % (cpt), - valuesMinus[chromosome][slice]))
							cpt += 1
				else:
					for chromosome in values[sample]:
						for slice in values[sample][chromosome]:
							writer.addTranscript(setTranscript(chromosome, 1, slice, slice + self.sliceSize, sample, "region%d" % (cpt), values[chromosome][slice]))
							cpt += 1
			writer.write()
			
	def run(self):
		self._findSizes()
		self._initializeBins()
		self._populateBins()
		if not self.rawData:
			_populateDensities()
		self._plot()
		if self.csvFileName != None:
			self._printCsv()
		if self.gffFileName != None:
			self._printGff()

if __name__ == "__main__":
	
	# parse command line
	description = "Get Distribution v1.0.1: Get the distribution of the genomic coordinates on a genome. [Category: Visualization]"

	parser = OptionParser(description = description)
	parser.add_option("-i", "--input",         dest="inputFileNames",    action="store",						   type="string", help="input files, separated by commas [compulsory] [format: string]")
	parser.add_option("-f", "--format",        dest="format",	         action="store",						   type="string", help="format of the input file [compulsory] [format: transcript or mapping file format]")
	parser.add_option("-o", "--output",        dest="outputFileName",    action="store",						   type="string", help="output file [compulsory] [format: output file in GFF3 format]")
	parser.add_option("-r", "--reference",     dest="referenceFileName", action="store",	  default=None,		   type="string", help="file containing the genome [compulsory] [format: file in FASTA format]")
	parser.add_option("-a", "--names",         dest="names",             action="store",	  default=None,		   type="string", help="name of the samples [default: None] [format: string]")
	parser.add_option("-n", "--nbBins",        dest="nbBins",            action="store",	  default=1000,		   type="int",    help="number of bins [default: 1000] [format: int]")
	parser.add_option("-b", "--binSize",       dest="binSize",           action="store",	  default=None,		   type="int",    help="size of the bins [default: None] [format: int]")
	parser.add_option("-2", "--bothStrands",   dest="bothStrands",       action="store_true", default=False,                      help="plot one curve per strand [format: bool] [default: false]")
	parser.add_option("-w", "--raw",		   dest="raw",               action="store_true", default=False,                      help="plot raw number of occurrences instead of density [format: bool] [default: false]")
	parser.add_option("-l", "--normalization", dest="normalization",     action="store",      default=None,        type="string", help="normalization factors (floats separated by commas) [format: string]")
	parser.add_option("-c", "--chromosome",    dest="chromosome",        action="store",	  default=None,		   type="string", help="plot only a chromosome [format: string]")
	parser.add_option("-s", "--start",         dest="start",             action="store",	  default=None,		   type="int",    help="start from a given region [format: int]")
	parser.add_option("-e", "--end",		   dest="end",               action="store",	  default=None,		   type="int",    help="end from a given region [format: int]")
	parser.add_option("-y", "--yMin",          dest="yMin",              action="store",	  default=None,		   type="int",    help="minimum value on the y-axis to plot [format: int]")
	parser.add_option("-Y", "--yMax",          dest="yMax",              action="store",	  default=None,		   type="int",    help="maximum value on the y-axis to plot [format: int]")
	parser.add_option("-g", "--gff",		   dest="gffFileName",       action="store",      default=None,        type="string", help="also write GFF3 file [format: output file in GFF format] [default: None]")
	parser.add_option("-x", "--csv",           dest="csvFileName",       action="store",      default=None,        type="string", help="write a .csv file [format: output file in CSV format] [format: None]")
	parser.add_option("-H", "--height",        dest="height",            action="store",	  default=None,		   type="int",	  help="height of the graphics [format: int] [default: 300]")
	parser.add_option("-W", "--width",         dest="width",             action="store",	  default=None,		   type="int",	  help="width of the graphics [format: int] [default: 1000]")
	parser.add_option("-v", "--verbosity",     dest="verbosity",         action="store",	  default=1,		   type="int",	  help="trace level [default: 1] [format: int]")
	parser.add_option("-D", "--directory",     dest="working_Dir",       action="store",	  default=os.getcwd(), type="string", help="the directory to store the results [format: directory]")
	(options, args) = parser.parse_args()

	gd = GetDistribution(options.verbosity)
	gd.setInputFiles(options.inputFileNames.split(","), options.format)
	gd.setOutputFile(options.outputFileName)
	gd.setNames(options.names)
	gd.setReferenceFileName(options.referenceFileName)
	gd.setBinSize(options.binSize)
	gd.setNbBins(options.nbBins)
	gd.setBothStrands(options.bothStrands)
	gd.setNormalization(options.normalization)
	gd.setRawData(options.raw)
	gd.setRegion(options.chromosome, options.start, options.end)
	gd.setYWidth(options.yMin, options.yMax)
	gd.setImageSize(options.width, options.height)
	gd.addGff(options.gffFileName)
	gd.addCsv(options.csvFileName)
	gd.run()
