#!/usr/bin/env python

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import os
import subprocess
from commons.core.LoggerFactory import LoggerFactory
from commons.core.checker.ConfigChecker import ConfigChecker
from commons.core.checker.ConfigChecker import ConfigRules
from commons.core.seq.FastaUtils import FastaUtils
from commons.core.sql.DbFactory import DbFactory
from commons.core.utils.FileUtils import FileUtils
from commons.core.utils.RepetOptionParser import RepetOptionParser
from commons.core.seq.BioseqUtils import BioseqUtils
from commons.launcher.LaunchtRNAscanSE import LaunchtRNAscanSE

LOG_DEPTH = "commons.tools"
#LOG_FORMAT = "%(message)s"

## Detect TE features
#
class DetectTEFeatures(object):
    
    def __init__(self, fastaFileName = "", projectName = "", configFileName = "", doClean = False, verbosity = 0):
        self._fastaFileName = fastaFileName
        self._projectName = projectName
        self._configFileName = configFileName
        self._doClean = doClean
        self._verbosity = verbosity
        self._log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, self.__class__.__name__), self._verbosity)
        self._doTerminalRepeatSearch = True
        self._doTandemRepeatSearch = True
        self._doPolyAtailSearch = True
        self._doORFSearch = True
        self._doHmmer = True
        self._doTEBLRn = True
        self._doTEBLRtx = True
        self._doTEBLRx = True
        self._doHGBLRn = True
        self._dorDNABLRn = True
        self._dotRNASearch = True
        #TODO: doTEclass ? was difficult to install...
        self._blastType = "ncbi"
        self._profilesBank = ""
        self._knownTEnuclBank = ""
        self._knownTEprotBank = ""
        self._HGbank = ""
        self._rDNAbank = ""
        self._maxProfilesEvalue = "10"
        self._sectionName = "detect_features"
        
    def setAttributesFromCmdLine(self):
        usage = "DetectTEFeatures.py [options]"
        description = "Detect TE features on the input sequences file (fasta format).\n"
        description += "These features are terminal repeats, tandem repeats, poly A tail, ORF, profiles, TE in banks, rDNA, tRNA etc..."
        epilog = "\nExample 1: launch with default parameters\n"
        epilog += "\t$ DetectTEFeatures.py -i consensus.fa -P project -C TEclassifier.cfg\n"
        parser = RepetOptionParser(description = description, epilog = epilog, usage = usage)
        parser.add_option("-i", "--fasta",      dest = "fastaFileName", action = "store",       type = "string", help = "input fasta file name [compulsory] [format: fasta]", default = "")
        parser.add_option("-P", "--project",    dest = "projectName",   action = "store",       type = "string", help = "project name [compulsory] - not from config", default = "")
        parser.add_option("-C", "--config",     dest = "configFileName",action = "store",       type = "string", help = "configuration file name (e.g. TEdenovo.cfg)", default = "")
        parser.add_option("-c", "--clean",      dest = "doClean",       action = "store_true",                   help = "clean temporary files [optional] [default: False]", default = False)
        parser.add_option("-v", "--verbosity",  dest = "verbosity",     action = "store",       type = "int",    help = "verbosity [optional] [default: 1]", default = 1)
        options = parser.parse_args()[0]
        self._setAttributesFromOptions(options)
        
    def _setAttributesFromOptions(self, options):
        self.setFastaFileName(options.fastaFileName)
        self.setProjectName(options.projectName)
        self.setConfigFileName(options.configFileName)
        self.setDoClean(options.doClean)
        self.setVerbosity(options.verbosity)

    def _checkConfig(self):
        iConfigRules = ConfigRules()
        iConfigRules.addRuleSection(section=self._sectionName, mandatory=True)
        iConfigRules.addRuleOption(section=self._sectionName, option ="clean", mandatory=True, type="bool")
        iConfigRules.addRuleOption(section=self._sectionName, option ="term_rep", mandatory=True, type="bool")
        iConfigRules.addRuleOption(section=self._sectionName, option ="tand_rep", mandatory=True, type="bool")
        iConfigRules.addRuleOption(section=self._sectionName, option ="polyA", mandatory=True, type="bool")
        iConfigRules.addRuleOption(section=self._sectionName, option ="orf", mandatory=True, type="bool")
        iConfigRules.addRuleOption(section=self._sectionName, option ="TE_HMMER", mandatory=True, type="bool")
        iConfigRules.addRuleOption(section=self._sectionName, option ="TE_BLRn", mandatory=True, type="bool")
        iConfigRules.addRuleOption(section=self._sectionName, option ="TE_BLRtx", mandatory=True, type="bool")
        iConfigRules.addRuleOption(section=self._sectionName, option ="TE_BLRx", mandatory=True, type="bool")
        iConfigRules.addRuleOption(section=self._sectionName, option ="HG_BLRn", mandatory=True, type="bool")
        iConfigRules.addRuleOption(section=self._sectionName, option ="rDNA_BLRn", mandatory=True, type="bool")
        iConfigRules.addRuleOption(section=self._sectionName, option ="blast", mandatory=True, type="string", set = ("ncbi", "blastplus", "wu"))
        iConfigRules.addRuleOption(section=self._sectionName, option ="tRNA_scan", mandatory=True, type="bool")
        iConfigRules.addRuleOption(section=self._sectionName, option ="TE_nucl_bank", type="string")
        iConfigRules.addRuleOption(section=self._sectionName, option ="TE_prot_bank", type="string")
        iConfigRules.addRuleOption(section=self._sectionName, option ="HG_nucl_bank", type="string")
        iConfigRules.addRuleOption(section=self._sectionName, option ="rDNA_bank", type="string")
        iConfigRules.addRuleOption(section=self._sectionName, option ="TE_HMM_profiles", type="string")
        iConfigRules.addRuleOption(section=self._sectionName, option ="TE_HMMER_evalue", type="string")
        iConfigRules.addRuleOption(section=self._sectionName, option ="TRFmaxPeriod", mandatory=True, type="int")
        iConfigChecker = ConfigChecker(self._configFileName, iConfigRules)
        iConfig = iConfigChecker.getConfig()
        self._setAttributesFromConfig(iConfig)
        
    def _setAttributesFromConfig(self, iConfig):
        self.setDoClean(iConfig.get(self._sectionName, "clean"))
        self._doTerminalRepeatSearch = iConfig.get(self._sectionName, "term_rep")
        self._doTandemRepeatSearch = iConfig.get(self._sectionName, "tand_rep")
        self._doPolyAtailSearch = iConfig.get(self._sectionName, "polyA")
        self._doORFSearch = iConfig.get(self._sectionName, "orf")
        self._doHmmer = iConfig.get(self._sectionName, "TE_HMMER")
        self._doTEBLRn = iConfig.get(self._sectionName, "TE_BLRn")
        self._doTEBLRtx = iConfig.get(self._sectionName, "TE_BLRtx")
        self._doTEBLRx = iConfig.get(self._sectionName, "TE_BLRx")
        self._doHGBLRn = iConfig.get(self._sectionName, "HG_BLRn")
        self._dorDNABLRn = iConfig.get(self._sectionName, "rDNA_BLRn")
        self._dotRNASearch = iConfig.get(self._sectionName, "tRNA_scan")
        self._blastType = iConfig.get(self._sectionName, "blast")
        
        if self._doTEBLRn or self._doTEBLRtx:
            self._knownTEnuclBank = iConfig.get(self._sectionName, "TE_nucl_bank")
            self._knownTEnuclBank, lBanksNT = FileUtils.concatenateFileNamesFromString(self._knownTEnuclBank, splitSep = ",", joinSep = "_")
            if len(lBanksNT) > 1:
                FileUtils.catFilesFromList(lBanksNT, self._knownTEnuclBank)
                
        if self._doTEBLRx:
            self._knownTEprotBank = iConfig.get(self._sectionName, "TE_prot_bank")
            self._knownTEprotBank, lBanksAA = FileUtils.concatenateFileNamesFromString(self._knownTEprotBank, splitSep = ",", joinSep = "_")
            if len(lBanksAA) > 1:
                FileUtils.catFilesFromList(lBanksAA, self._knownTEprotBank)
        
        self._HGbank = iConfig.get(self._sectionName, "HG_nucl_bank")
        self._rDNAbank = iConfig.get(self._sectionName, "rDNA_bank")
        
        if self._doHmmer:
            self._profilesBank = iConfig.get(self._sectionName, "TE_HMM_profiles")
            self._profilesBank, lProfilesBanks = FileUtils.concatenateFileNamesFromString(self._profilesBank, splitSep = ",", joinSep = "_")
            if len(lProfilesBanks) > 1:
                FileUtils.catFilesFromList(lProfilesBanks, self._profilesBank)
        
        self._maxProfilesEvalue = iConfig.get(self._sectionName, "TE_HMMER_evalue")
        self._TRFmaxPeriod = iConfig.get(self._sectionName, "TRFmaxPeriod")
    
    def setFastaFileName(self, fastaFileName):
        self._fastaFileName = fastaFileName
        
    def setProjectName(self, projectName):
        self._projectName = projectName
        
    def setConfigFileName(self, configFileName):
        self._configFileName = configFileName
        
    def setDoClean(self, doClean):
        self._doClean = doClean
        
    def setVerbosity(self, verbosity):
        self._verbosity = verbosity
        
    def _checkOptions(self):
        if self._configFileName == "" or not FileUtils.isRessourceExists(self._configFileName):
            self._logAndRaise("ERROR: Missing configuration file")
        if self._fastaFileName == "" or not FileUtils.isRessourceExists(self._fastaFileName):
            self._logAndRaise("ERROR: Missing input fasta file")
            
    def _logAndRaise(self, errorMsg):
        self._log.error(errorMsg)
        raise Exception(errorMsg)
                    
    def run(self):
        LoggerFactory.setLevel(self._log, self._verbosity)
        self._checkOptions()
        self._checkConfig()
        self._log.info("START DetectTEFeatures")
        self._log.debug("Fasta file name: %s" % self._fastaFileName)
        
        self._detectFeatures()
        
        if self._doClean:
            lFilesToRemove = ["formatdb.log", "last_time_stamp.log"]
            FileUtils.removeFilesFromListIfExist(lFilesToRemove)
            FileUtils.removeFilesByPattern("*_cut*")
            FileUtils.removeFilesByPattern("*Nstretch.map")

        if not self._checkIfFeaturesTablesExist():
            self._logAndRaise("ERROR: no table recording TE features exists")
            
        self._log.info("END DetectTEFeatures")
        
    def _detectFeatures(self):
        if self._doTerminalRepeatSearch:
            self._detectTRsearch()
        if self._doPolyAtailSearch:
            self._detectPolyA()
        if self._doTandemRepeatSearch:
            self._detectTRF()
        if self._doORFSearch:
            self._detectORFs()
        if self._dotRNASearch:
            self._detectTRNA()
        if self._doTEBLRn:
            self._detectBlaster(self._knownTEnuclBank, "blastn", "TE_BLRn")
        if self._doTEBLRtx:
            self._detectBlaster(self._knownTEnuclBank, "tblastx", "TE_BLRtx")
        if self._doTEBLRx:
            self._detectBlaster(self._knownTEprotBank, "blastx", "TE_BLRx")
        if self._doHGBLRn:
            self._detectBlaster(self._HGbank, "blastn", "HG_BLRn")
        if self._dorDNABLRn:
            self._detectBlaster(self._rDNAbank, "blastn", "rDNA_BLRn")
        if self._doHmmer:
            self._detectHmmProfiles()
    
    def _detectTRsearch(self):
        self._log.info("Detect terminal repeats...")
        logFileName = "TRsearch.log"
        cmd = "TRsearch"
        cmd += " %s" % self._fastaFileName
        cmd += " > %s" % logFileName
        process = subprocess.Popen(cmd, shell = True)
        process.communicate()
        if process.returncode != 0:
            self._logAndRaise("ERROR when launching '%s'" % cmd)
        if self._doClean:
            if os.path.exists(logFileName):
                os.remove(logFileName)
        iDb = DbFactory.createInstance()
        iDb.createTable("%s_TR_set" % self._projectName, "set", "%s.TR.set" % self._fastaFileName, True)
        iDb.close()
    
    def _detectPolyA(self):
        self._log.info("Detect polyA tail...")
        logFileName = "polyAtail.log"
        cmd = "polyAtail"
        cmd += " %s" % self._fastaFileName
        cmd += " > %s" % logFileName
        process = subprocess.Popen(cmd, shell = True)
        process.communicate()
        if process.returncode != 0:
            self._logAndRaise("ERROR when launching '%s'" % cmd)
        if self._doClean:
            os.remove(logFileName)
        iDb = DbFactory.createInstance()
        iDb.createTable("%s_polyA_set" % self._projectName, "set", "%s.polyA.set" % self._fastaFileName, True)
        iDb.close()
        
    def _detectTRF(self):
        self._log.info("Detect SSR...")
        logFileName = "launchTRF.log"
        cmd = "LaunchTRF.py"
        cmd += " -i %s" % self._fastaFileName
        cmd += " -o %s.SSR.set" % self._fastaFileName
        cmd += " -m %d" % self._TRFmaxPeriod
        cmd += " -c"
        cmd += " -v 0"
        cmd += " > %s" % logFileName
        process = subprocess.Popen(cmd, shell = True)
        process.communicate()
        if process.returncode != 0:
            self._logAndRaise("ERROR when launching '%s'" % cmd)
        if self._doClean:
            os.remove(logFileName)
        iDb = DbFactory.createInstance()
        iDb.createTable("%s_SSR_set" % self._projectName, "set", "%s.SSR.set" % self._fastaFileName, True)
        iDb.close()
    
    def _detectORFs(self):
        self._log.info("Detect ORF...")
        FastaUtils.dbORF(self._fastaFileName, 10, 30, verbose = 0)
        iDb = DbFactory.createInstance()
        iDb.createTable("%s_ORF_map" % self._projectName, "map", "%s.ORF.map" % self._fastaFileName, True)
        iDb.close()
    
    def _detectTRNA(self):
        self._log.info("Detect tRNA...")
        iLtRNA = LaunchtRNAscanSE(self._fastaFileName, override = True, doClean = self._doClean, verbosity = 0)
        iLtRNA.run()
        iDb = DbFactory.createInstance()
        iDb.createTable("%s_tRNA_map" % self._projectName, "map", "%s.tRNA.map" % self._fastaFileName, True)
        iDb.close()
    
    def _detectBlaster(self, bank, blast, acronym):
        self._log.info("Detect known sequences (%s, %s)..." % (blast, bank))
        logFileName = "%s.log" % acronym
        if not os.path.exists("%s_cut" % bank):
            self._log.debug("prepare bank '%s'" % bank)
            prg = "blaster"
            cmd = prg
            cmd += " -s %s" % bank
            cmd += " -n %s" % blast
            if self._blastType == "ncbi": 
                cmd += " -N"
                cmd += " -p '-a 1'"
            elif self._blastType == "wu": 
                cmd += " -W"
                cmd += " -p -cpus=1"
            elif self._blastType == "blastplus":
                cmd += " -X"
                cmd += " -p '-num_threads 1'"
            cmd += " -r"
            cmd += " -P"
            cmd += " -v %i >> %s" % (self._verbosity - 2, logFileName)
            process = subprocess.Popen(cmd, shell = True)
            process.communicate()
            if process.returncode != 0:
                self._logAndRaise("ERROR when launching '%s'" % cmd)
            FileUtils.removeFilesByPattern("%s-%s-*.param" % (bank, blast))
            
        cmd = "blaster"
        cmd += " -q %s" % self._fastaFileName
        cmd += " -s %s" % bank
        cmd += " -B %s_%s_%s" % (self._fastaFileName, acronym, bank)
        cmd += " -n %s" % blast
        if self._blastType == "ncbi": 
            cmd += " -N"
            cmd += " -p '-a 1'"
        elif self._blastType == "wu": 
            cmd += " -W"
            cmd += " -p -cpus=1"
        elif self._blastType == "blastplus":
            cmd += " -X"
            cmd += " -p '-num_threads 1'"
        cmd += " -r"
        cmd += " -v 1 >> %s" % logFileName
        process = subprocess.Popen(cmd, shell = True)
        process.communicate()
        if process.returncode != 0:
            self._logAndRaise("ERROR when launching '%s'" % cmd)
        
        cmd = "matcher"
        cmd += " -m %s_%s_%s.align" % (self._fastaFileName, acronym, bank)
        #TODO: '.tab' files ('match' tables) are unused for now
#        cmd += " -q %s" % self._fastaFileName
#        cmd += " -s %s" % bank
        cmd += " -j"
        #TODO: clean conflicts AFTER 'join' procedure instead of BEFORE ?
#            cmd += " -x"
        cmd += " -v 1 >> %s" % logFileName
        process = subprocess.Popen(cmd, shell = True)
        process.communicate()
        if process.returncode != 0:
            self._logAndRaise("ERROR when launching '%s'" % cmd)
            
        if self._doClean:
            os.remove(logFileName)
            os.remove("%s_%s_%s.param" % (self._fastaFileName, acronym, bank))
            os.remove("%s_%s_%s.raw" % (self._fastaFileName, acronym, bank))
            os.remove("%s_%s_%s.seq_treated" % (self._fastaFileName, acronym, bank))
            os.remove("%s_%s_%s.align" % (self._fastaFileName, acronym, bank))
            os.remove("%s_%s_%s.align.clean_match.map" % (self._fastaFileName, acronym, bank))
            os.remove("%s_%s_%s.align.clean_match.param" % (self._fastaFileName, acronym, bank))
            
        iDb = DbFactory.createInstance()
        iDb.createTable("%s_%s_path" % (self._projectName, acronym), "path", "%s_%s_%s.align.clean_match.path" % (self._fastaFileName, acronym, bank), True)
        iDb.close()
        
    def _detectHmmProfiles(self):
        self._log.info("Detect profiles...")
        #NOTE: hmmer2 no more handle

        logFileName = "profiles.log"

        if not (os.path.exists("%s.h3m" % self._profilesBank) \
            and os.path.exists("%s.h3i" % self._profilesBank) \
            and os.path.exists("%s.h3f" % self._profilesBank) \
            and os.path.exists("%s.h3p" % self._profilesBank)) :
            self._log.debug("prepare bank '%s'" % self._profilesBank)
            cmd = "hmmpress"
            cmd += " -f"
            cmd += " %s" % self._profilesBank
            cmd += " >> %s" % logFileName
            process = subprocess.Popen(cmd, shell = True)
            process.communicate()
            if process.returncode != 0:
                raise Exception("ERROR when launching '%s'" % cmd)

        lBioseqs = BioseqUtils.extractBioseqListFromFastaFile(self._fastaFileName)
        lTranslatedBioseqs = BioseqUtils.translateBioseqListInAllFrames(lBioseqs)
        lTranslatedBioseqsWithoutStop = BioseqUtils.replaceStopCodonsByXInBioseqList(lTranslatedBioseqs)
        BioseqUtils.writeBioseqListIntoFastaFile(lTranslatedBioseqsWithoutStop, "%s_translated" % self._fastaFileName)
        
        cmd = "hmmscan"
        cmd += " -o %s_tr.hmmScanOut" % self._fastaFileName
        cmd += " --domtblout %s_tr.hmmScanOutTab" % self._fastaFileName
        cmd += " --noali -E %s" % self._maxProfilesEvalue
        cmd += " --cpu 1"
        cmd += " %s" % self._profilesBank
        cmd += " %s_translated" % self._fastaFileName
        process = subprocess.Popen(cmd, shell = True)
        process.communicate()
        if process.returncode != 0:
            raise Exception("ERROR when launching '%s'" % cmd)
        
        cmd = "HmmOutput2alignAndTransformCoordInNtAndFilterScores_script.py"
        cmd += " -i %s_tr.hmmScanOutTab" % self._fastaFileName
        cmd += " -o %s_Profiles_%s.align" % (self._fastaFileName, self._profilesBank)
        cmd += " -T %s" % self._fastaFileName
        cmd += " -p hmmscan"
        cmd += " -c"
        process = subprocess.Popen(cmd, shell = True)
        process.communicate()
        if process.returncode != 0:
            raise Exception("ERROR when launching '%s'" % cmd)
        
        cmd = "matcher"
        cmd += " -m %s_Profiles_%s.align" % (self._fastaFileName, self._profilesBank)
        cmd += " -j"
        cmd += " -E 10"
        cmd += " -L 0"
        cmd += " -v 1 >> %s" % logFileName
        process = subprocess.Popen(cmd, shell = True)
        process.communicate()
        if process.returncode != 0:
            raise Exception("ERROR when launching '%s'" % cmd)
            
        if self._doClean:
            os.remove(logFileName)
            os.remove("%s_translated" % self._fastaFileName)
            os.remove("%s_tr.hmmScanOut" % self._fastaFileName)
            os.remove("%s_Profiles_%s.align" % (self._fastaFileName, self._profilesBank))
            os.remove("%s_Profiles_%s.align.clean_match.map" % (self._fastaFileName, self._profilesBank))
            os.remove("%s_Profiles_%s.align.clean_match.param" % (self._fastaFileName, self._profilesBank))
            FileUtils.removeFilesByPattern("*.hmm.h3*")
             
        iDb = DbFactory.createInstance()
        iDb.createTable("%s_Profiles_path" % self._projectName, "path", "%s_Profiles_%s.align.clean_match.path" % (self._fastaFileName, self._profilesBank), True)
        iDb.close()
       
            
    def _checkIfFeaturesTablesExist(self):
        self._log.info("Check tables...")
        createdTablesNb = 0
        tableSuffix = ["_TR_set",
                       "_SSR_set",
                       "_polyA_set",
                       "_ORF_map",
                       "_Profiles_path",
                       "_TE_BLRn_path",
                       "_TE_BLRtx_path",
                       "_TE_BLRx_path",
                       "_HG_BLRn_path",
                       "_rDNA_BLRn_path",
                       "_tRNA_map"]
        iDb = DbFactory.createInstance()
        for suffix in tableSuffix:
            if iDb.doesTableExist(self._projectName + suffix):
                self._log.info("table '%s%s' exists (%i lines)" % (self._projectName, suffix, iDb.getSize(self._projectName + suffix)))
                createdTablesNb += 1
            else:
                self._log.warning("table '%s%s' doesn't exist" % (self._projectName, suffix))
        iDb.close()
        return createdTablesNb

if __name__ == "__main__":
    iLaunch = DetectTEFeatures()
    iLaunch.setAttributesFromCmdLine()
    iLaunch.run()