#!/usr/bin/env python

##@file FilterClassifiedSequences.py

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import sys
from commons.core.seq.Bioseq import Bioseq
from commons.core.utils.RepetOptionParser import RepetOptionParser
from commons.core.utils.FileUtils import FileUtils
from commons.core.seq.FastaUtils import FastaUtils
from commons.core.sql.DbFactory import DbFactory
from PASTEC.StatPastec import StatPastec

## Filter classified sequences (output data-bank from PASTEC)
#
class FilterClassifiedSequences(object):
    
    def __init__(self, inFastaFileName = "", isFilterSSR = False, maxLengthToFilterSSR = 0, isFilterHostGenes = False, isFilterChim = False, isFilterNoCat = False, minMemberMSA = 0, isFilterIncomp = False, isFilterRDNA = False, outFastaFileName = "", verbose = 0):
        self._inFastaFileName = inFastaFileName
        self._isFilterSSR = isFilterSSR
        self._maxLengthToFilterSSR = maxLengthToFilterSSR
        self._isFilterHostGenes = isFilterHostGenes
        self._isFilterChim = isFilterChim
        self._isFilterNoCat = isFilterNoCat
        self._minMemberMSA = minMemberMSA
        self._isFilterIncomp = isFilterIncomp
        self._isFilterRDNA = isFilterRDNA
        self._outFastaFileName = outFastaFileName
        self._verbose = verbose
        
    def setAttributesFromCmdLine(self):
        parser = RepetOptionParser()
        parser.add_option("-i", "--fasta",      dest = "inFastaFileName",     action = "store",     type = "string", help = "input fasta file",                                 default = "")
        parser.add_option("-S", "--ssr",        dest = "filterSSR",           action = "store_true",                 help = "filter the consensus classified as 'SSR'",         default = False)
        parser.add_option("-s", "--length",     dest = "maxLengthToFilterSSR",action = "store",     type = "int",    help = "length below which a SSR is filtered (e.g. 300, default=0 : all SSR are filtered)", default = 0)
        parser.add_option("-g", "--gene",       dest = "isFilterHostGenes",   action = "store_true",                 help = "filter the consensus classified as 'PotentialHostGene'", default = False)
        parser.add_option("-c", "--chim",       dest = "isFilterChim",        action = "store_true",                 help = "filter the consensus classified as 'PotentialChimeric'", default = False)
        parser.add_option("-N", "--nocat",      dest = "isFilterNoCat",       action = "store_true",                 help = "filter the consensus classified as 'noCat'", default = False)
        parser.add_option("-n", "--num",        dest = "minMemberMSA",        action = "store",     type = "int",    help = "minimum number of sequences in the MSA from which the noCat consensus has been built (default=0 : avoid)", default = 0)
        parser.add_option("-u", "--incomp",     dest = "isFilterIncomp",      action = "store_true",                 help = "filter the consensus classified as 'incomplete'", default = False)
        parser.add_option("-r", "--rdna",       dest = "isFilterRDNA",        action = "store_true",                 help = "filter the consensus classified as 'rDNA'", default = False)
        parser.add_option("-o", "--out",        dest = "outFastaFileName",    action = "store",     type = "string", help = "name of the output fasta file (default=inFileName+'.filtered')",                                 default = "")
        parser.add_option("-v", "--verbose",    dest = "verbose",             action = "store",     type = "int",    help = "verbosity level (default=0/1/2)",                  default = 0)
        options = parser.parse_args()[0]
        self._setAttributesFromOptions(options)
        
    def _setAttributesFromOptions(self, options):
        self.setInFastaName(options.inFastaFileName)
        self.setFilterSSR(options.filterSSR)
        self.setLengthSSR(options.maxLengthToFilterSSR)
        self.setFilterHostGenes(options.isFilterHostGenes)
        self.setFilterChim(options.isFilterChim)
        self.setIsFilterNoCat(options.isFilterNoCat)
        self.setMinMemberMSA(options.minMemberMSA)
        self.setIsFilterIncomp(options.isFilterIncomp)
        self.setIsFilterRdna(options.isFilterRDNA)
        self.setOutFastaFileName(options.outFastaFileName)
        self.setVerbose(options.verbose)
        
    def setInFastaName(self, inFastaFileName):
        self._inFastaFileName = inFastaFileName
        
    def setFilterSSR(self, filterSSR):
        self._isFilterSSR = filterSSR
        
    def setLengthSSR(self, lengthSSR):
        self._maxLengthToFilterSSR = lengthSSR
        
    def setFilterHostGenes(self, isFilterHostGenes):
        self._isFilterHostGenes = isFilterHostGenes
        
    def setFilterChim(self, isFilterChim):
        self._isFilterChim = isFilterChim
        
    def setIsFilterNoCat(self, isFilterNoCat):
        self._isFilterNoCat = isFilterNoCat
        
    def setMinMemberMSA(self, minMemberMSA):
        self._minMemberMSA = minMemberMSA
        
    def setIsFilterIncomp(self, isFilterIncomp):
        self._isFilterIncomp = isFilterIncomp
        
    def setIsFilterRdna(self, isFilterRDNA):
        self._isFilterRDNA = isFilterRDNA
        
    def setOutFastaFileName(self, outFastaFileName):
        self._outFastaFileName = outFastaFileName
        
    def setVerbose(self, verbose):
        self._verbose = verbose
        
    def checkAttributes(self):
        if self._inFastaFileName != "":
            if not FileUtils.isRessourceExists(self._inFastaFileName):
                raise Exception("ERROR: Input fasta file does not exist!")
        else:
            raise Exception("ERROR: No specified -i option!")
        
        if self._minMemberMSA < 0:
            raise Exception("ERROR: invalid value for -n option: %s" % self._minMemberMSA)
        
        if self._outFastaFileName == "":
            self._outFastaFileName = "%s.filtered" % self._inFastaFileName
            
    def filterClassifiedConsensus(self):
        inFile = open(self._inFastaFileName, "r")
        outFile = open(self._outFastaFileName, "w")
        bs = Bioseq()
        nbInSeq = 0
        nbRmv = 0
        
        while True:
            bs.read(inFile)
            if bs.header == None:
                break
            nbInSeq += 1
            if self._verbose > 1:
                print bs.header
                
            if self._isFilterSSR and ("SSR_" in bs.header or "SSR-" in bs.header) and ( self._maxLengthToFilterSSR == 0 or bs.getLength() <= self._maxLengthToFilterSSR ):
                nbRmv += 1
                if self._verbose > 1: print "\tfiltered SSR !"
                
            elif self._isFilterHostGenes and ("PotentialHostGene_" in bs.header or "PotentialHostGene-" in bs.header):
                nbRmv += 1
                if self._verbose > 1: print "\tfiltered PotentialHostGene !"
                
            elif self._isFilterChim and "-chim_" in bs.header:
                nbRmv += 1
                if self._verbose > 1: print "\tfiltered PotentialChimeric !"
                
            elif self._isFilterNoCat and ("noCat_" in bs.header or "noCat-" in bs.header):
                keep = False
                if self._minMemberMSA != 0:
                    algoMSA = ""
                    for i in ["Map", "MAP", "Malign", "Mafft", "Prank", "Clustalw", "Muscle", "Tcoffee"]:
                        if i in bs.header:
                            algoMSA = i
                    if algoMSA:
                        nbAlignSeq = int(bs.header.split(algoMSA)[1].split('\n')[0].split("_")[0])
                        if nbAlignSeq > self._minMemberMSA:
                            keep = True
                    else:
                        #Not TEdenovo-like headers classified as "noCat" are always deleted, regardless of the number of sequences used to compute them.
                        #TODO: If you want to do something else, place your code here.
                        #Example: add a switch in config to filter them or not.
                        keep = False 
                if keep:
                    bs.write( outFile )
                else:
                    nbRmv += 1
                    if self._verbose > 1: print "\tfiltered noCat !"
                    
            elif self._isFilterIncomp and ("-incomp_" in bs.header or "-incomp-" in bs.header):
                nbRmv += 1
                if self._verbose > 1: print "\tfiltered incomplete !"
                    
            elif self._isFilterRDNA and ("rDNA_" in bs.header or "rDNA-" in bs.header):
                nbRmv += 1
                if self._verbose > 1: print "\tfiltered rDNA !"
                
            else:
                bs.write(outFile)
                
        inFile.close()
        outFile.close()
        
        if self._verbose > 0:
            print "nb of input seq: %i" % nbInSeq
            print "nb of filtered seq: %i" % nbRmv
            sys.stdout.flush()
            
    def start(self):
        self.checkAttributes()
        if self._verbose > 0:
            msg = "START FilterClassifiedSequences.py"
            sys.stdout.write( "%s\n" % msg)
            print "input file: %s" % self._inFastaFileName
            print "output file: %s" % self._outFastaFileName
            if self._isFilterSSR:
                if self._maxLengthToFilterSSR == 0:
                    print "filter SSRs"
                else:
                    print "filter SSRs (<%ibp)" % self._maxLengthToFilterSSR
            if self._isFilterHostGenes:
                print "filter host's genes"
            if self._isFilterNoCat != "0":
                print "filter NoCat"
            if self._isFilterChim:
                print "filter confused"
            if self._isFilterIncomp:
                print "filter incomplete"
            if self._isFilterRDNA:
                print "filter rDNA"
            sys.stdout.flush()
            
    def end(self):
        if self._verbose > 0:
            msg = "END FilterClassifiedSequences.py"
            sys.stdout.write("%s\n" % msg)
            sys.stdout.flush()
            
    def postTraitement(self, fastaFileName):
        iDB=DbFactory.createInstance()
        classifTableName = "%s_consensus_classif" % fastaFileName.split('_denovoLibTEs_filtered.fa')[0]
        lSeqHeaders = []
        lSeqHeaders = FastaUtils.dbHeaders(fastaFileName)
        classifFileFromList_name = "classifFileFromList.classif"
        classifFileFromList = open(classifFileFromList_name, "w")
        for seqHeader in lSeqHeaders:
            sqlCmd = 'SELECT * FROM %s WHERE seq_name = "%s"' % (classifTableName, seqHeader)
            iDB.execute(sqlCmd)
            lFields = iDB.fetchall()[0]
            classifFileFromList.write("%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n" % (lFields[0],lFields[1],lFields[2],lFields[3],lFields[4],lFields[5],lFields[6],lFields[7]))
        classifFileFromList.close()
        iDB.close()
        iSP = StatPastec(inFileName = classifFileFromList_name)
        iSP.run()

    def run(self):
        self.start()
        self.filterClassifiedConsensus()
        self.postTraitement(self._outFastaFileName)
        self.end()
        
if __name__ == "__main__":
    i = FilterClassifiedSequences()
    i.setAttributesFromCmdLine()
    i.run()