#!/usr/bin/env python

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import os
import subprocess
from commons.core.LoggerFactory import LoggerFactory
from commons.core.checker.RepetException import RepetException
from commons.core.checker.CheckerUtils import CheckerUtils
from commons.core.utils.FileUtils import FileUtils
from commons.core.utils.RepetOptionParser import RepetOptionParser
from commons.core.coord.SlidingWindow import SlidingWindow
from commons.core.coord.MergeFromOverlaps import MergeFromOverlaps
from commons.core.seq.BioseqDB import BioseqDB
from commons.core.seq.SequenceModifications import SequenceModifications
from commons.core.seq.SequenceModificationsCollection import SequenceModificationsCollection

LOG_DEPTH = "repet.tools"

##Deletions detection based on coverage
#
class GetDeletionsFromCoverage(object):
    
    def __init__(self, fastaFileName = "", BAMFileName = "", outVCF = "", windowsLength = 200, overlap = 1, minSize = 500, thresholdN = 0.1, doClean = False, verbosity = 3):
        self._fastaFileName = fastaFileName
        self._BAMFileName = BAMFileName
        self.setOutVCF(outVCF)
        self._windowLength = windowsLength
        self._overlap = overlap
        self._minSize = minSize
        self._thresholdN = thresholdN
        
        self._thresholdDown = 0.1
        self._thresholdUp = 5
        
        self._coverageFileName = None
        self._dlCov = None
        
        self._doClean = doClean
        self._verbosity = verbosity
        self._log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, self.__class__.__name__), self._verbosity)
        
    def setAttributesFromCmdLine(self):
        usage = "GetDeletionsFromCoverage.py [options]"
        description = "Detect deletions using Depth Of Coverage method (DOC)."
        epilog = "\nExample 1: launch without verbosity.\n"
        epilog += "\t$ GetDeletionsFromCoverage.py -i genome.fa -b mapping.bam -v 0\n"
        parser = RepetOptionParser(description = description, epilog = epilog, usage = usage)
        parser.add_option("-i", "--fasta",      dest = "fastaFileName", action = "store",       type = "string", help = "input genome file name [compulsory] [format: fasta]", default = "")
        parser.add_option("-b", "--bam",        dest = "bamFileName",   action = "store",       type = "string", help = "input sorted BAM file name [compulsory] [format: sorted BAM]", default = "")
        parser.add_option("-o", "--out",        dest = "outVCF",        action = "store",       type = "string", help = "output VCF file name [optional] [default: <input_bam>.vcf]", default = "")
        parser.add_option("-w", "--window",     dest = "windowsLength", action = "store",       type = "int",    help = "sliding window length [optional] [default: 200]", default = 200)
        parser.add_option("-O", "--overlap",    dest = "overlap",       action = "store",       type = "int",    help = "overlap between sliding windows [optional] [default: 1]", default = 1)
        parser.add_option("-s", "--size",       dest = "minSize",       action = "store",       type = "int",    help = "minimal deletion size to detect [optional] [default: 500]", default = 500)
        parser.add_option("-n", "--nratio",     dest = "thresholdN",    action = "store",       type = "float",  help = "maximum unknown nucleotide coverage allowed [optional] [default: 0.1]", default = 0.1)
        parser.add_option("-c", "--clean",      dest = "doClean",       action = "store_true",                   help = "clean temporary files [optional] [default: False]", default = False)
        parser.add_option("-v", "--verbosity",  dest = "verbosity",     action = "store",       type = "int",    help = "verbosity [optional] [default: 3]", default = 3)
        options = parser.parse_args()[0]
        self._setAttributesFromOptions(options)
        
    def _setAttributesFromOptions(self, options):
        self.setFastaFileName(options.fastaFileName)
        self.setBAMFileName(options.bamFileName)
        self.setOutVCF(options.outVCF)
        self.setWindowLength(options.windowsLength)
        self.setOverlap(options.overlap)
        self.setMinSize(options.minSize)
        self.setThresholdN(options.thresholdN)
        self.setDoClean(options.doClean)
        self.setVerbosity(options.verbosity)

    def setFastaFileName(self, fastaFileName):
        self._fastaFileName = fastaFileName
        
    def setBAMFileName(self, bamFileName):
        self._BAMFileName = bamFileName
        
    def setOutVCF(self, outVCF):
        if outVCF == "" and self._BAMFileName:
            self._outVCF = "%s.vcf" % os.path.splitext(os.path.basename(self._BAMFileName))[0]
        else:
            self._outVCF = outVCF
        
    def setWindowLength(self, windowsLength):
        self._windowLength = windowsLength
        
    def setOverlap(self, overlap):
        self._overlap = overlap
        
    def setMinSize(self, minSize):
        self._minSize = minSize
        
    def setThresholdN(self, thresholdN):
        self._thresholdN = thresholdN
        
    def setDoClean(self, doClean):
        self._doClean = doClean
        
    def setVerbosity(self, verbosity):
        self._verbosity = verbosity
        
    def _checkOptions(self):
        if self._fastaFileName != "":
            if not FileUtils.isRessourceExists(self._fastaFileName):
                self._logAndRaise("ERROR: Genome file: '%s' does not exist!" % self._fastaFileName)
        else:
            self._logAndRaise("ERROR: Missing input fasta file name")
            
        if self._BAMFileName != "":
            if not FileUtils.isRessourceExists(self._BAMFileName):
                self._logAndRaise("ERROR: BAM file: '%s' does not exist!" % self._BAMFileName)
        else:
            self._logAndRaise("ERROR: Missing input BAM file name")
        
        if self._overlap >= self._windowLength:
            self._logAndRaise("ERROR: Windows overlaps must be lower than window size")
        
        if self._thresholdN < 0 or 1 < self._thresholdN :
            self._logAndRaise("ERROR: maximal unknown nucleotide ratio ('-n' option) must be in [0;1]")
            
        if not self._outVCF:
            self.setOutVCF("")
            
    def _logAndRaise(self, errorMsg):
        self._log.error(errorMsg)
        raise RepetException(errorMsg)
    
    # Algorithm far from optimum!
    def detectDeletions(self, iSeqModif):
        lCoverages = self._dlCov.get(iSeqModif.getOriginalHeader())
        genomeSize = len(lCoverages)
        iSlidingWindows = SlidingWindow(self._windowLength, self._overlap)
        meanCoverageInLastWindow = sum(lCoverages[iSlidingWindows.getStart() - 1:iSlidingWindows.getEnd()]) / float(iSlidingWindows.getLength())
        hasLowerCoverage = False
        hasHigherCoverage = False
        start = 1
        end = None
        
        while iSlidingWindows.getEnd() + self._windowLength - self._overlap < genomeSize:
            iSlidingWindows.slideWindowOnce()
            meanCoverageInWindow = sum(lCoverages[iSlidingWindows.getStart() - 1:iSlidingWindows.getEnd()]) / float(iSlidingWindows.getLength())
            
            if meanCoverageInWindow == 0:
                meanCoverageInWindow = 2.2250738585072014e-308

            if meanCoverageInLastWindow == 0:
                meanCoverageInLastWindow = 2.2250738585072014e-308
            
            hasLowerCoverage = meanCoverageInWindow / float(meanCoverageInLastWindow) < self._thresholdDown
            hasHigherCoverage = meanCoverageInWindow / float(meanCoverageInLastWindow) > self._thresholdUp
            if hasLowerCoverage:
                start = iSlidingWindows.getMiddle()
                
            if hasHigherCoverage:
                end = iSlidingWindows.getMiddle()
                if end > genomeSize:
                    end = genomeSize
                    
                iSeqModif.addDeletion(start, end)

            meanCoverageInLastWindow = meanCoverageInWindow
            
    def _loadCoverage(self):
        self._coverageFileName = "%s.coverage" % os.path.basename(self._BAMFileName)
        
        cmd = "bedtools genomecov -d -ibam %s | cut -f1,3 > %s" % (self._BAMFileName, self._coverageFileName)
        process = subprocess.Popen(cmd, shell = True)
        self._log.debug("Running : '%s'" % cmd)
        process.communicate()
        if process.returncode != 0:
            self._logAndRaise("ERROR when launching '%s'" % cmd)
        
        os.system("cat %s | cut -f1 | uniq > BAMsequencesNames.txt" % self._coverageFileName)
        with open("BAMsequencesNames.txt") as f:
            lSeqNames = [line.strip() for line in f.readlines()]
        
        self._dlCov = dict((seqName, []) for seqName in lSeqNames)
        
        self._log.debug("Loading coverage in memory")
        with open(self._coverageFileName) as f:
            for line in f:
                splittedLine = line.split()
                seqName = splittedLine[0]
                baseCov = int(splittedLine[1])
                self._dlCov.get(seqName).append(baseCov)
        
        if self._doClean:
            self._log.warning("Temporary files will be cleaned")
            os.remove("BAMsequencesNames.txt")
            os.remove(self._coverageFileName)
                
    def _loadCoverageSlim(self):
        self._coverageFileName = "%s.coverage" % os.path.basename(self._BAMFileName)

        cmd = "bedtools genomecov -bga -ibam %s | grep -w '[0]$'> %s" % (self._BAMFileName, self._coverageFileName)
        process = subprocess.Popen(cmd, shell = True)
        self._log.debug("Running : '%s'" % cmd)
        process.communicate()
        if process.returncode != 0:
            self._logAndRaise("ERROR when launching '%s'" % cmd)
            
    def detectDeletionsSlim(self, iSeqModif):
        [iSeqModif.addDeletion(int(line.split()[1])+1, int(line.split()[2])) for line in open(self._coverageFileName).readlines() if line.startswith(iSeqModif.getOriginalHeader())]
        
        if self._doClean:
            self._log.warning("Temporary files will be cleaned")
            os.remove(self._coverageFileName)

    def run(self):
        LoggerFactory.setLevel(self._log, self._verbosity)
        self._checkOptions()
        if not CheckerUtils.isExecutableInUserPath("bedtools"):
            self._logAndRaise("ERROR: 'bedtools' must be in your path")

        self._log.info("START GetDeletionsFromCoverage")
        self._log.debug("Fasta file name: %s" % self._fastaFileName)
        self._log.debug("BAM file name: %s" % self._BAMFileName)
        
        self._log.info("Computing base coverage from '%s'" % self._BAMFileName)
        self._loadCoverage()
#        self._loadCoverageSlim()

        iSeqModifCollection = SequenceModificationsCollection()

        iBSDB = BioseqDB(self._fastaFileName)
        
        self._log.info("Coverage list loaded; starting detection.")
        for iBS in iBSDB.db:
            iSeqModif = SequenceModifications(iBS.getHeader()) 
            
            self._log.debug("Current sequence name: '%s'" % iBS.getHeader())
            
            if iBS.getHeader() in self._dlCov.keys():
#                self.detectDeletionsSlim(iSeqModif)
                self.detectDeletions(iSeqModif)
            else:
                iSeqModif.addDeletion(1, iBS.getLength())

            self._log.debug("Detection done. Starting results cleaning.")
            
            self._log.debug("Merging overlapping results.")
            lClusteredRanges = MergeFromOverlaps(iSeqModif.getDeletions()).clusterize()
            
            iCleanSeqModif = SequenceModifications(iBS.getHeader())
            
            for iRange in lClusteredRanges:
                if iRange.getLength() >= self._minSize:
                    if iBS.subseq(iRange.getStart(), iRange.getEnd()).propNt("N") < self._thresholdN:
                        iCleanSeqModif.addDeletion(iRange.getStart(), iRange.getEnd())
            
            self._log.debug("Number of deletions in '%s' before cleaning: %i" % (iBS.getHeader(), len(iSeqModif.getDeletions())))
            self._log.debug("Number of deletions in '%s' after cleaning: %i" % (iBS.getHeader(), len(iCleanSeqModif.getDeletions())))
            
            iSeqModifCollection.add(iCleanSeqModif)
            
        iSeqModifCollection.sort()
        iSeqModifCollection.writeVCF(self._outVCF, self._fastaFileName, "GetDeletionsFromCoverage")

        self._log.info("END GetDeletionsFromCoverage")

if __name__ == "__main__":
    iGetDeletionsFromCoverage = GetDeletionsFromCoverage()
    iGetDeletionsFromCoverage.setAttributesFromCmdLine()
    iGetDeletionsFromCoverage.run()        