#!/usr/bin/env python

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import os
import time
import shutil
import subprocess
from commons.core.LoggerFactory import LoggerFactory
from commons.core.checker.CheckerUtils import CheckerUtils
from commons.core.utils.FileUtils import FileUtils
from commons.core.utils.RepetOptionParser import RepetOptionParser

LOG_DEPTH = "commons.launcher"

##Launches Velvet-Columbus assembler
#
class LaunchVelvet(object):
    #TODO: add advanced options string for velveth and velvetg
    #TODO: add possibility to scaffold contigs, either by Velvet or by SSPace
    
    def __init__(self, refGenomeFileName = "", outFileName = "", BAMFileName = "", kmerSize = 25, nbThreads = 1, doClean = False, verbosity = 3):
        self._refGenomeFileName = refGenomeFileName
        self.setOutFileName(outFileName)
        self._BAMFileName = BAMFileName
        self._kmerSize = kmerSize
        
        self._nbThreads = nbThreads
        self._doClean = doClean
        self._verbosity = verbosity
        self._log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, self.__class__.__name__), self._verbosity)
        
    def setAttributesFromCmdLine(self):
        usage = "LaunchVelvet.py [options]"
        description = "Launch Velvet-Columbus reference-guided assembler.\n"
        epilog = "\nExample 1: launch with info log level, overriding K-mer size, using 10 threads and keeping temporary files.\n"
        epilog += "\t$ LaunchVelvet.py -i refGenome.fa -b mapping.bam -k 31 -t 10 -v 3\n"
        epilog += "IMPORTANT NOTE ABOUT THREADS: to use several threads, Velvet must have been compiled with the 'OPENMP=1' flag.\n"
        epilog += "For an unknown reason, when multithreading is activated and only 1 thread is required, Velvet seems to run indefinitely.\n"
        epilog += "If you are in this case, you have to use at least 2 threads.\n"
        parser = RepetOptionParser(description = description, epilog = epilog, usage = usage)
        parser.add_option("-i", "--ref",        dest = "refGenome",     action = "store",       type = "string", help = "reference genome [compulsory] [format: fasta]", default = "")
        parser.add_option("-o", "--out",        dest = "outFileName",   action = "store",       type = "string", help = "output contigs file name [default: <input>_contigs.fa]", default = "")
        parser.add_option("-b", "--bam",        dest = "bamFileName",   action = "store",       type = "string", help = "reads mapping file [compulsory] [format: BAM]", default = "")
        parser.add_option("-k", "--kmer",       dest = "kmerSize",      action = "store",       type = "int",    help = "K-mer size [optional] [default: 25]", default = 25)
        parser.add_option("-t", "--thread",     dest = "nbThreads",     action = "store",       type = "int",    help = "Number of threads used by Velvet [optional] [default: 1]", default = 1)
        parser.add_option("-c", "--clean",      dest = "doClean",       action = "store_true",                   help = "clean temporary files [optional] [default: False]", default = False)
        parser.add_option("-v", "--verbosity",  dest = "verbosity",     action = "store",       type = "int",    help = "verbosity [optional] [default: 1]", default = 3)
        options = parser.parse_args()[0]
        self._setAttributesFromOptions(options)
        
    def _setAttributesFromOptions(self, options):
        self.setRefGenomeFileName(options.refGenome)
        self.setOutFileName(options.outFileName)
        self.setBAMFileName(options.bamFileName)
        self.setKmerSize(options.kmerSize)
        self.setNbThreads(options.nbThreads)
        self.setDoClean(options.doClean)
        self.setVerbosity(options.verbosity)

    def setRefGenomeFileName(self, refGenome):
        self._refGenomeFileName = refGenome
        
    def setOutFileName(self, outFileName):
        if outFileName == "":
            self._outFileName = "%s/%s_contigs.fa" % (os.getcwd(), os.path.splitext(os.path.basename(self._refGenomeFileName))[0])
        else:
            self._outFileName = outFileName
        
    def setBAMFileName(self, bamFileName):
        self._BAMFileName = bamFileName
        
    def setKmerSize(self, kmerSize):
        self._kmerSize = kmerSize
        
    def setNbThreads(self, nbThreads):
        self._nbThreads = nbThreads
        
    def setDoClean(self, doClean):
        self._doClean = doClean
        
    def setVerbosity(self, verbosity):
        self._verbosity = verbosity
        
    def _checkOptions(self):
        if self._refGenomeFileName == "":
            self._logAndRaise("ERROR: Missing input reference genome")
        else:
            if FileUtils.isRessourceExists(self._refGenomeFileName):
                self._refGenomeFileName = os.path.abspath(self._refGenomeFileName)
                self._log.info("Using '%s' as reference genome" % self._refGenomeFileName)
            else:
                self._logAndRaise("Reference genome file '%s' does not exist!" % self._refGenomeFileName)
                
        if self._BAMFileName == "":
            self._logAndRaise("Missing reads mapping file")
        else:
            if FileUtils.isRessourceExists(self._BAMFileName):
                self._BAMFileName = os.path.abspath(self._BAMFileName)
                self._log.info("Using '%s' as reads mapping file" % self._BAMFileName)
            else:
                self._logAndRaise("Reads mapping file '%s' does not exist!" % self._BAMFileName)
        
        self._kmerSize = int(self._kmerSize)
        if self._kmerSize % 2 == 0:
            self._log.warning("K-mer size must be an odd number: using '%i' instead of '%i'" % (self._kmerSize + 1, self._kmerSize))
            self._kmerSize += 1
        self._log.info("Using %i as a value for K" % self._kmerSize)
        
        self._nbThreads = int(self._nbThreads)
        if self._nbThreads < 1:
            self._log.warning("Bad number of threads '%i'. Using only 1 thread." % self._nbThreads)
            self._nbThreads = 1
        s = ""
        if self._nbThreads > 1:
            s = "s"
        self._log.info("Using %i thread%s" % (self._nbThreads, s))
            
    def _logAndRaise(self, errorMsg):
        self._log.error(errorMsg)
        raise Exception(errorMsg)
    
    def launchVelveth(self):
        advancedOptions = ""
        
        cmd = "export OMP_THREAD_LIMIT=%s; export OMP_NUM_THREADS=%s; velveth " % (self._nbThreads, self._nbThreads)
        cmd += "velvet_k_%i " % self._kmerSize
        cmd += "%i " % self._kmerSize 
        cmd += "-reference %s " % self._refGenomeFileName
        cmd += "-shortPaired "
        cmd += "-bam %s " % self._BAMFileName
        cmd += "%s 2>&1 > velveth.log" % advancedOptions
        
        process = subprocess.Popen(cmd, shell = True)
        self._log.debug("Running : '%s'" % cmd)
        process.communicate()
        if process.returncode != 0:
            self._logAndRaise("ERROR when launching '%s'" % cmd)

    def launchVelvetg(self):
        advancedOptions = ""
        
        cmd = "export OMP_THREAD_LIMIT=%s; export OMP_NUM_THREADS=%s; velvetg " % (self._nbThreads, self._nbThreads)
        cmd += "velvet_k_%i " % self._kmerSize
        cmd += "-ins_length auto "
        cmd += "-scaffolding no "
        cmd += "-min_contig_lgth 100 "
        cmd += "-exp_cov auto "
        cmd += "%s 2>&1 > velvetg.log" % advancedOptions
        
        process = subprocess.Popen(cmd, shell = True)
        self._log.debug("Running : '%s'" % cmd)
        process.communicate()
        if process.returncode != 0:
            self._logAndRaise("ERROR when launching '%s'" % cmd)
                    
    def run(self):
        LoggerFactory.setLevel(self._log, self._verbosity)
        self._log.info("START LaunchVelvet")
        if not CheckerUtils.isExecutableInUserPath("velveth"):
            self._logAndRaise("ERROR: velveth must be in your path")
        if not CheckerUtils.isExecutableInUserPath("velvetg"):
            self._logAndRaise("ERROR: velvetg must be in your path")
        self._checkOptions()
        
        self._workingDir = "%s/LaunchVelvet_%s" % (os.getcwd(), time.strftime("%Y%m%d%H%M%S"))
        os.makedirs(self._workingDir)
        os.chdir(self._workingDir)
        
        for fileName in [self._refGenomeFileName, self._BAMFileName]:
            os.symlink(fileName, os.path.basename(fileName))
        
        self._refGenomeFileName = os.path.abspath(os.path.basename(self._refGenomeFileName))
        self._BAMFileName = os.path.abspath(os.path.basename(self._BAMFileName))
        
        self._log.info("Hashing genome...")
        self.launchVelveth()
        self._log.info("Generating De Bruijn graph and assembly...")
        self.launchVelvetg()
        
        self._log.debug("Moving output contigs file to '%s'" % self._outFileName)
        os.chdir("..")
        outputDirectory = "%s/velvet_k_%s" % (self._workingDir, self._kmerSize)
        shutil.move("%s/contigs.fa" % outputDirectory, self._outFileName)
        
        if self._doClean:
            self._log.warning("Temporary files will be cleaned")
            shutil.rmtree(self._workingDir)
            
        self._log.info("END LaunchVelvet")

if __name__ == "__main__":
    iLaunchVelvet = LaunchVelvet()
    iLaunchVelvet.setAttributesFromCmdLine()
    iLaunchVelvet.run()