#!/usr/bin/env python

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import os
import subprocess
from commons.core.LoggerFactory             import LoggerFactory
from commons.core.utils.FileUtils           import FileUtils
from commons.core.utils.RepetOptionParser   import RepetOptionParser
from commons.core.checker.ConfigChecker     import ConfigRules
from commons.core.checker.ConfigChecker     import ConfigChecker
from commons.core.checker.CheckerUtils      import CheckerUtils
from commons.core.checker.RepetException    import RepetException
from commons.core.seq.BioseqDB import BioseqDB
import shutil
from operator import attrgetter
from commons.tools.LaunchBlaster import LaunchBlaster
from commons.launcher.LaunchMatcher import LaunchMatcher
from commons.core.coord.PathUtils import PathUtils
from commons.core.coord.Path import Path
from commons.core.coord.Range import Range
from commons.core.parsing.FastaParser import FastaParser

LOG_DEPTH = "commons.tools"

class NoCatBestHitClassifier(object):
    
    def __init__(self, fastaFileName = "", classifFileName="", outFileName = "", doClean = False, verbosity = 3):
        self._inFastaFileName = fastaFileName
        self._inClassifFileName = classifFileName
        self._outFastaFileName = outFileName
        self._doClean = doClean
        self._verbosity = verbosity
        self._log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, self.__class__.__name__), self._verbosity)
        
    def setAttributesFromCmdLine(self):
        usage = "NoCatBestHitClassifier.py [options]"
        description = "Add information to noCat classified sequences. These one are compared with valid classified sequences.\n"
        description += "The best hit information is added to the noCat sequence header in fasta file and in classif file\n"
        description += "eg: In fasta and classif files the sequence header is noCat_DmelCaf1_2_2-B-G143-Map12 will become noCat-RLX-like_DmelCaf1_2_2-B-G143-Map12\n"
        description += "that means this noCat has the best hit with an RLX sequence\n"
        description += "In classif file, the 'like' field will be created like=(bestHitName:RLX-incomp_DmelCaf1_2_2-L-B230-Map1_reversed CI:78) with RLX sequence features\n"
        epilog = "Use case 1: launch without verbosity and keep temporary files.\n"
        epilog += "\t$ NoCatBestHitClassifier.py -i file.fa -v 0\n"
        epilog += "\n"
        epilog += "Use case 2: launch with verbosity to have errors (level 1) and basic information (level 2), and delete temporary files.\n"
        epilog += "\t$ NoCatBestHitClassifier.py -i file.fa -c -v 2\n"
        parser = RepetOptionParser(description = description, epilog = epilog, usage = usage)
        parser.add_option("-i", "--fasta",      dest = "fastaFileName", action = "store",       type = "string", help = "input fasta file name [compulsory] [format: fasta]", default = "")
        parser.add_option("-o", "--out",        dest = "outFileName",   action = "store",       type = "string", help = "output fasta file name [optional] [default: <input>.fa][format: fasta]", default = "")
        parser.add_option("-C", "--classif",    dest = "classifFileName",action = "store",       type = "string", help = "classif file name [compulsory] (e.g. <input>.classif)", default = "")
        parser.add_option("-c", "--clean",      dest = "doClean",       action = "store_true",                   help = "clean temporary files [optional] [default: False]", default = False)
        parser.add_option("-v", "--verbosity",  dest = "verbosity",     action = "store",       type = "int",    help = "verbosity [optional] [default: 3]", default = 3)
        options = parser.parse_args()[0]
        self._setAttributesFromOptions(options)
        
    def _setAttributesFromOptions(self, options):
        self.setFastaFileName(options.fastaFileName)
        self.setOutFileName(options.outFileName)
        self.setClassifFileName(options.classifFileName)
        self.setDoClean(options.doClean)
        self.setVerbosity(options.verbosity)

    def setFastaFileName(self, FileName):
        self._inFastaFileName = FileName
        
    def setOutFileName(self, FileName):
        self._outFastaFileName = FileName
        
    def setClassifFileName(self, FileName):
        self._inClassifFileName = FileName
        
    def setDoClean(self, doClean):
        self._doClean = doClean
        
    def setVerbosity(self, verbosity):
        self._verbosity = verbosity
        
    def _checkOptions(self):
        if self._inFastaFileName == "":
            self._logAndRaise("Missing input fasta file name")
        else:
            if FileUtils.isRessourceExists(self._inFastaFileName):
                self._inFastaFileName = os.path.abspath(self._inFastaFileName)
            else:
                self._logAndRaise("Input fasta file '%s' does not exist!" % self._inFastaFileName)
        if self._inClassifFileName == "":
            self._logAndRaise("Missing input classif file name")
        else:
            if FileUtils.isRessourceExists(self._inClassifFileName):
                self._inClassifFileName = os.path.abspath(self._inClassifFileName)
            else:
                self._logAndRaise("Input classif file '%s' does not exist!" % self._inClassifFileName)
        if self._outFastaFileName == "":
            prefixe = os.path.basename(self._inFastaFileName)
            prefixe = os.path.splitext(prefixe)
            self._outFastaFileName = "%s_noCatBestHit%s" % (prefixe[0],prefixe[1])
            
    def _logAndRaise(self, errorMsg):
        self._log.error(errorMsg)
        raise RepetException(errorMsg)
    
    def _analyzeInputClassifFile(self):
        self._log.info("Started analyzing input classification file")
        self._dClassifOk = {}
        self._lClassifOk = []
        
        self._dClassifChim = {}
        self._lClassifChim = []
        
        self._dClassifNoCat = {}
        self._lClassifNoCat= []
        
        inClassifFile = open(self._inClassifFileName, "r")
        dName2Classif = {}
        line = inClassifFile.readline()
        totalClassifs = 0
        
        while True:
            if line == "":
                break
            data = line.split("\t")
            name = data[0]
            isChim = (data[3] != "ok")
            isNoCat = (data[4] == "noCat")
            if isNoCat and not isChim:
                self._dClassifNoCat[name] = data;
                self._lClassifNoCat.append(line)
            elif isChim:
                self._dClassifChim[name] = data;
                self._lClassifChim.append( line )
            else:
                self._dClassifOk[name] = data
                self._lClassifOk.append(line)
            
            if not dName2Classif.has_key(name):
                dName2Classif[name] = data
            else:
                self._logAndRaise("*ERROR: two consensus have the same name")
            
            line = inClassifFile.readline()
            totalClassifs += 1
        inClassifFile.close()
        
        self._log.info("Number of Ok: %d, Number of NoCats %d, Number of Chimeric: %d , Total: %d" % (len(self._lClassifOk), len(self._lClassifNoCat), len(self._lClassifChim), totalClassifs))
        self._log.info("Finished analyzing input classification file")
    
    def _getSequencesByHeaders(self):
        self._log.info("Started fetching sequences by pre-selected headers")

        uniqDB = BioseqDB(self._inFastaFileName)
        #Step2:#get sequences of headers
        lBioseqClassifOk = uniqDB.fetchList(self._dClassifOk.keys())
        bioseqDbOk = BioseqDB()
        bioseqDbOk.setData( lBioseqClassifOk )
        bioseqDbOk.save("classifOk.fa")
        
        lBioseqClassifNoCat = uniqDB.fetchList(self._dClassifNoCat.keys())
        bioseqDbNoCat = BioseqDB()
        bioseqDbNoCat.setData( lBioseqClassifNoCat)
        bioseqDbNoCat.save("classifNoCat.fa")
        self._log.info("Finished fetching sequences by pre-selected headers")
    
    def _alignFiles(self):
        self._log.info("Started aligning noCats with valid classifications")
        #step3: LaunchBlaster
        iLaunchBlaster = LaunchBlaster("classifNoCat.fa", "classifOk.fa",doAllByall=True, doClean=True)# sensitivity, evalue, identity, length, doAllByall, type, nbCPU, program, extraParams, doClean, verbosity)
        iLaunchBlaster.run()
        
        iLaunchMatcher = LaunchMatcher("classifNoCat.fa.align", "classifNoCat.fa", "classifOk.fa", doJoin=True, cleanConflicts=True, keepConflict=True, doClean=True)#, keepConflict, prefix, doClean, verbosity)
        iLaunchMatcher.run()
        self._log.info("finished aligning noCats with valid classifications")

    def sumScoreFromPathListOnSamePath(self, pathList):
        newPath = Path()
        if len(PathUtils.getListOfDistinctIdentifiers(pathList)) != 1:
            self._log.warning("The pathList contains several pathId")
        else: 
            pathListLength = len(pathList)
            newQueryStart, newQueryEnd = PathUtils.getQueryMinMaxFromPathList(pathList)
            newSubjectStart, newSubjectEnd = PathUtils.getSubjectMinMaxFromPathList(pathList)
            pathId = pathList[0].id
            queryName = pathList[0].range_query.seqname
            subjectName = pathList[0].range_subject.seqname
            sumScore = 0
            for i in range(0, pathListLength):
                sumScore += pathList[i].score      
            newPath = Path(pathId, Range(queryName, newQueryStart, newQueryEnd), Range(subjectName, newSubjectStart, newSubjectEnd), 0, sumScore, 0)
        return  newPath   

    def findBestPath_basedOnSubjectName(self, dPathQueryNameToPath, dPathIdToNewLength):
        lNewClassificationPaths = []
        
        for queryName, pathList in dPathQueryNameToPath.items():
            if len(pathList) == 1:
                lNewClassificationPaths.append(pathList[0])
            else:
#                pathList sorted by decreased score
                sortedItems = sorted(pathList, key=attrgetter('score'), reverse=True)
#                manage egality of scores
                currentForBestPath = sortedItems.pop(0)
                for path in sortedItems:
                    if path == currentForBestPath:
                        currentForBestPath = path
                    else:
                        if (path.score == currentForBestPath.score) and (dPathIdToNewLength[path.id] > dPathIdToNewLength[currentForBestPath.id]):
                            currentForBestPath = path
                bestPath = currentForBestPath
                lNewClassificationPaths.append(bestPath)
        return lNewClassificationPaths

    def _classifyAlignements(self):
        self._log.info("Started classifying alignments")
        #step4 : parse & analyze path file
        self._lOutputClassifs = []
        
        lPaths = PathUtils.getPathListFromFile("classifNoCat.fa.align.match.path")
        
        lModifiedNoCat = []
        lUnModifiedNoCat = []
        dOldToNewHeaders = {}

#        Calculate cumulative score for each pathId with several match part
        lSeqNames = PathUtils.getListOfDistinctQueryNames(lPaths)
        lPathIds = PathUtils.getListOfDistinctIdentifiers(lPaths)
        dPathIdToNewLength = dict((int(pathId), 0) for pathId in lPathIds)
        dIdpathToPath = PathUtils.getDictOfListsWithIdAsKey(lPaths)
        self._dPathQueryNameToPath = dict((seqName, []) for seqName in lSeqNames)
        for pathId, pathList in dIdpathToPath.items():
            if len(pathList) > 1:
                newPath = self.sumScoreFromPathListOnSamePath(pathList)
                if newPath.id != -1:
                    self._dPathQueryNameToPath[newPath.range_query.seqname].append(newPath)
            elif len(pathList) == 1:
                self._dPathQueryNameToPath[pathList[0].range_query.seqname].append(pathList[0])
            dPathIdToNewLength[pathId]= PathUtils.getLengthOnQueryFromPathList(pathList)

#        Build Best Path list from 
        self._lNewClassificationPaths = []       
        self._lNewClassificationPaths = self.findBestPath_basedOnSubjectName(self._dPathQueryNameToPath, dPathIdToNewLength)
        
        
        for path in self._lNewClassificationPaths:
            refName = path.range_subject.seqname
            refWickerCode = refName.split("_")[0]
            refWickerCode = refWickerCode.split("-comp")[0]
            refWickerCode = refWickerCode.split("-incomp")[0]
            
            queryName = path.range_query.seqname
            data = self._dClassifNoCat[queryName]
            queryNameWithClassif = queryName.split("_")
            queryNameWithClassif[0] = "noCat-%s-like" % refWickerCode
            queryNameWithClassif = "_".join(queryNameWithClassif)
            data[0] = queryNameWithClassif
            dOldToNewHeaders[queryName] = queryNameWithClassif
            
            refData = self._dClassifOk[refName]
            refCI = (refData[7].split(";")[0]).replace("=",":")
            line = "\t".join(data)
            line = line.replace("\n","")
            line += "; like=(bestHitName:%s %s)\n" % (refName,refCI)
            lModifiedNoCat.append( line )
        
        
        unModifiedNoCatNames = list(set(self._dClassifNoCat.keys()) - set(self._dPathQueryNameToPath.keys()))
        
        for queryName in unModifiedNoCatNames:
            data = self._dClassifNoCat[queryName]
            line  = "\t".join(data)
            lUnModifiedNoCat.append(line)
    
        self.dOldToNewHeaders = dOldToNewHeaders
    
        #merge all lists
        self._lOutputClassifs.extend(self._lClassifOk)
        self._lOutputClassifs.extend(self._lClassifChim)
        self._lOutputClassifs.extend(lModifiedNoCat)
        self._lOutputClassifs.extend(lUnModifiedNoCat)
        
        self._log.info("Finished classifying alignments")
      
    def run(self):
        LoggerFactory.setLevel(self._log, self._verbosity)
        self._checkOptions()
#        the new fasta and classif files have the same name
        outClassifFileName = "%s.classif" % os.path.splitext(os.path.basename(self._outFastaFileName))[0]        
        self._log.info("START %s" % self.__class__.__name__)
        self._log.debug("input classification file: %s" % self._inClassifFileName)
        self._log.debug("input fasta file: %s" % self._inFastaFileName)
        self._log.debug("output fasta file: %s" % self._outFastaFileName)
        self._log.debug("output file: %s" % outClassifFileName)
        
        if os.path.exists(outClassifFileName):
            self._logAndRaise("Output classif file name already exists")
        
        self._analyzeInputClassifFile()
        self._getSequencesByHeaders()
        self._alignFiles()
        self._classifyAlignements()
        
        #write output classif file name with new seq names
        with open(outClassifFileName, "w") as outClassifFile:
            for line in self._lOutputClassifs:
                outClassifFile.write(line)
    
        #write output fasta file name, with new headers
        with open(self._outFastaFileName,"w") as outFastaFile:
            iFastaParser = FastaParser(self._inFastaFileName)
            for iSequence in iFastaParser.getIterator():
                oldSeqName = iSequence.getName()
                if self.dOldToNewHeaders.has_key(oldSeqName):
                    newSeqName = self.dOldToNewHeaders[oldSeqName]
                    iSequence.setName(newSeqName)
                outFastaFile.write(iSequence.printFasta())
            
        FileUtils.sortFileContent(outClassifFileName, "tmp")
        shutil.move("tmp", outClassifFileName)
        
        self._log.info("END %s" % self.__class__.__name__)
       
       

if __name__ == "__main__":
    iNoCatBestHitClassifier = NoCatBestHitClassifier()
    iNoCatBestHitClassifier.setAttributesFromCmdLine()
    iNoCatBestHitClassifier.run()        