#!/usr/bin/env python

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import os
import sys
import time
import shutil
import ConfigParser
import glob
import subprocess
from commons.tools.PASTEClassifier import PASTEClassifier

if not "REPET_PATH" in os.environ.keys():
    print "ERROR: no environment variable REPET_PATH"
    sys.exit(1)
sys.path.append(os.environ["REPET_PATH"])
if not "PYTHONPATH" in os.environ.keys():
    os.environ["PYTHONPATH"] = os.environ["REPET_PATH"]
else:
    os.environ["PYTHONPATH"] = "%s:%s" % (os.environ["REPET_PATH"], os.environ["PYTHONPATH"])

from ConfigParser import NoOptionError
from ConfigParser import NoSectionError
from ConfigParser import MissingSectionHeaderError
from commons.core.LoggerFactory import LoggerFactory
from commons.core.launcher.Launcher import Launcher
from commons.core.launcher.LauncherUtils import LauncherUtils
from commons.core.utils.FileUtils import FileUtils
from commons.core.utils.RepetOptionParser import RepetOptionParser
from commons.core.checker.CheckerUtils import CheckerUtils
from commons.core.checker.ConfigChecker import ConfigRules
from commons.core.checker.ConfigChecker import ConfigChecker
from commons.core.checker.ConfigException import ConfigException
from commons.core.checker.CheckerException import CheckerException
from commons.core.seq.BioseqDB import BioseqDB
from commons.core.seq.FastaUtils import FastaUtils
from commons.core.sql.DbFactory import DbFactory
from commons.core.sql.TableJobAdaptatorFactory import TableJobAdaptatorFactory
from commons.tools.PrepareBatches import PrepareBatches
from commons.tools.GetClassifUniq import GetClassifUniq
from commons.tools.RenameHeaderClassif import RenameHeaderClassif
from commons.tools.RemoveRedundancyBasedOnCI import RemoveRedundancyBasedOnCI
from commons.tools.DetectTEFeatures_parallelized import DetectTEFeatures_parallelized
from commons.tools.ReverseComplementAccordingToClassif import ReverseComplementAccordingToClassif
from commons.tools.NoCatBestHitClassifier import NoCatBestHitClassifier
from commons.tools.LaunchGrouperInParallel import LaunchGrouperInParallel
from PASTEC.StatPastec import StatPastec

LOG_DEPTH = "repet.TEdenovo"

## Setup the required environment.
#
# @param config ConfigParser instance
#
def setup_env(config):
    os.environ["REPET_HOST"] = config.get("repet_env", "repet_host")
    os.environ["REPET_USER"] = config.get("repet_env", "repet_user")
    os.environ["REPET_PW"] = config.get("repet_env", "repet_pw")
    os.environ["REPET_DB"] = config.get("repet_env", "repet_db")
    os.environ["REPET_PORT"] = config.get("repet_env", "repet_port")
    os.environ["REPET_JOB_MANAGER"] = config.get("repet_env", "repet_job_manager")
    os.environ["REPET_QUEUE"] = config.get("repet_env", "repet_job_manager")
    os.environ["REPET_JOBS"] = "MySQL"

#------------------------------------------------------------------------------

def main():

    """
    This program is a pipeline for the de novo identification of transposable elements (TEs) in genomic sequences.
    """

    description = "Launch TEdenovo pipeline"
    examples = "\nExample 1: launch TEdenovo step 1 on DmelChr4.fa\n"
    examples += "\t$ TEdenovo.py -P DmelChr4 -C TEdenovo.cfg -S 1"
    examples += "\n\t"
    examples += "\nExample 2: launch TEdenovo step 2 structural\n"
    examples += "\t$ TEdenovo.py -P DmelChr4 -C TEdenovo.cfg -S 2 --struct"
    examples += "\n\t"
    examples += "\nExample 3: launch TEdenovo step 4 after Grouper\n"
    examples += "\t$ TEdenovo.py -P DmelChr4 -C TEdenovo.cfg -S 4 -s Blaster -c Grouper -m Map"
    examples += "\n\t"
    examples += "\nExample 4: launch TEdenovo step 6 including LTRharvest's consensus\n"
    examples += "\t$ TEdenovo.py -P DmelChr4 -C TEdenovo.cfg -S 6 -s Blaster -c GrpRecPil -m Map --struct\n"
    parser = RepetOptionParser(description = description, version = "v2.1", epilog = examples)
    parser.add_option("-P", "--project", dest = "projectName", action = "store", type = "string", default = "", help = "project name (<=15 symbols, alphanumeric or underscore)")
    parser.add_option("-C", "--config", dest = "configFileName", action = "store", type = "string", default = "", help = "configuration file")
    parser.add_option("-S", "--step", dest = "step", action = "store", type = "string", default = "", help = "step (1/2/3/4/5/6/7/8)")
    parser.add_option("", "--struct", dest = "isStructuralSearch", action = "store_true", default = False, help = "search for LTR using LTRharvest (default=False)")
    parser.add_option("-s", "--align", dest = "smplAlign", action = "store", type = "string", default = "", help = "program to self-align (Blaster)")
    parser.add_option("-c", "--clust", dest = "clustHsp", action = "store", type = "string", default = "", help = "program to cluster HSPs (Grouper/Recon/Piler or GrpRec/GrpPil/RecPil/GrpRecPil)")
    parser.add_option("-m", "--msa", dest = "multAlign", action = "store", type = "string", default = "", help = "program to build MSAs (Map)")
    parser.add_option("-f", "--family", dest = "clustMethod", action = "store", type = "string", default = "", help = "program to cluster consensus to find consensus families (Blastclust/MCL)")
    parser.add_option("-v", "--verbose", dest = "verbose", action = "store", type = "int", default = 1, help = "verbosity (0/default=1)")
    options = parser.parse_args()[0]

    global projectName
    projectName = options.projectName
    global configFileName
    configFileName = options.configFileName
    step = options.step
    smplAlign = options.smplAlign
    clustHsp = options.clustHsp
    multAlign = options.multAlign
    global verbose
    verbose = options.verbose
    isStructuralSearch = options.isStructuralSearch
    step8ClustMethod = options.clustMethod
    global configFileHandle

    if projectName == "" or configFileName == "" or step == "":
        print "ERROR: missing compulsory options"
        sys.exit(1)

    if not CheckerUtils.isMax15Char(projectName):
        print "ERROR: project name must have 15 character max"
        sys.exit(1)

    if not CheckerUtils.isCharAlphanumOrUnderscore(projectName):
        print "ERROR: project name must contain only alphanumeric or underscore character"
        sys.exit(1)

    global config
    configFileHandle = open(configFileName)
    config = ConfigParser.ConfigParser()

    configFileName = os.path.basename(configFileName)

    try :
        config.readfp(configFileHandle)
    except MissingSectionHeaderError:
        print "ERROR: config file " + configFileName + " must begin with a section header "
        sys.exit(1)

    isOptionErrorRaised = False
    sectionName = "repet_env"
    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionName)
    except NoSectionError:
        print "ERROR: the section " + sectionName + " must be in your config file : " + configFileName
        sys.exit(1)

    optionName = "repet_version"
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, optionName)
    except NoOptionError:
        print "ERROR: the option " + optionName + " must be defined in " + sectionName + " in your config file : " + configFileName
        isOptionErrorRaised = True
    optionName = "repet_host"
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, optionName)
    except NoOptionError:
        print "ERROR: the option " + optionName + " must be defined in " + sectionName + " in your config file : " + configFileName
        isOptionErrorRaised = True

    optionName = "repet_user"
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, optionName)
    except NoOptionError:
        print "ERROR: the option " + optionName + " must be defined in " + sectionName + " in your config file : " + configFileName
        isOptionErrorRaised = True

    optionName = "repet_pw"
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, optionName)
    except NoOptionError:
        print "ERROR: the option " + optionName + " must be defined in " + sectionName + " in your config file : " + configFileName
        isOptionErrorRaised = True

    optionName = "repet_db"
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, optionName)
    except NoOptionError:
        print "ERROR: the option " + optionName + " must be defined in " + sectionName + " in your config file : " + configFileName
        isOptionErrorRaised = True

    optionName = "repet_port"
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, optionName)
    except NoOptionError:
        print "ERROR: the option " + optionName + " must be defined in " + sectionName + " in your config file : " + configFileName
        isOptionErrorRaised = True

    if isOptionErrorRaised:
        sys.exit(1)

    setup_env(config)

    changeLogFileName = os.environ["REPET_PATH"] + "/doc/CHANGELOG"
    changeLogFileHandle = open(changeLogFileName, "r")

    try:
        CheckerUtils.checkConfigVersion(changeLogFileHandle, config)
    except CheckerException, e:
        print e.msg
        sys.exit(1)

    sectionName = "project"

    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionName)
    except NoSectionError:
        print "ERROR: the section " + sectionName + " must be in your config file : " + configFileName
        sys.exit(1)

    isOptionErrorRaised = False
    optionName = "project_name"
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, optionName)
    except NoOptionError:
        print "ERROR: the option " + optionName + " must be defined in " + sectionName + " in your config file : " + configFileName
        isOptionErrorRaised = True

    optionName = "project_dir"
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, optionName)
    except NoOptionError:
        print "ERROR: the option " + optionName + " must be defined in " + sectionName + " in your config file : " + configFileName
        isOptionErrorRaised = True

    if isStructuralSearch:
        sectionName = "structural_search_clustering"
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "type")
        except NoOptionError:
            print "ERROR: the option 'type' must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
            isOptionErrorRaised = True

    if isOptionErrorRaised :
        sys.exit(1)

    if verbose > 0:
        print "\nSTART TEdenovo.py (%s)" % time.strftime("%Y-%m-%d %H:%M:%S")
        print "version %s" % config.get("repet_env", "repet_version")
        sys.stdout.flush()

    if config.get("project", "project_name") != projectName:
        print "ERROR: project name different between configuration file and command-line option"
        sys.exit(1)
    if verbose > 0:
        print "project name = %s" % projectName
        sys.stdout.flush()

    global projectDir
    projectDir = config.get("project", "project_dir")
    if projectDir[-1] == "/":
        projectDir = projectDir[:-1]
    if os.path.exists(projectDir):
        if verbose > 0:
            print "project directory = %s" % projectDir
            sys.stdout.flush()
        os.chdir(projectDir)
    else:
        print "ERROR: project directory %s doesn't exist" % projectDir
        sys.exit(1)

    if isStructuralSearch and smplAlign == "":
        smplAlign = "LTRharvest"
        clustHsp = config.get("structural_search_clustering", "type").lower()
        if clustHsp == "blastclust":
            clustHsp = "Blastclust"
        elif clustHsp == "mcl":
            clustHsp = "MCL"

    if step == "1":
        iPrepareBatches = PrepareBatches("TEdenovo", projectDir, projectName, config, verbose)
        iPrepareBatches.run()

    if step == "2" and not isStructuralSearch:
        smplAlign = "Blaster"
        selfAlign(smplAlign)

    if step == "3" and not isStructuralSearch:
        smplAlign = "Blaster"
        clusterHsp(smplAlign, clustHsp)

    if step == "2" and isStructuralSearch:
        iTEStructuralSearch = TEStructuralSearch(projectDir, projectName, config, verbose)
        iTEStructuralSearch.run()

    if step == "3" and isStructuralSearch:
        iTEClustering = TEClustering(projectDir, projectName, config, verbose)
        iTEClustering.run()

    if step == "4":
        buildConsensus(smplAlign, clustHsp, multAlign, isStructuralSearch)

    if step == "5":
        detectFeatures(smplAlign, clustHsp, multAlign, isStructuralSearch)

    if step == "6":
        iClassifConsensus = ClassifConsensus(smplAlign, clustHsp, multAlign, isStructuralSearch, configFileName, verbose)
        iClassifConsensus.run()

    if step == "7":
        filterConsensus(smplAlign, clustHsp, multAlign, isStructuralSearch)

    if step == "8":
        if step8ClustMethod.lower() not in ["blastclust", "mcl"]:
            print "ERROR: missing compulsory options for consensus clustering."
            sys.exit(1)
        clusteringConsensus(smplAlign, clustHsp, multAlign, isStructuralSearch, step8ClustMethod)

    if verbose > 0:
        print "version %s" % config.get("repet_env", "repet_version")
        print "END TEdenovo.py (%s)\n" % time.strftime("%Y-%m-%d %H:%M:%S")
        sys.stdout.flush()

    return 0

#------------------------------------------------------------------------------

def getStructuralSearchClusterTool():
    sectionName = "structural_search_clustering"
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "type")
    except NoOptionError:
        print "ERROR: the option 'type' must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
        sys.exit(1)
    structClustering = config.get(sectionName, "type").lower()
    if structClustering == "mcl":
        structClustering = "MCL"
    elif structClustering == "blastclust":
        structClustering = "Blastclust"
    else:
        print "ERROR: the option 'type' must be set to 'MCL' or 'Blastclust' in '%s' in your config file : %s" % (sectionName, configFileName)
        sys.exit(1)

    return structClustering

#------------------------------------------------------------------------------

class TEStructuralSearch(object):

    def __init__(self, projectDir, projectName, iConfig, verbose):
        self._sectionName = "structural_search"
        self._projectDir = projectDir
        self._projectName = projectName
        self._iConfig = iConfig
        self._verbose = verbose
        self._minLTRSize = None
        self._maxLTRSize = None
        self._minElementSize = None
        self._maxElementSize = None
        self._LTRSimilarity = None
        self._overlaps = None

        self._minDistLTRHarvestParam = None
        self._maxDistLTRHarvestParam = None

    def setStructParams(self, minLTRSize, maxLTRSize, minElementSize, maxElementSize, LTRsimilarity, overlaps):
        self._minLTRSize = minLTRSize
        self._maxLTRSize = maxLTRSize
        self._minElementSize = minElementSize
        self._maxElementSize = maxElementSize
        self._LTRSimilarity = LTRsimilarity
        self._overlaps = overlaps

    def getLTRharvestStructParams(self):
        return self._minLTRSize, self._maxLTRSize, self._minDistLTRHarvestParam, self._maxDistLTRHarvestParam, self._LTRSimilarity, self._overlaps

    def run(self):
        if verbose > 0:
            print "beginning of step 2 (LTRharvest)"
            sys.stdout.flush()
        if FileUtils.isRessourceExists("%s_LTRharvest" % self._projectName):
            print "ERROR: directory '%s_LTRharvest' already exists" % self._projectName
            sys.exit(1)

        os.mkdir("%s_LTRharvest" % self._projectName)
        os.chdir("%s_LTRharvest" % self._projectName)

        self._checkConfig()
        self.setStructParams(int(self._iConfig.get(self._sectionName, "minLTRSize")), \
                             int(self._iConfig.get(self._sectionName, "maxLTRSize")), \
                             int(self._iConfig.get(self._sectionName, "minElementSize")), \
                             int(self._iConfig.get(self._sectionName, "maxElementSize")), \
                             int(self._iConfig.get(self._sectionName, "LTR_similarity")), \
                             self._iConfig.get(self._sectionName, "overlaps_handling"))
        self.checkConfigStructParamsConsistency()
        self.convertStructParamsForLTRHarvest()

        resources = self._iConfig.get(self._sectionName, "resources")
        cDir = os.getcwd()
        if self._iConfig.get(self._sectionName, "tmpDir") != "":
            tmpDir = self._iConfig.get(self._sectionName, "tmpDir")
        else:
            tmpDir = cDir

        groupid = "%s_TEdenovo_LTRharvest" % self._projectName
        acronym = "LTRharvest"
        iDb = DbFactory.createInstance()
        jobdb = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
        iLauncher = Launcher(jobdb, cDir, tmpDir, resources, groupid, acronym)

        lCmdsTuples = []
        lFiles = FileUtils.getFileNamesList("../%s_db/batches" % self._projectName, "batch*")
        for file in lFiles:
            lCmds = []
            lCmds.append(self._createLTRharvestCommand(file, iLauncher))
            lCmdStart = []
            lCmdStart.append("shutil.copy('%s/%s_db/batches/%s', '.' )" % (self._projectDir, self._projectName, file))
            lCmdFinish = []
            lCmdFinish.append("shutil.move('%s_LTRharvest.fa', '%s')" % (os.path.splitext(file)[0], cDir))
            lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart, lCmdFinish))

        cleanMustBeDone = False
        if config.get(self._sectionName, "clean") == "yes":
            cleanMustBeDone = True
        iLauncher.runLauncherForMultipleJobs(acronym, lCmdsTuples, cleanMustBeDone)

        # Concatenate all LTRharvest prediction in a single file
        FileUtils.catFilesByPattern("*.fa", "%s_LTRharvest.fa" % self._projectName)
        if cleanMustBeDone:
            FileUtils.removeFilesByPattern("batch*")

        os.chdir("..")
        if verbose > 0:
            print "step 2 (LTRharvest) finished successfully"
            sys.stdout.flush()

    def convertStructParamsForLTRHarvest(self):
        self._minDistLTRHarvestParam = self._minElementSize - self._minLTRSize
        self._maxDistLTRHarvestParam = self._maxElementSize - self._maxLTRSize

    def checkConfigStructParamsConsistency(self):
        if self._minLTRSize > self._maxLTRSize:
            raise ConfigException("ERROR: option 'minLTRSize' > 'maxLTRSize'. Check your config values in section 'structural_search'")
        if self._minElementSize > self._maxElementSize:
            raise ConfigException("ERROR: option 'minElementSize' > 'maxElementSize'. Check your config values in section 'structural_search'")
        if 2 * self._minLTRSize > self._minElementSize:
            raise ConfigException("ERROR: inconsistency between 'minLTRSize' and 'minElementSize' options. Check your config values in section 'structural_search'")
        if 2 * self._maxLTRSize > self._maxElementSize:
            raise ConfigException("ERROR: inconsistency between 'maxLTRSize' and 'maxElementSize' options. Check your config values in section 'structural_search'")
        if self._LTRSimilarity < 0 or self._LTRSimilarity > 100:
            raise ConfigException("ERROR: 'LTR_similarity' option must be between 0 and 100. Check your config value in section 'structural_search'")
        if self._overlaps not in ["no", "best", "all"]:
            raise ConfigException("ERROR: bad 'overlaps_handling' option in section 'structural_search': '%s'. Possible values are 'no', 'best' and 'all'" % self._overlaps)

    def _createLTRharvestCommand(self, file, iLauncher):
        lArgs = []
        lArgs.append("--db %s" % file)
        lArgs.append("--minlen %s" % self._minLTRSize)
        lArgs.append("--maxlen %s" % self._maxLTRSize)
        lArgs.append("--mindist %s" % self._minDistLTRHarvestParam)
        lArgs.append("--maxdist %s" % self._maxDistLTRHarvestParam)
        lArgs.append("--sim %s" % self._LTRSimilarity)
        lArgs.append("--over %s" % self._overlaps)
        return iLauncher.getSystemCommand("LaunchLTRharvest.py", lArgs)

    def _checkConfig(self):
        try:
            CheckerUtils.checkSectionInConfigFile(self._iConfig, self._sectionName)
        except NoSectionError:
            print "ERROR: the section %s must be in your configuration file" % self._sectionName
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(self._iConfig, self._sectionName, "resources")
        except NoOptionError:
            print "ERROR: the option 'resources' must be defined in %s in your configuration file" % self._sectionName
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(self._iConfig, self._sectionName, "tmpDir")
        except NoOptionError:
            print "ERROR: the option 'tmpDir' must be defined in %s in your configuration file" % self._sectionName
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(self._iConfig, self._sectionName, "minLTRSize")
        except NoOptionError:
            print "ERROR: the option 'minLTRSize' must be defined in %s in your configuration file" % self._sectionName
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(self._iConfig, self._sectionName, "maxLTRSize")
        except NoOptionError:
            print "ERROR: the option 'maxLTRSize' must be defined in %s in your configuration file" % self._sectionName
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(self._iConfig, self._sectionName, "minElementSize")
        except NoOptionError:
            print "ERROR: the option 'minElementSize' must be defined in %s in your configuration file" % self._sectionName
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(self._iConfig, self._sectionName, "maxElementSize")
        except NoOptionError:
            print "ERROR: the option 'maxElementSize' must be defined in %s in your configuration file" % self._sectionName
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(self._iConfig, self._sectionName, "LTR_similarity")
        except NoOptionError:
            print "ERROR: the option 'LTR_similarity' must be defined in %s in your configuration file" % self._sectionName
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(self._iConfig, self._sectionName, "overlaps_handling")
        except NoOptionError:
            print "ERROR: the option 'overlaps_handling' must be defined in %s in your configuration file" % self._sectionName
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(self._iConfig, self._sectionName, "clean")
        except NoOptionError:
            print "ERROR: the option 'clean' must be defined in %s in your configuration file" % self._sectionName
            sys.exit(1)

#------------------------------------------------------------------------------

class TEClustering(object):

    def __init__(self, projectDir, projectName, iConfig, verbose):
        self._projectDir = projectDir
        self._projectName = projectName
        self._iConfig = iConfig
        self._verbosity = verbose
        self._clustHsp = ""
        self._log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, self.__class__.__name__), self._verbosity)

    def _logAndRaise(self, errorMsg):
        self._log.error(errorMsg)
        raise Exception(errorMsg)

    def run(self):
        # TODO: set level to verbosity attribute when everything is using logger
        LoggerFactory.setLevel(self._log, self._verbosity + 2)
        self._log.info("beginning of step 3")

        self._clustHsp = getStructuralSearchClusterTool()
        sectionName = "structural_search_clustering"
        self._checkConfig(sectionName)

        if FileUtils.isRessourceExists("%s_LTRharvest_%s" % (self._projectName, self._clustHsp)):
            self._logAndRaise("ERROR: directory '%s_LTRharvest_%s' already exists" % (self._projectName, self._clustHsp))

        os.mkdir("%s_LTRharvest_%s" % (self._projectName, self._clustHsp))
        os.chdir("%s_LTRharvest_%s" % (self._projectName, self._clustHsp))

        self._log.info("launching clustering program: %s" % self._clustHsp)

        resources = self._iConfig.get(sectionName, "resources")
        cDir = os.getcwd()
        if self._iConfig.get(sectionName, "tmpDir") != "":
            tmpDir = self._iConfig.get(sectionName, "tmpDir")
        else:
            tmpDir = cDir
        groupid = "%s_TEdenovo_LTRharvest_%s" % (self._projectName, self._clustHsp)

        iDb = DbFactory.createInstance()
        jobdb = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
        iLauncher = Launcher(jobdb, cDir, tmpDir, resources, groupid)

        fileName = "%s_LTRharvest.fa" % self._projectName
        clusteringFastaFileName = "%s_%s.fa" % (os.path.splitext(fileName)[0], self._clustHsp)
        mapFilename = "%s_LTRharvest_%s.map" % (projectName, self._clustHsp)

        if self._clustHsp == 'Blastclust':
            blastclustOptions = "-b F -o %s_LTRharvest -m" % self._projectName
            lCmds = [self.getBlastclustCommand(fileName, iLauncher, blastclustOptions)]
        elif self._clustHsp == 'MCL':
            inflation = self._iConfig.get(sectionName, "MCL_inflation")
            coverage = self._iConfig.get(sectionName, "MCL_coverage")
            MCLOptions = "-I %s -T %s -j -o %s_LTRharvest -m" % (inflation, coverage, self._projectName)
            lCmds = [self.getMCLCommand(fileName, iLauncher, MCLOptions)]

        lCmdStart = ["shutil.copy('%s/%s_LTRharvest/%s', '.' )" % (self._projectDir, self._projectName, fileName)]
        lCmdFinish = ["shutil.move('%s', '%s/%s')" % (clusteringFastaFileName, cDir, clusteringFastaFileName)]
        lCmdFinish.append("shutil.move('%s', '%s/%s')" % (mapFilename, cDir, mapFilename))
        lCmdsTuples = [iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart, lCmdFinish)]

        cleanMustBeDone = False
        if config.get(sectionName, "clean") == "yes":
            cleanMustBeDone = True
        iLauncher.runLauncherForMultipleJobs("%s" % self._clustHsp, lCmdsTuples, cleanMustBeDone)

        self._log.info("Done launching clustering program:  %s " % self._clustHsp)

        self._log.info("Filtering clusters")
        minSeq = 1
        maxSeq = 20

        prg = "FilterSeqClusters.py"
        cmdFilter = prg
        cmdFilter += " -i %s" % mapFilename
        cmdFilter += " -m %s" % minSeq
        cmdFilter += " -M %s" % maxSeq
        cmdFilter += " -L -1"
        cmdFilter += " -v %i" % (self._verbosity - 2)

        os.system(cmdFilter)

        input = "%s_%s_%s.map.filtered-%s-%s" % (projectName, "LTRharvest", self._clustHsp, minSeq, maxSeq)

        bankName = projectName
        bankName += "_chunks.fa"
        pathToFasta = "%s/%s_db/%s" % (projectDir, projectName, bankName)

        prg = "map2db "
        cmd = prg
        cmd += input
        cmd += " %s" % pathToFasta
        os.system(cmd)

        self._log.info("Done filtering clusters ")
        self._log.info("Orienting clusters ")

        filteredFastaFileName = "%s.filtered-%s-%s.flank_size0.fa" % (mapFilename, minSeq, maxSeq)

        prg = "splitSeqPerCluster.py "
        cmd = prg
        cmd += "-i %s " % filteredFastaFileName
        cmd += "-c %s " % self._clustHsp
        cmd += "-v 0 "
        os.system(cmd)

        lFiles = glob.glob("seqCluster*.fa")

        lFileSizeTuples = []
        for fileName in lFiles:
                dbSize = FastaUtils.dbSize(fileName)
                if dbSize > 1:
                    lFileSizeTuples.append((fileName, dbSize))

        lFilesToReorient = [file for file, size in lFileSizeTuples]
        lFilesNotToReorient = list(set(lFiles) - set(lFilesToReorient))

        lFilesToWrite = lFilesNotToReorient
        lFilesToWrite.extend(["%s.oriented" % file for file in lFilesToReorient])

        if not lFilesToReorient:
            print "\tNo cluster has more than 1 member - No sequence re-orientation needed"
        else:
            lFileSizeList = LauncherUtils.createHomogeneousSizeList(lFileSizeTuples, 20)
            lCmdsTuples = []
            groupid = "%s_TEdenovo_S3_orienter" % projectName
            acronym = "orientClusterSequences"
            iDb = DbFactory.createInstance()
            jobdb = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
            iLauncher = Launcher(jobdb, cDir, tmpDir, resources, groupid)
            for lFileNames in lFileSizeList:
                lCmds = []
                lCmdStart = []
                lCmdFinish = []
                for file in lFileNames:
                    lCmds.append(self.getOrienterCommand(file, iLauncher))
                    lCmdStart.append("os.symlink(\"%s/%s\", \"%s\")" % (cDir, file, file))
                    lCmdFinish.append("shutil.move('%s.oriented', '%s/%s.oriented')" % (file, cDir, file))
                lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart, lCmdFinish))
            iLauncher.runLauncherForMultipleJobs(acronym, lCmdsTuples, cleanMustBeDone)

        cleanOrientedFastaFileName = "%s.oriented" % filteredFastaFileName
        FileUtils.catFilesFromList(lFilesToWrite, cleanOrientedFastaFileName)

        FileUtils.removeFilesByPattern("seqCluster*.fa*")
        os.symlink(cleanOrientedFastaFileName, "%s_LTRharvest_%s_%selem_%sseq.fa" % (self._projectName, self._clustHsp, minSeq, maxSeq))
        self._log.info("Done orienting clusters ")

        os.chdir("..")
        self._log.info("step 3 with %s finished successfully" % self._clustHsp)

    def getBlastclustCommand(self, file, iLauncher, blastclustOptions):
        lArgs = []
        lArgs.append("-i %s" % file)
        lArgs.append(blastclustOptions)
        return iLauncher.getSystemCommand("LaunchBlastclust.py", lArgs)

    def getMCLCommand(self, file, iLauncher, MCLOptions):
        lArgs = []
        lArgs.append("-i %s" % file)
        lArgs.append(MCLOptions)
        return iLauncher.getSystemCommand("LaunchMCL.py", lArgs)

    def getOrienterCommand(self, file, iLauncher):
        lArgs = []
        lArgs.append(file)
        lArgs.append("-v 0")
        return iLauncher.getSystemCommand("orienter", lArgs)

    def _checkConfig(self, sectionName):
        try:
            CheckerUtils.checkSectionInConfigFile(self._iConfig, sectionName)
        except NoSectionError:
            print "ERROR: the section %s must be in your configuration file" % sectionName
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(self._iConfig, sectionName, "resources")
        except NoOptionError:
            print "ERROR: the option 'resources' must be defined in %s in your configuration file" % sectionName
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(self._iConfig, sectionName, "tmpDir")
        except NoOptionError:
            print "ERROR: the option 'tmpDir' must be defined in %s in your configuration file" % sectionName
            sys.exit(1)

#------------------------------------------------------------------------------

def getFilterAlignCmd(verbose, sectionName, config, iLauncher):
    lArgs = []
    lArgs.append("-i %s.align.not_over" % projectName)
    lArgs.append("-E %s" % config.get(sectionName, "min_Evalue"))
    lArgs.append("-I %s" % config.get(sectionName, "min_identity"))
    lArgs.append("-l %s" % config.get(sectionName, "min_length"))
    lArgs.append("-L %s" % config.get(sectionName, "max_length"))
    lArgs.append("-v %d" % (verbose - 1))

    return iLauncher.getSystemCommand("FilterAlign.py", lArgs)

#------------------------------------------------------------------------------

def selfAlign(smplAlign):
    """
    step 2: self-alignment of the input data bank

    @param smplAlign: name of the alignment program (Blaster or Pals)
    @type smplAlign: string
    """

    if verbose > 0:
        print "beginning of step 2"
        print "self-alignment with %s" % smplAlign
        sys.stdout.flush()

    if os.path.exists(projectName + "_" + smplAlign):
        print "ERROR: directory " + projectName + "_" + smplAlign + "/ already exists"
        sys.exit(1)

    sectionName = "self_align"
    isCopyOnNode = False

    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionName)
    except NoSectionError:
        print "ERROR: the section " + sectionName + " must be in your config file : " + configFileName
        sys.exit(1)

    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "blast")
    except NoOptionError:
        print "ERROR: the option blast must be defined in " + sectionName + " in your config file : " + configFileName
        sys.exit(1)

    blastToUse = config.get(sectionName, "blast")

    if blastToUse not in ["ncbi", "blastplus", "wu"]:
        print "ERROR: 'blast: %s' in your configuration file is not authorized" % blastToUse
        sys.exit(1)

    if blastToUse == "blastplus":
        if not (CheckerUtils.isExecutableInUserPath("blastn") or CheckerUtils.isExecutableInUserPath("makeblastdb")):
            print "ERROR: ncbi blast plus (blastn executable) must be in your path"
            sys.exit(1)
        else:
            process = subprocess.Popen("blastn -h", shell = True, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
            output = process.communicate()[0]
            if "washu" in output.lower():
                print "ERROR: you are trying to use NCBI BlastPlus, but the blastn executable is pointing to WuBlast. Please check your system path."
                sys.exit(1)
    elif blastToUse == "wu":
        if not (CheckerUtils.isExecutableInUserPath("blastn") or CheckerUtils.isExecutableInUserPath("xdformat")):
            print "ERROR: WuBlast (blastn executable) must be in your path"
            sys.exit(1)
        else:
            process = subprocess.Popen("blastn -h", shell = True, stdout = subprocess.PIPE, stderr = subprocess.PIPE)
            output = process.communicate()[0]
            if not "washu" in output.lower():
                print "ERROR: you are trying to use WuBlast, but the blastn executable is pointing to NCBI BlastPlus. Please check your system path."
                sys.exit(1)
    else:
        if not (CheckerUtils.isExecutableInUserPath("blastall") or CheckerUtils.isExecutableInUserPath("formatdb")):
            print "ERROR: ncbi blastall and formatdb must be in your path"
            sys.exit(1)

    if smplAlign == "Blaster":
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "Evalue")
        except NoOptionError:
            print "ERROR: the option Evalue must be defined in " + sectionName + " in your config file : " + configFileName
            sys.exit(1)

    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "clean")
    except NoOptionError:
        print "ERROR: the option 'clean' must be defined in " + sectionName + " in your config file : " + configFileName
        sys.exit(1)
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "length")
    except NoOptionError:
        print "ERROR: the option 'length' must be defined in " + sectionName + " in your config file : " + configFileName
        sys.exit(1)
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "identity")
    except NoOptionError:
        print "ERROR: the option 'identity' must be defined in " + sectionName + " in your config file : " + configFileName
        sys.exit(1)

    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "filter_HSP")
    except NoOptionError:
        print "ERROR: the option 'filter_HSP' must be defined in " + sectionName + " in your config file : " + configFileName
        sys.exit(1)
    if config.get(sectionName, "filter_HSP") == "yes":
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "min_Evalue")
        except NoOptionError:
            print "ERROR: the option 'min_Evalue' must be defined in " + sectionName + " in your config file : " + configFileName
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "min_identity")
        except NoOptionError:
            print "ERROR: the option 'min_identity' must be defined in " + sectionName + " in your config file : " + configFileName
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "min_length")
        except NoOptionError:
            print "ERROR: the option 'min_length' must be defined in " + sectionName + " in your config file : " + configFileName
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "max_length")
        except NoOptionError:
            print "ERROR: the option 'max_length' must be defined in " + sectionName + " in your config file : " + configFileName
            sys.exit(1)

    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "resources")
    except NoOptionError:
        print "ERROR: the option 'resources' must be defined in " + sectionName + " in your config file : " + configFileName
        sys.exit(1)
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "tmpDir")
    except NoOptionError:
        print "ERROR: the option 'tmpDir' must be defined in " + sectionName + " in your config file : " + configFileName
        sys.exit(1)

    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "copy")
    except NoOptionError:
        print "ERROR: the option 'copy' must be defined in " + sectionName + " in your config file : " + configFileName
        sys.exit(1)
    if config.get(sectionName, "copy") == "yes" and config.get(sectionName, "tmpDir") != "":
        isCopyOnNode = True

    if verbose > 0:
        print "The copy option is: %s " % isCopyOnNode
        sys.stdout.flush()

    os.mkdir("%s_%s" % (projectName, smplAlign))
    os.chdir("%s_%s" % (projectName, smplAlign))
    os.system("cp ../%s ." % configFileName)

    queue = config.get(sectionName, "resources")
    cDir = os.getcwd()
    if config.get(sectionName, "tmpDir") != "":
        tmpDir = config.get(sectionName, "tmpDir")
    else:
        tmpDir = cDir

    # if the program is Blaster
    if smplAlign == "Blaster":

        # TODO: remove temporary config file when TEdenovo.cfg and TEannot.cfg have been merged
        tempConfigfileName = "%s/blaster.cfg" % os.getcwd()
        with open(tempConfigfileName, "w") as fh:
            fh.write("[jobs]\n")
            fh.write("resources: %s\n" % config.get("self_align", "resources"))
            fh.write("tmpDir: %s\n" % config.get("self_align", "tmpDir"))
            fh.write("copy: %s\n" % config.get("self_align", "copy"))
            fh.write("clean: %s\n" % config.get("self_align", "clean"))
            fh.write("\n")
            fh.write("[prepare_data]\n")
            fh.write("chunk_length: %s\n" % config.get("prepare_batches", "chunk_length"))
            fh.write("chunk_overlap: %s\n" % config.get("prepare_batches", "chunk_overlap"))
            fh.write("\n")
            fh.write("[alignment]\n")
            fh.write("blast: %s\n" % config.get("self_align", "blast"))
            fh.write("Evalue: %s\n" % config.get("self_align", "Evalue"))
            fh.write("length: %s\n" % config.get("self_align", "length"))
            fh.write("identity: %s\n" % config.get("self_align", "identity"))

        batchesDir = "%s/%s_db/batches" % (projectDir, projectName)
        subjectFilePath = "%s/%s_db/%s_chunks.fa" % (projectDir, projectName, projectName)
        outFileName = "%s.align.not_over" % projectName
        groupid = "%s_TEdenovo_S2_Blaster" % projectName
        cmd = "LaunchBlasterInParallel.py -q %s -s %s -a -o %s -C %s -g %s -v %d" % (batchesDir, subjectFilePath, outFileName, tempConfigfileName, groupid, verbose - 1)
        process = subprocess.Popen(cmd, shell = True)
        process.communicate()
        if process.returncode != 0:
            raise Exception("ERROR when launching '%s'" % cmd)
        os.remove(tempConfigfileName)

    # filter the HSPs
    if config.get(sectionName, "filter_HSP") == "yes":
        groupid = "%s_TEdenovo_%s_FilterAlign" % (projectName, smplAlign)
        acronym = "FilterAlign"
        iDb = DbFactory.createInstance()
        jobdb = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
        iLauncher = Launcher(jobdb, cDir, tmpDir, queue, groupid)

        lCmdsTuples = []
        lCmds = []
        lCmds.append(getFilterAlignCmd(verbose, sectionName, config, iLauncher))
        lCmdStart = []
        lCmdStart.append("os.symlink(\"%s/%s.align.not_over\", \"%s.align.not_over\")" % (cDir, projectName, projectName))
        lCmdFinish = []
        lCmdFinish.append("shutil.move(\"%s.align.not_over.filtered\", \"%s/.\")" % (projectName, cDir))
        lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart, lCmdFinish))

        cleanMustBeDone = False
        if config.get(sectionName, "clean") == "yes":
            cleanMustBeDone = True
        iLauncher.runLauncherForMultipleJobs(acronym, lCmdsTuples, cleanMustBeDone)

    if config.get(sectionName, "clean") == "yes":

        # remove the temporary file(s) with the HSPs
        if config.get(sectionName, "filter_HSP") == "yes":
            cmd = "rm -f %s.align %s.align.not_over" % (projectName, projectName)
            log = os.system(cmd)
            if log != 0:
                print "ERROR while removing an intermediary file"
                sys.exit(1)

        os.chdir("%s/%s_db" % (projectDir, projectName))
        # remove the files "_cut"
        cmd = "find . -name \"%s*_cut*\" -exec rm {} \;" % projectName
        log = os.system(cmd)
        if log != 0:
            print "ERROR while removing the '_cut' files"
            sys.exit(1)

    os.chdir("..")

    if verbose > 0:
        print "step 2 finished successfully"
        sys.stdout.flush()

#------------------------------------------------------------------------------

def getLaunchReconOrPilerCmd(smplAlign, clustHsp, verbose, sectionNameClusterHSP, config, projectName, alignFileName, bankName, iLauncher):
    lArgs = []
    lArgs.append("-p %s_%s" % (projectName, smplAlign))
    lArgs.append("-f %s" % bankName)
    lArgs.append("-a %s" % alignFileName)
    if config.get(sectionNameClusterHSP, "clean") == "yes":
        lArgs.append("-c")
    lArgs.append("-v %i" % (verbose - 1))
    return iLauncher.getSystemCommand("launch%s.py" % clustHsp, lArgs)

def getFilterSeqClusterCmd(dirName, minSeq, maxSeq, sectionNameClusterHSP, verbose, logFileName, iLauncher):
    lArgs = []
    lArgs.append("-i %s.map" % dirName)
    lArgs.append("-m %s" % minSeq)
    lArgs.append("-M %s" % maxSeq)
    lArgs.append("-L %s" % config.get(sectionNameClusterHSP, "maxSeqLength"))
    lArgs.append("-d")
    lArgs.append("-v 3 > %s" % logFileName)
    return iLauncher.getSystemCommand("FilterSeqClusters.py", lArgs)

def getMap2dbCmd(dirName, bankName, minSeq, maxSeq, iLauncher):
    lArgs = []
    lArgs.append("%s.map.filtered-%s-%s" % (dirName, minSeq, maxSeq))
    lArgs.append(bankName)
    return iLauncher.getSystemCommand("map2db", lArgs)

#------------------------------------------------------------------------------

def clusterHsp(smplAlign, clustHsp):
    """
    step 3: cluster the resulting HSPs (high-scoring segment pairs)

    @param smplAlign: name of the alignment program (Blaster or Pals)
    @type smplAlign: string

    @param clustHsp: name of the clustering program (Grouper, Recon or Piler)
    @type clustHsp: string
    """

    if verbose > 0:
        print "beginning of step 3"
        print "self-alignment with %s" % smplAlign
        print "clustering with %s" % clustHsp
        sys.stdout.flush()

    selfAlignSectionName = "self_align"
    sectionName = "cluster_HSPs"

    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionName)
    except NoSectionError:
        print "ERROR: '%s' must be in your config file : %s" % (sectionName, configFileName)
        sys.exit(1)

    if clustHsp == "Grouper" or clustHsp == "Recon" or clustHsp == "Piler":
        dirName = "%s_%s_%s" % (projectName, smplAlign, clustHsp)
        if os.path.exists(dirName):
            print "ERROR: directory %s already exists" % dirName
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "minNbSeqPerGroup")
        except NoOptionError:
            print "ERROR: the option minNbSeqPerGroup must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "nbLongestSeqPerGroup")
        except NoOptionError:
            print "ERROR: the option nbLongestSeqPerGroup must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, selfAlignSectionName, "filter_HSP")
        except NoOptionError:
            print "ERROR: the option filter_HSP must be defined in '%s' in your config file : %s" % (selfAlignSectionName, configFileName)
            sys.exit(1)

        if int(config.get(sectionName, "minNbSeqPerGroup")) < 3:
            print "ERROR: groups with less than %i members should be filtered out" % int(config.get(sectionName, "minNbSeqPerGroup"))
            print "as they are likely to correspond to segmental duplications"
            print "change this parameter value in your configuration file"
            sys.exit(1)

        os.mkdir(dirName)
        os.chdir(dirName)

        alignFileName = "%s.align.not_over" % projectName
        if config.get(selfAlignSectionName, "filter_HSP") == "yes":
            alignFileName += ".filtered"
        pathToAlignFile = "%s/%s_%s/%s" % (projectDir, projectName, smplAlign, alignFileName)

#        sectionNamePrepareBatches = "prepare_batches"
        bankName = "%s_chunks.fa" % projectName
        pathToBank = "%s/%s_db/%s" % (projectDir, projectName, bankName)
        logFileName = "%s_%sFiltering.log" % (projectName, clustHsp)

        #----------------------------------------------------------------------

        minSeq = config.get(sectionName, "minNbSeqPerGroup")
        maxSeq = config.get(sectionName, "nbLongestSeqPerGroup")

        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "clean")
        except NoOptionError:
            print "ERROR: the option 'clean' must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "maxSeqLength")
        except NoOptionError:
            print "ERROR: the option 'maxSeqLength' must be defined in '%s' in your config file :  %s" % (sectionName, configFileName)
            sys.exit(1)

        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "resources")
        except NoOptionError:
            print "ERROR: the option 'resources' must be defined in '%s' in your config file :  %s" % (sectionName, configFileName)
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "tmpDir")
        except NoOptionError:
            print "ERROR: the option 'tmpDir' must be defined in '%s' in your config file :  %s" % (sectionName, configFileName)
            sys.exit(1)
        os.symlink("../%s" % configFileName, configFileName)
        queue = config.get(sectionName, "resources")
        cDir = os.getcwd()
        if config.get(sectionName, "tmpDir") != "":
            tmpDir = config.get(sectionName, "tmpDir")
        else:
            tmpDir = cDir

        groupid = "%s_TEdenovo_%s_%s" % (projectName, smplAlign, clustHsp)
        acronym = clustHsp

        if clustHsp == "Grouper":
            try:
                CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "Grouper_nbGroup")
            except NoOptionError:
                print "ERROR: the option Grouper_nbGroup must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
                sys.exit(1)
            try:
                CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "Grouper_coverage")
            except NoOptionError:
                print "ERROR: the option Grouper_coverage must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
                sys.exit(1)
            try:
                CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "Grouper_include")
            except NoOptionError:
                print "ERROR: the option Grouper_include must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
                sys.exit(1)
            try:
                CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "Grouper_maxJoinLength")
            except NoOptionError:
                print "ERROR: the option Grouper_maxJoinLength must be defined in '%s' in your config file : %s" % (sectionName, configFileName)
                sys.exit(1)

            os.symlink(pathToBank, bankName)
            os.symlink(pathToAlignFile, alignFileName)
            iLGIP = LaunchGrouperInParallel(configFileName, bankName, alignFileName, groupId = groupid, verbosity = verbose - 1)
            iLGIP.run()
            os.remove(bankName)
            os.remove(alignFileName)

        else:
            iDb = DbFactory.createInstance()
            jobdb = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
            iLauncher = Launcher(jobdb, cDir, tmpDir, queue, groupid)
            lCmdsTuples = []
            # remove the chunk overlaps
            lCmds = []
            lCmdStart = []
            lCmdStart.append("os.symlink(\"%s\", \"%s\")" % (pathToBank, bankName))
            lCmdStart.append("os.symlink(\"%s\", \"%s\")" % (pathToAlignFile, alignFileName))
            lCmdFinish = []

            if clustHsp == "Recon":
                if not CheckerUtils.isExecutableInUserPath("recon.pl") :
                    print "ERROR: recon.pl must be in your path"
                    sys.exit(1)
                lCmdFinish.append("shutil.move(\"%s_%s_Recon.map\", \"%s/.\")" % (projectName, smplAlign, cDir))
                lCmdFinish.append("shutil.move(\"%s_%s_Recon.fa\", \"%s/.\")" % (projectName, smplAlign, cDir))
                lCmdFinish.append("shutil.move(\"%s_%s_MSP_file\", \"%s/.\")" % (projectName, smplAlign, cDir))
                lCmdFinish.append("shutil.move(\"%s_%s_seq_list\", \"%s/.\")" % (projectName, smplAlign, cDir))
                if config.get(sectionName, "clean") == "no":
                    lCmdFinish.append("shutil.move(\"images\", \"%s/images\")" % cDir)
                    lCmdFinish.append("shutil.move(\"ele_def_res\", \"%s/ele_def_res\")" % cDir)
                    lCmdFinish.append("shutil.move(\"ele_redef_res\", \"%s/ele_redef_res\")" % cDir)
                    lCmdFinish.append("shutil.move(\"edge_redef_res\", \"%s/edge_redef_res\")" % cDir)
                    lCmdFinish.append("shutil.move(\"summary\", \"%s/summary\")" % cDir)

            elif clustHsp == "Piler":
                if not CheckerUtils.isExecutableInUserPath("piler") :
                    print "ERROR: piler must be in your path"
                    sys.exit(1)
                lCmdFinish.append("shutil.move(\"%s_%s_Piler.map\", \"%s/.\")" % (projectName, smplAlign, cDir))
                lCmdFinish.append("shutil.move(\"%s_%s_Piler.fa\", \"%s/.\")" % (projectName, smplAlign, cDir))
                lCmdFinish.append("shutil.move(\"%s.gff\", \"%s/.\")" % (alignFileName, cDir))
                lCmdFinish.append("shutil.move(\"%s_%s_Piler-trs.gff\", \"%s/.\")" % (projectName, smplAlign, cDir))

            lCmdFinish.append("shutil.move(\"%s.map.filtered-%s-%s\", \"%s/.\")" % (dirName, minSeq, maxSeq, cDir))
            lCmdFinish.append("shutil.move(\"%s\", \"%s\")" % (logFileName, cDir))
            lCmdFinish.append("shutil.move(\"%s.map.distribC\", \"%s/.\")" % (dirName, cDir))
            lCmdFinish.append("shutil.move(\"%s.map.filtered-%s-%s.flank_size0.fa\", \"%s/.\")" % (dirName, minSeq, maxSeq, cDir))

            cmd = getLaunchReconOrPilerCmd(smplAlign, clustHsp, verbose, sectionName, config, projectName, alignFileName, bankName, iLauncher)
            lCmds.append(cmd)
            cmd = getFilterSeqClusterCmd(dirName, minSeq, maxSeq, sectionName, verbose, logFileName, iLauncher)
            lCmds.append(cmd)
            cmd = getMap2dbCmd(dirName, bankName, minSeq, maxSeq, iLauncher)
            lCmds.append(cmd)

            lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart, lCmdFinish))

            cleanMustBeDone = False
            if config.get(sectionName, "clean") == "yes":
                cleanMustBeDone = True
            iLauncher.runLauncherForMultipleJobs(acronym, lCmdsTuples, cleanMustBeDone)

            if clustHsp in ["Recon", "Piler"]:
                os.symlink("%s.map.filtered-%s-%s.flank_size0.fa" % (dirName, minSeq, maxSeq), "%s_%selem_%sseq.fa" % (dirName, minSeq, maxSeq))

        os.chdir("..")

    #--------------------------------------------------------------------------

    elif clustHsp == "":
        print "ERROR: missing option -c"
        sys.exit(1)

    else:
        print "ERROR: unknown option -c %s" % clustHsp
        sys.exit(1)

    if verbose > 0:
        print "step 3 finished successfully"
        sys.stdout.flush()

#------------------------------------------------------------------------------

def getConvertFastaHeadersFromChkToChrCmd(file, mapChunkFile, iLauncher):
    lArgs = []
    lArgs.append("-i %s" % file)
    lArgs.append("-m %s" % mapChunkFile)
    lArgs.append("-v %i" % (verbose - 1))
    return iLauncher.getSystemCommand("ConvertFastaHeadersFromChkToChr.py", lArgs)

def getLaunchMapCmd(inFile, sectionNameBuildConsensus, iLauncher):
    lArgs = []
    lArgs.append("-i %s" % inFile)
    if config.get(sectionNameBuildConsensus, "clean") == "yes":
        lArgs.append("-c")
    lArgs.append("-v %i" % verbose)
    return iLauncher.getSystemCommand("LaunchMap.py", lArgs)

def getMafftProgramLauncherCmd(inFile, sectionNameBuildConsensus, iLauncher):
    lArgs = []
    lArgs.append("-i %s" % inFile)
    if config.get(sectionNameBuildConsensus, "clean") == "yes":
        lArgs.append("-c")
    lArgs.append("-v %i" % verbose)
    return iLauncher.getSystemCommand("MafftProgramLauncher.py", lArgs)

def getBuildConsLibCmd(sectionNameBuildConsensus, prefix, multAlign, iLauncher):
    lArgs = []
    lArgs.append("-d %s" % os.getcwd())
    lArgs.append("-r \'*.fa_aln\'")
    lArgs.append("-n %s" % config.get(sectionNameBuildConsensus, "minBasesPerSite"))
    lArgs.append("-o %s_%s_consensus.fa" % (prefix, multAlign))
    if config.get(sectionNameBuildConsensus, "clean") == "yes":
        lArgs.append("-c")
    lArgs.append("-v %i" % (verbose - 1))
    return iLauncher.getSystemCommand("buildConsLib.py", lArgs)

#------------------------------------------------------------------------------

def buildConsensus(smplAlign, clustHsp, multAlign, isStructuralSearch):

    """
    step 4: build consensus from multiple alignments

    @param smplAlign: name of the alignment program (Blaster or Pals)
    @type smplAlign: string

    @param clustHsp: name of the clustering program (Grouper, Recon or Piler)
    @type clustHsp: string

    @param multAlign: name of the multiple alignment program MSA (MAP, Mafft, Muscle, Tcoffee, ClustalW or Prank)
    @type multAlign: string
    """

    isOptionErrorRaised = False

    if isStructuralSearch:
        clustHsp = getStructuralSearchClusterTool()

    if verbose > 0:
        print "beginning of step 4"
        print "multiple alignment with %s" % multAlign
        sys.stdout.flush()

    sectionNameClusterHSPs = "cluster_HSPs"
    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionNameClusterHSPs)
    except NoSectionError:
        print "ERROR: the section '%s' must be defined in your config file : %s" % (sectionNameClusterHSPs, configFileName)
        sys.exit(1)
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionNameClusterHSPs, "minNbSeqPerGroup")
    except NoOptionError:
        print "ERROR: the option 'minNbSeqPerGroup' must be defined in '%s' in your config file : %s" % (sectionNameClusterHSPs, configFileName)
        isOptionErrorRaised = True
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionNameClusterHSPs, "nbLongestSeqPerGroup")
    except NoOptionError:
        print "ERROR: the option 'nbLongestSeqPerGroup' must be defined in '%s' in your config file : %s" % (sectionNameClusterHSPs, configFileName)
        isOptionErrorRaised = True

    sectionNameBuildConsensus = "build_consensus"
    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionNameBuildConsensus)
    except NoSectionError:
        print "ERROR: the section '%s' must be defined in your config file : %s" % (sectionNameBuildConsensus, configFileName)
        sys.exit(1)
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionNameBuildConsensus, "resources")
    except NoOptionError:
        print "ERROR: the option 'resources' must be defined in '%s' in your config file : %s" % (sectionNameBuildConsensus, configFileName)
        isOptionErrorRaised = True
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionNameBuildConsensus, "tmpDir")
    except NoOptionError:
        print "ERROR: the option 'tmpDir' must be defined in '%s' in your config file : %s" % (sectionNameBuildConsensus, configFileName)
        isOptionErrorRaised = True
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionNameBuildConsensus, "minBasesPerSite")
    except NoOptionError:
        print "ERROR: the option 'minBasesPerSite' must be defined in '%s' in your config file : %s" % (sectionNameBuildConsensus, configFileName)
        isOptionErrorRaised = True
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionNameBuildConsensus, "clean")
    except NoOptionError:
        print "ERROR: the option 'clean' must be defined in  in your config file : %s" % (sectionNameBuildConsensus, configFileName)
        isOptionErrorRaised = True

    if isOptionErrorRaised:
        sys.exit(1)

    prefix1 = "%s_%s_%s" % (projectName, smplAlign, clustHsp)
    prefix2 = prefix1
    if clustHsp in ["Blastclust", "MCL"]:
        prefix2 += "_1elem_20seq"
    else:
        prefix2 += "_%selem_%sseq" % (config.get(sectionNameClusterHSPs, "minNbSeqPerGroup"), config.get(sectionNameClusterHSPs, "nbLongestSeqPerGroup"))

    if clustHsp not in ["Grouper", "Recon", "Piler", "Blastclust", "MCL"]:
        print "ERROR: %s not implememented" % clustHsp
        sys.exit(1)
    else:
        if not os.path.exists("%s/%s.fa" % (prefix1, prefix2)):
            print "ERROR: %s/%s.fa not found" % (prefix1, prefix2)
            sys.exit(1)

        if os.path.exists("%s_%s" % (prefix1, multAlign)):
            print "ERROR: directory %s_%s/ already exists" % (prefix1, multAlign)
            sys.exit(1)

        if multAlign not in ["Map", "MAP", "Mafft"]:
            print "ERROR: %s not implememented" % (multAlign)
            sys.exit(1)

        os.mkdir("%s_%s" % (prefix1, multAlign))
        os.chdir("%s_%s" % (prefix1, multAlign))
        os.symlink("../%s/%s.fa" % (prefix1, prefix2), "%s.fa" % prefix2)

        cmd = "splitSeqPerCluster.py"
        cmd += " -i %s.fa" % prefix2
        cmd += " -c %s" % clustHsp
        cmd += " -v %i" % (verbose - 1)
        sys.stdout.flush()
        process = subprocess.Popen(cmd, shell = True)
        process.communicate()
        if process.returncode != 0:
            raise Exception("ERROR when launching '%s'" % cmd)

        os.system("ln -s ../%s ." % (configFileName))
        queue = config.get(sectionNameBuildConsensus, "resources")
        cDir = os.getcwd()
        if config.get(sectionNameBuildConsensus, "tmpDir") != "":
            tmpDir = config.get(sectionNameBuildConsensus, "tmpDir")
        else:
            tmpDir = cDir

        cleanMustBeDone = False
        if config.get(sectionNameBuildConsensus, "clean") == "yes":
            cleanMustBeDone = True

#         Convert FASTA headers
        mapChunkFile = "%s/%s_db/%s_chunks.map" % (projectDir, projectName, projectName)
        lFiles = glob.glob("seqCluster*.fa")
        if lFiles:
            lFileSizeTuples = []
            for fileName in lFiles:
                fileSize = os.path.getsize(fileName)
                lFileSizeTuples.append((fileName, fileSize))
            lFileSizeList = LauncherUtils.createHomogeneousSizeList(lFileSizeTuples, 1000000)
            lCmdsTuples = []
            groupid = "%s_TEdenovo_S4_%s_convertHeadersFromChkToChr" % (projectName, clustHsp)
            acronym = "convertFastaHeadersFromChkToChr"
            iDb = DbFactory.createInstance()
            jobdb = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
            iLauncher = Launcher(jobdb, cDir, tmpDir, queue, groupid)
            for lFileNames in lFileSizeList:
                lCmds = []
                lCmdStart = []
                lCmdFinish = []
                for file in lFileNames:
                    lCmds.append(getConvertFastaHeadersFromChkToChrCmd(file, mapChunkFile, iLauncher))
                    lCmdStart.append("os.symlink(\"%s/%s\", \"%s\")" % (cDir, file, file))
                    lCmdFinish.append("if os.path.exists(\"%s.onChr\"):" % file)
                    lCmdFinish.append("\tshutil.move(\"%s/%s\", \"%s/%s.onChk\")" % (cDir, file, cDir, file))
                    lCmdFinish.append("\tshutil.move(\"%s.onChr\", \"%s/%s\")" % (file, cDir, file))
                lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart, lCmdFinish))
            iLauncher.runLauncherForMultipleJobs(acronym, lCmdsTuples, cleanMustBeDone)

            if multAlign in [ "Map", "MAP" ] and not CheckerUtils.isExecutableInUserPath("rpt_map") :
                print "ERROR: Map must be in your path"
                sys.exit(1)
            if multAlign == "Mafft" and not CheckerUtils.isExecutableInUserPath("mafft") :
                print "ERROR: mafft must be in your path"
                sys.exit(1)

            lFileSizeTuples = []
            for fileName in lFiles:
                iBioseqDB = BioseqDB(fileName)
                if iBioseqDB.getSize() == 1:
                    iBioseqDB.upCase()
                    iBioseqDB.save("%s.fa_aln" % fileName)
                else:
                    fileSize = os.path.getsize(fileName)
                    lFileSizeTuples.append((fileName, fileSize))
            lFileSizeList = LauncherUtils.createHomogeneousSizeList(lFileSizeTuples, 100000)

            inFileDirectory = os.getcwd()
            groupid = "%s_TEdenovo_%s_%s_%s" % (projectName, smplAlign, clustHsp, multAlign)
            acronym = multAlign
            iDb = DbFactory.createInstance()
            iTJA = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
            iLauncher = Launcher(iTJA, cDir, tmpDir, queue, groupid)
            lCmdsTuples = []

            for lFiles in lFileSizeList:
                lCmds = []
                lCmdStart = []
                lCmdFinish = []
                for file in lFiles:
                    if multAlign in [ "Map", "MAP" ]:
                        lCmds.append(getLaunchMapCmd(file, sectionNameBuildConsensus, iLauncher))
                    elif multAlign == "Mafft":
#                         TODO: test Mafft in functional test
                        lCmds.append(getMafftProgramLauncherCmd(file, iLauncher))
                    lCmdStart.append("os.symlink(\"%s/%s\", \"%s\")" % (inFileDirectory, file, file))
                    lCmdFinish.append("shutil.copy(\"%s.fa_aln\", \"%s/.\")" % (file, inFileDirectory))
                lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart, lCmdFinish))
            if lCmdsTuples:
                iLauncher.runLauncherForMultipleJobs(acronym, lCmdsTuples, cleanMustBeDone)

#             build the consensus for each MSA and concatenate them into a single fasta file
            groupid = "%s_TEdenovo_S4_%s_buildConsLib" % (projectName, clustHsp)
            iDb = DbFactory.createInstance()
            jobdb = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
            iLauncher = Launcher(jobdb, cDir, tmpDir, queue, groupid)
            lCmds = []
            lCmds.append(getBuildConsLibCmd(sectionNameBuildConsensus, prefix1, multAlign, iLauncher))

            lFiles = glob.glob(cDir + "/*.fa_aln")
            lCmdStart = []
            for file in lFiles:
                lCmdStart.append("os.symlink(\"%s\", os.getcwd() + \"/%s\")" % (file, os.path.basename(file)))
            lCmdsTuples = []
            lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart))
            iLauncher.runLauncherForMultipleJobs("BuildConsLib_%s" % multAlign, lCmdsTuples, cleanMustBeDone)

        os.chdir("..")

    if config.get(sectionNameBuildConsensus, "clean") == "yes":
        FileUtils.removeFilesByPattern("profil*")

    if verbose > 0:
        print "step 4 finished successfully"
        sys.stdout.flush()

#------------------------------------------------------------------------------

def launchShortenConsensusHeader(projectName, smplAlign, clustHSP, multAlign, isStructuralSearch):

    if isStructuralSearch:
        inFile = "%s/%s_LTRharvest_%s_%s/%s_LTRharvest_%s_%s_consensus.fa" % (projectDir, projectName, clustHSP, multAlign, projectName, clustHSP, multAlign)
        outFile = "%s_LTRharvest_%s_%s_consensus_shortH.fa" % (projectName, clustHSP, multAlign)
    else:
        inFile = "%s/%s_%s_%s_%s/%s_%s_%s_%s_consensus.fa" % (projectDir, projectName, smplAlign, clustHSP, multAlign, projectName, smplAlign, clustHSP, multAlign)
        outFile = "%s_%s_%s_%s_consensus_shortH.fa" % (projectName, smplAlign, clustHSP, multAlign)

    if not os.path.exists(inFile):
        msg = "ERROR: consensus file '%s' doesn't exist" % (inFile.replace("%s/" % projectDir, ""))
        sys.stderr.write("%s\n" % msg)
        sys.exit(1)

    cmd = "ShortenConsensusHeader.py"
    cmd += " -p %s" % projectName
    cmd += " -s %s" % smplAlign
    cmd += " -c %s" % clustHSP
    cmd += " -i %s" % inFile
    cmd += " -o %s" % outFile
    cmd += " -m %s" % multAlign
    cmd += " -v %i" % (verbose - 2)
    process = subprocess.Popen(cmd, shell = True)
    process.communicate()
    if process.returncode != 0:
        raise Exception("ERROR when launching '%s'" % cmd)

#------------------------------------------------------------------------------

def detectFeatures(smplAlign, clustHsp, multAlign, isStructuralSearch):

    """
    step 5: detect TE features on the consensus

    @param smplAlign: name of the alignment program (Blaster or Pals)
    @type smplAlign: string

    @param clustHsp: name of the clustering program (Grouper, Recon or Piler)
    @type clustHsp: string

    @param multAlign: name of the multiple alignment program MSA (Map, Mafft, Muscle, Tcoffee, ClustalW or Prank)
    @type multAlign: string
    """

    if isStructuralSearch:
        structClustering = getStructuralSearchClusterTool()

    if verbose > 0:
        print "beginning of step 5"
        if smplAlign == 'Blaster':
            print "self-alignment with %s" % smplAlign
        if isStructuralSearch:
            print "structural detection with LTRharvest"
            if smplAlign == "LTRharvest":
                print "clustering with %s" % structClustering
            else:
                print "clustering with %s and %s" % (clustHsp, structClustering)
        else:
            print "clustering with %s" % clustHsp
        print "multiple alignment with %s" % multAlign
        sys.stdout.flush()

    if verbose > 0:
        print "detect features on the consensus"
        sys.stdout.flush()

    if isStructuralSearch and smplAlign == "Blaster":
        prefix = "%s_%s_%s_Struct_%s" % (projectName, smplAlign, clustHsp, multAlign)
        projectNameDetectFeature = projectName
    else:
        prefix = "%s_%s_%s_%s" % (projectName, smplAlign, clustHsp, multAlign)
        if isStructuralSearch:
            projectNameDetectFeature = "%s_struct" % projectName
        else:
            projectNameDetectFeature = "%s_sim" % projectName

    if isStructuralSearch:
        isStructDir = os.path.exists("%s_LTRharvest_%s_%s" % (projectName, structClustering, multAlign))
        if not isStructDir:
            print "ERROR: run step 4 struct before step 5"
            sys.exit(1)

    clustHspList = []
    if "Grouper" in clustHsp or "Grp" in clustHsp:
        clustHspList.append("Grouper")
    if "Recon" in clustHsp or "Rec" in clustHsp:
        clustHspList.append("Recon")
    if "Piler" in clustHsp or "Pil" in clustHsp:
        clustHspList.append("Piler")

    for clust in clustHspList:
        if not os.path.exists("%s_%s_%s_%s" % (projectName, smplAlign, clust, multAlign)):
            print "ERROR: run step 4 %s before step 5" % clust
            sys.exit(1)

    if os.path.exists("%s_TEclassif" % prefix):
        print "ERROR: directory '%s_TEclassif/' already exists" % prefix
        sys.exit(1)
    os.mkdir("%s_TEclassif" % prefix)
    os.chdir("%s_TEclassif" % prefix)

    if not os.path.exists("detectFeatures"):
        os.mkdir("detectFeatures")
    else:
        if verbose > 0:
            print "** Warning: directory detectFeatures/ already exists"
            sys.stdout.flush()
    os.chdir("detectFeatures")

    if not os.path.exists(configFileName):
        if not os.path.exists("../../" + configFileName):
            print "ERROR: %s not in project directory" % (configFileName)
            sys.exit(1)
        else:
            os.system("ln -s ../../%s ." % (configFileName))

    sectionName = "detect_features"

    if (config.get(sectionName, "TE_BLRtx") == "yes" or config.get(sectionName, "TE_BLRn") == "yes"):
        TE_nucl_bank = config.get(sectionName, "TE_nucl_bank")
        for nuclBank in TE_nucl_bank.split(","):
            if not os.path.exists(nuclBank):
                if not os.path.exists("../../%s" % nuclBank):
                    print "ERROR: %s not in project directory" % nuclBank
                    sys.exit(1)
                else:
                    os.symlink("../../%s" % nuclBank, nuclBank)

    if config.get(sectionName, "TE_BLRx") == "yes":
        TE_prot_bank = config.get(sectionName, "TE_prot_bank")
        for protBank in TE_prot_bank.split(","):
            if not os.path.exists(protBank):
                if not os.path.exists("../../%s" % protBank):
                    print "ERROR: %s not in project directory" % protBank
                    sys.exit(1)
                else:
                    os.symlink("../../%s" % protBank, protBank)

    if config.get(sectionName, "HG_BLRn") == "yes":
        HG_nucl_bank = config.get(sectionName, "HG_nucl_bank")
        if not os.path.exists(HG_nucl_bank):
            if not os.path.exists("../../%s" % HG_nucl_bank):
                print "ERROR: %s not in project directory" % HG_nucl_bank
                sys.exit(1)
            else:
                os.symlink("../../%s" % HG_nucl_bank, HG_nucl_bank)

    if config.get(sectionName, "te_hmmer") == "yes":
        HMM_profiles = config.get(sectionName, "te_hmm_profiles")
        for hmmBank in HMM_profiles.split(","):
            if not os.path.exists(hmmBank):
                if not os.path.exists("../../%s" % hmmBank):
                    print "ERROR: %s not in project directory" % hmmBank
                    sys.exit(1)
                else:
                    os.symlink("../../%s" % hmmBank, hmmBank)

    if config.get(sectionName, "rDNA_BLRn") == "yes":
        rDNA_bank = config.get(sectionName, "rDNA_bank")
        if not os.path.exists(rDNA_bank):
            if not os.path.exists("../../%s" % rDNA_bank):
                print "ERROR: %s not in project directory" % rDNA_bank
                sys.exit(1)
            else:
                os.symlink("../../%s" % rDNA_bank, rDNA_bank)

    repScoutBank = ""
    if config.get(sectionName, "RepScout") == "yes":
        repScoutBank = config.get(sectionName, "RepScout_bank")
        if not os.path.exists(repScoutBank):
            repScoutBank = os.path.basename(repScoutBank)
            if not os.path.exists("../../%s" % repScoutBank):
                print "ERROR: %s not in project directory" % repScoutBank
                sys.exit(1)
            else:
                os.symlink("../../%s" % repScoutBank, repScoutBank)

    doClean = False
    if config.get(sectionName, "clean") == "yes":
        doClean = True

    shortenConsensusFile = "%s_consensus.fa" % prefix

    if not os.path.exists(shortenConsensusFile):
        # shorten the headers of the consensus
        if isStructuralSearch:
            launchShortenConsensusHeader(projectName, "LTRharvest", structClustering, multAlign, isStructuralSearch)

        for clust in clustHspList:
            launchShortenConsensusHeader(projectName, smplAlign, clust, multAlign, False)

        # concatenate all the consensus
        lFiles = glob.glob("%s_*_%s_consensus_shortH*.fa" % (projectName, multAlign))
        if repScoutBank:
            lFiles.append(repScoutBank)
            print "successfully added consensus from RepeatScout bank '%s'" % repScoutBank; sys.stdout.flush()
        FileUtils.catFilesFromList(lFiles, shortenConsensusFile, skipHeaders = False, separator = "")

        # clean
        if os.path.exists("%s_%s_Grouper_%s_consensus_shortH.fa" % (projectName, smplAlign, multAlign)):
            os.system("rm -f %s_%s_Grouper_%s_consensus_shortH.fa" % (projectName, smplAlign, multAlign))
        if os.path.exists("%s_%s_Recon_%s_consensus_shortH.fa" % (projectName, smplAlign, multAlign)):
            os.system("rm -f %s_%s_Recon_%s_consensus_shortH.fa" % (projectName, smplAlign, multAlign))
        if os.path.exists("%s_%s_Piler_%s_consensus_shortH.fa" % (projectName, smplAlign, multAlign)):
            os.system("rm -f %s_%s_Piler_%s_consensus_shortH.fa" % (projectName, smplAlign, multAlign))
        if os.path.exists("%s_LTRharvest_Blastclust_%s_consensus_shortH.fa" % (projectName, multAlign)):
            os.system("rm -f %s_LTRharvest_Blastclust_%s_consensus_shortH.fa" % (projectName, multAlign))

    iDTF = DetectTEFeatures_parallelized(shortenConsensusFile, projectNameDetectFeature, configFileName, verbosity = verbose)
    iDTF.run()

    if (config.get(sectionName, "TE_BLRtx") == "yes" or config.get(sectionName, "TE_BLRn") == "yes"):
        concatRepNTFileName = FileUtils.concatenateFileNamesFromString(TE_nucl_bank, splitSep = ",", joinSep = "_")[0]
        if not FileUtils.isRessourceExists("../../%s" % concatRepNTFileName):
            shutil.move(concatRepNTFileName, "../../%s" % concatRepNTFileName)

    if config.get(sectionName, "TE_BLRx") == "yes":
        concatRepAAFileName = FileUtils.concatenateFileNamesFromString(TE_prot_bank, splitSep = ",", joinSep = "_")[0]
        if not FileUtils.isRessourceExists("../../%s" % concatRepAAFileName):
            shutil.move(concatRepAAFileName, "../../%s" % concatRepAAFileName)

    if config.get(sectionName, "TE_HMMER") == "yes":
        concatHMMFileName = FileUtils.concatenateFileNamesFromString(HMM_profiles, splitSep = ",", joinSep = "_")[0]
        if not FileUtils.isRessourceExists("../../%s" % concatHMMFileName):
            shutil.move(concatHMMFileName, "../../%s" % concatHMMFileName)

    os.chdir ("../..")

    if verbose > 0:
        print "step 5 finished successfully"; sys.stdout.flush()

#------------------------------------------------------------------------------

class ClassifConsensus(object):

    def __init__(self, smplAlign, clustHsp, multAlign, isStructuralSearch, configFileName, verbose):
        self._smplAlign = smplAlign
        self._clustHsp = clustHsp
        self._multAlign = multAlign
        self._isStructuralSearch = isStructuralSearch
        self._branch = ''
        self._configFileName = configFileName
        self._iConfigValue = None
        self._sectionName = "classif_consensus"
        self._projectName = ""
        self._projectDir = ""
        self._verbose = verbose
        self._clean = False

    def run(self):

        if self._isStructuralSearch:
            structClustering = getStructuralSearchClusterTool()

        if self._verbose > 0:
            print "beginning of step 6"
            if self._smplAlign == 'Blaster':
                print "self-alignment with %s" % self._smplAlign
            if self._isStructuralSearch:
                print "structural detection with LTRharvest"
                if self._smplAlign == "LTRharvest":
                    print "clustering with %s" % structClustering
                else:
                    print "clustering with %s and %s" % (self._clustHsp, structClustering)
            else:
                print "clustering with %s" % self._clustHsp
            print "multiple alignment with %s" % self._multAlign
            print "classify the consensus according to their features (PASTEC)"
            sys.stdout.flush()

        self._checkConfig()
        self._projectName = self._iConfigValue.get("project", "project_name")
        self._projectDir = self._iConfigValue.get("project", "project_dir")
        if self._iConfigValue.get(self._sectionName, "clean") == "yes":
            self._clean = True

        if self._isStructuralSearch and self._smplAlign == "Blaster":
            self._prefix = "%s_%s_%s_Struct_%s" % (self._projectName, self._smplAlign, self._clustHsp, self._multAlign)
            projectNameDetectFeature = projectName
        if self._isStructuralSearch and self._smplAlign == "LTRharvest":
            self._prefix = "%s_%s_%s_%s" % (self._projectName, self._smplAlign, structClustering, self._multAlign)
            projectNameDetectFeature = "%s_struct" % projectName
        if not self._isStructuralSearch:
            self._prefix = "%s_%s_%s_%s" % (self._projectName, self._smplAlign, self._clustHsp, self._multAlign)
            projectNameDetectFeature = "%s_sim" % projectName

        if not os.path.exists("%s_TEclassif/detectFeatures" % self._prefix):
            print "ERROR: run step 5 before step 6"
            sys.exit(1)
        os.chdir("%s_TEclassif" % self._prefix)
        if os.path.exists("classifConsensus"):
            print "ERROR: directory %s_TEclassif/classifConsensus already exists" % self._prefix
            sys.exit(1)
        os.mkdir("classifConsensus")
        os.chdir("classifConsensus")
        os.symlink("../../%s" % self._configFileName, self._configFileName)
        consensusFileName = "%s_consensus.fa" % self._prefix
        os.symlink("../detectFeatures/%s" % consensusFileName, consensusFileName)

        if self._checkFeatureTables(projectNameDetectFeature) == 0:
            message = "ERROR: no table recording TE features exists"
            if self._verbose > 0:
                print message
                sys.stdout.flush()
            sys.exit(1)
        iPC = PASTEClassifier(steps = "2", parallel = True)
        iPC.setProjectName(projectNameDetectFeature)
        iPC.setConfigFileName(configFileName)
        iPC.setFastaFileName(consensusFileName)
        iPC.setDoClean(self._clean)
        iPC.run()

        if self._verbose > 0:
            print "step 6 finished successfully"
            sys.stdout.flush()

    def _checkConfig(self):
        iConfigRules = ConfigRules()
        iConfigRules.addRuleSection(self._sectionName, True)
        iConfigRules.addRuleOption(self._sectionName, "min_redundancy_identity", True)
        iConfigRules.addRuleOption(self._sectionName, "min_redundancy_coverage", True)
        iConfigRules.addRuleOption(self._sectionName, "max_profiles_evalue")
        iConfigRules.addRuleOption(self._sectionName, "min_TE_profiles_coverage")
        iConfigRules.addRuleOption(self._sectionName, "min_HG_profiles_coverage")
        iConfigRules.addRuleOption(self._sectionName, "max_helitron_extremities_evalue")
        iConfigRules.addRuleOption(self._sectionName, "min_TE_bank_coverage")
        iConfigRules.addRuleOption(self._sectionName, "min_HG_bank_coverage")
        iConfigRules.addRuleOption(self._sectionName, "min_rDNA_bank_coverage")
        iConfigRules.addRuleOption(self._sectionName, "min_HG_bank_identity")
        iConfigRules.addRuleOption(self._sectionName, "min_rDNA_bank_identity")
        iConfigRules.addRuleOption(self._sectionName, "min_SSR_coverage")
        iConfigRules.addRuleOption(self._sectionName, "max_SSR_size")
        iConfigRules.addRuleOption(self._sectionName, "rev_complement")
        iConfigRules.addRuleOption(self._sectionName, "add_wicker_code")
        iConfigRules.addRuleOption(self._sectionName, "add_noCat_bestHitClassif")
        iConfigRules.addRuleOption(self._sectionName, "clean", True)
        iConfigRules.addRuleOption(self._sectionName, "tmpDir", True)
        iConfigRules.addRuleOption(self._sectionName, "resources", True)

        iConfigChecker = ConfigChecker(self._configFileName, iConfigRules)
        self._iConfigValue = iConfigChecker.getConfig()

    ## Check if the different tables recording sequence features (tandem repeats, matches with known TEs...) exist or not.
    def _checkFeatureTables(self, projectNameDetectFeature):
        exist = 0
        tableSuffix = ["_TR_set",
                       "_SSR_set",
                       "_polyA_set",
                       "_ORF_map",
                       "_TE_BLRn_path",
                       "_TE_BLRtx_path",
                       "_TE_BLRx_path",
                       "_RepbaseNT_seq",
                       "_RepbaseAA_seq",
                       "_HG_BLRn_path",
                       "_rDNA_BLRn_path",
                       "_Profiles_path",
                       "_ProfilesDB_map"]
        iDb = DbFactory.createInstance()
        for suffix in tableSuffix:
            if iDb.doesTableExist(projectNameDetectFeature + suffix):
                message = "table '%s%s' exists" % (projectNameDetectFeature, suffix)
                message += " (%i lines)" % (iDb.getSize(projectNameDetectFeature + suffix))
                if self._verbose > 0:
                    print message
                    sys.stdout.flush()
                exist += 1
            else:
                message = "table '%s%s' doesn't exist" % (projectNameDetectFeature, suffix)
                if self._verbose > 0:
                    print message
                    sys.stdout.flush()
        iDb.close()
        return exist

    def getPASTECcommand(self, projectNameDetectFeature, iLauncher, fileName):
        lArgs = []
        lArgs.append("-C %s" % self._configFileName)
        lArgs.append("-P %s" % projectNameDetectFeature)
        lArgs.append("-S 2")
        lArgs.append("-i %s" % fileName)
        lArgs.append("-v %s" % self._verbose)
        return iLauncher.getSystemCommand("LaunchPASTEC.py", lArgs)

#------------------------------------------------------------------------------

def filterConsensus(smplAlign, clustHsp, multAlign, isStructuralSearch):

    """
    step 7: filter the consensus: SSR, NoCat, HostGenes, PotentialChimeric, incomplete, rDNA 

    @param smplAlign: name of the alignment program (Blaster or Pals)
    @type smplAlign: string

    @param clustHsp: name of the clustering program (Grouper, Recon or Piler)
    @type clustHsp: string

    @param multAlign: name of the multiple alignment program MSA (Map, Mafft, Muscle, Tcoffee, ClustalW or Prank)
    @type multAlign: string
    """

    if isStructuralSearch:
        structClustering = getStructuralSearchClusterTool()

    if verbose > 0:
        print "beginning of step 7"
        if smplAlign == 'Blaster':
            print "self-alignment with %s" % smplAlign
        if isStructuralSearch:
            print "structural detection with LTRharvest"
            if smplAlign == "LTRharvest":
                print "clustering with %s" % structClustering
            else:
                print "clustering with %s and %s" % (clustHsp, structClustering)
        else:
            print "clustering with %s" % clustHsp
        print "multiple alignment with %s" % multAlign
        print "filter the consensus (SSR and noCat)"
        sys.stdout.flush()

    if isStructuralSearch and smplAlign == "Blaster":
        prefix = "%s_%s_%s_Struct_%s" % (projectName, smplAlign, clustHsp, multAlign)
        projectNameFromClassif = projectName

    if isStructuralSearch and smplAlign == "LTRharvest":
        prefix = "%s_%s_%s_%s" % (projectName, smplAlign, clustHsp, multAlign)
        projectNameFromClassif = "%s_struct" % projectName

    if not isStructuralSearch:
        prefix = "%s_%s_%s_%s" % (projectName, smplAlign, clustHsp, multAlign)
        projectNameFromClassif = "%s_sim" % projectName

    if not os.path.exists("%s_TEclassif/classifConsensus" % prefix):
        print "ERROR: run step 6 before step 7"
        sys.exit(1)
    else:
        os.mkdir("%s_TEclassif_Filtered" % prefix)
        os.chdir("%s_TEclassif_Filtered" % prefix)

    if not os.path.exists("%s/%s" % (projectDir, configFileName)):
        print "ERROR: %s not in project directory" % configFileName
        sys.exit(1)
    else:
        os.symlink("%s/%s" % (projectDir, configFileName), configFileName)

    sectionName = "filter_consensus"
    if config.get(sectionName, "filter_noCat") == "yes":
        if config.get(sectionName, "filter_noCat_max_fragments") == "":
            print "ERROR: incorrect 'filter_noCat_max_fragments' value in config"
            sys.exit(1)

    fileName = "%s_denovoLibTEs.fa" % projectNameFromClassif
    pathAndFileName = "%s/%s_TEclassif/classifConsensus/%s" % (projectDir, prefix, fileName)
    if not os.path.exists(pathAndFileName):
        msg = "ERROR: fasta file '%s' doesn't exist" % pathAndFileName
        sys.stderr.write("%s\n" % msg)
        sys.exit(1)

    cmd = "FilterClassifiedSequences.py"
    cmd += " -i %s" % pathAndFileName
    cmd += " -o %s_filtered.fa" % os.path.splitext(fileName)[0]
    if config.get(sectionName, "filter_SSR") == "yes":
        cmd += " -S"
        cmd += " -s %s" % config.get(sectionName, "length_SSR")
    if config.get(sectionName, "filter_noCat") == "yes":
        maxMembers = config.get(sectionName, "filter_noCat_max_fragments")
        cmd += " -N -n %s" % maxMembers
    if config.get(sectionName, "filter_host_gene") == "yes":
        cmd += " -g"
    if config.get(sectionName, "filter_potential_chimeric") == "yes":
        cmd += " -c"
    if config.get(sectionName, "filter_incomplete") == "yes":
        cmd += " -u"
    if config.get(sectionName, "filter_rDNA") == "yes":
        cmd += " -r"
    cmd += " -v %i" % (verbose - 2)
    process = subprocess.Popen(cmd, shell = True)
    process.communicate()
    if process.returncode != 0:
        raise Exception("ERROR when launching '%s'" % cmd)

    os.chdir ("../..")

    if verbose > 0:
        print "step 7 finished successfully"
        sys.stdout.flush()

#------------------------------------------------------------------------------

def getLaunchPostAnalyzeTELib(filteredFileName, sectionName, iLauncher, clusterMethod):
    lArgs = []
    lArgs.append("-i %s" % filteredFileName)

    if clusterMethod == "Blastclust":
        lArgs.append("-S %s" % config.get(sectionName, "Blastclust_identity"))
        lArgs.append("-L %s" % config.get(sectionName, "Blastclust_coverage"))
        lArgs.append("-b F")
    elif clusterMethod == "MCL":
        lArgs.append("-M %s" % clusterMethod)
        lArgs.append("-I %s" % config.get(sectionName, "MCL_inflation"))
        lArgs.append("-T %s" % config.get(sectionName, "MCL_coverage"))
    lArgs.append("--isClusterConsensusHeaders")
    lArgs.append("-v %i" % (verbose - 1))

    return iLauncher.getSystemCommand("PostAnalyzeTELib.py", lArgs)

def clusteringConsensus(smplAlign, clustHsp, multAlign, isStructuralSearch, step8ClustMethod):

    """
    step 8: clustering the consensus 
    
    @param smplAlign: name of the alignment program (Blaster or Pals)
    @type smplAlign: string

    @param clustHsp: name of the clustering program (Grouper, Recon or Piler)
    @type clustHsp: string

    @param multAlign: name of the multiple alignment program MSA (Map, Mafft, Muscle, Tcoffee, ClustalW or Prank)
    @type multAlign: string
    """
    _log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, "clusteringConsensus"), verbose)
    # TODO: set level to verbosity attribute when everything is using logger
    LoggerFactory.setLevel(_log, verbose + 2)

    if isStructuralSearch:
        structClustering = getStructuralSearchClusterTool()

    if step8ClustMethod.lower() == "blastclust":
        step8ClustMethod = "Blastclust"
    elif step8ClustMethod.lower() == "mcl":
        step8ClustMethod = "MCL"
    else:
        print "%s isn't supported." % step8ClustMethod
        sys.exit(1)

    if verbose > 0:
        _log.info("beginning of step 8")
        if smplAlign == 'Blaster':
            _log.info("self-alignment with %s" % smplAlign)
        if isStructuralSearch:
            _log.info("structural detection with LTRharvest")
            if smplAlign == "LTRharvest":
                _log.info("clustering with %s" % structClustering)
            else:
                _log.info("clustering with %s and %s" % (clustHsp, structClustering))
        else:
            _log.info("clustering with %s" % clustHsp)
        _log.info("multiple alignment with %s" % multAlign)
        _log.info("clustering the filtered consensus with %s" % step8ClustMethod)
        sys.stdout.flush()

    if isStructuralSearch and smplAlign == "Blaster":
        prefix = "%s_%s_%s_Struct_%s" % (projectName, smplAlign, clustHsp, multAlign)
        projectNameFromFiltering = projectName

    if isStructuralSearch and smplAlign == "LTRharvest":
        prefix = "%s_%s_%s_%s" % (projectName, smplAlign, clustHsp, multAlign)
        projectNameFromFiltering = "%s_struct" % projectName

    if not isStructuralSearch:
        prefix = "%s_%s_%s_%s" % (projectName, smplAlign, clustHsp, multAlign)
        projectNameFromFiltering = "%s_sim" % projectName

    if not os.path.exists("%s_TEclassif_Filtered/%s_denovoLibTEs_filtered.fa" % (prefix, projectNameFromFiltering)):
        _log.info("ERROR: run step 7 before step 8")
        sys.exit(1)
    else:
        os.mkdir("%s_TEclassif_Filtered_%s" % (prefix, step8ClustMethod))
        os.chdir("%s_TEclassif_Filtered_%s" % (prefix, step8ClustMethod))

    if not os.path.exists("%s/%s" % (projectDir, configFileName)):
        _log.info("ERROR: %s not in project directory")
        sys.exit(1)
    else:
        os.symlink("%s/%s" % (projectDir, configFileName), configFileName)

    sectionName = "cluster_consensus"
    # TODO: these method has to be selfed when step 8 will be a class. Have a look on TEClustering class for example.
    _checkConfigForStep8(sectionName, step8ClustMethod)

#    filteredFileName = "%s_denovoLibTEs_filtered.fa" % projectName
    filteredFileName = "%s_denovoLibTEs_filtered.fa" % projectNameFromFiltering
    absPathFilteredFile = os.path.abspath("../%s_TEclassif_Filtered/%s" % (prefix, filteredFileName))
    os.symlink("../%s_TEclassif_Filtered/%s" % (prefix, filteredFileName), filteredFileName)

    queue = config.get(sectionName, "resources")
    cDir = os.getcwd()
    if config.get(sectionName, "tmpDir") != "":
        tmpDir = config.get(sectionName, "tmpDir")
    else:
        tmpDir = cDir
    groupid = "%s_TEdenovo_S8_%s" % (projectName, step8ClustMethod)
    acronym = "Launch_%s" % step8ClustMethod
    iDb = DbFactory.createInstance()
    jobdb = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
    iLauncher = Launcher(jobdb, cDir, tmpDir, queue, groupid)
    lCmds = []
    lCmds.append(getLaunchPostAnalyzeTELib(filteredFileName, sectionName, iLauncher, step8ClustMethod))
    lCmdStart = []
    lCmdStart.append("shutil.copy(\"%s\", \".\")" % absPathFilteredFile)
    lCmdFinish = []
    lCmdsTuples = []
    filteredFileName = os.path.splitext(filteredFileName)[0]
    if step8ClustMethod == 'Blastclust':
        lCmdFinish.append("shutil.move(\"%s.fa_Blastclust.fa\", \"%s/%s_Blastclust.fa\")" % (filteredFileName, cDir, filteredFileName))
        lCmdFinish.append("shutil.move(\"%s.fa_blastclust.tab\", \"%s/%s_Blastclust_clusterCons.tab\")" % (filteredFileName, cDir, filteredFileName))
        lCmdFinish.append("shutil.move(\"%s.fa_blastclust.globalStatsPerCluster.txt\", \"%s/%s_Blastclust_globalStatsPerCluster.txt\")" % (filteredFileName, cDir, filteredFileName))
        lCmdFinish.append("shutil.move(\"%s.fa_blastclust.statsPerCluster.tab\", \"%s/%s_Blastclust.statsPerCluster.tab\")" % (filteredFileName, cDir, filteredFileName))

    elif step8ClustMethod == 'MCL':
        lCmdFinish.append("shutil.move(\"%s_MCL.fa\", \"%s/%s_MCL.fa\")" % (filteredFileName, cDir, filteredFileName))
        lCmdFinish.append("shutil.move(\"%s.fa.newHlink\", \"%s/%s_MCL_clusterCons.tab\")" % (filteredFileName, cDir, filteredFileName))
        lCmdFinish.append("shutil.move(\"%s.globalStatsPerCluster.txt\", \"%s/%s_MCL_globalStatsPerCluster.txt\")" % (filteredFileName, cDir, filteredFileName))
        lCmdFinish.append("shutil.move(\"%s.statsPerCluster.tab\", \"%s/%s_MCL_statsPerCluster.tab\")" % (filteredFileName, cDir, filteredFileName))
    cleanMustBeDone = False

    if config.get(sectionName, "clean") == "yes":
        cleanMustBeDone = True

    lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart, lCmdFinish))
    iLauncher.runLauncherForMultipleJobs(acronym, lCmdsTuples, cleanMustBeDone)
    os.chdir("../..")

    if verbose > 0:
        print "step 8 finished successfully"
        sys.stdout.flush()

def _checkConfigForStep8(sectionName, step8ClustMethod):

    try:
        CheckerUtils.checkSectionInConfigFile(config, sectionName)
    except NoSectionError:
        print "ERROR: the section %s must be in your configuration file" % sectionName
        sys.exit(1)
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "resources")
    except NoOptionError:
        print "ERROR: the option 'resources' must be defined in %s in your configuration file" % sectionName
        sys.exit(1)
    try:
        CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "tmpDir")
    except NoOptionError:
        print "ERROR: the option 'tmpDir' must be defined in %s in your configuration file" % sectionName
        sys.exit(1)

    if step8ClustMethod == "Blastclust":
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "Blastclust_identity")
        except NoOptionError:
            print "ERROR: the option 'Blastclust_identity' must be defined in " + sectionName + " in your config file : " + configFileName
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "Blastclust_coverage")
        except NoOptionError:
            print "ERROR: the option 'Blastclust_coverage' must be defined in " + sectionName + " in your config file : " + configFileName
            sys.exit(1)

        Blastclust_identity = int(config.get(sectionName, "Blastclust_identity"))
        if Blastclust_identity < 0 or Blastclust_identity > 100:
            print "ERROR: bad value for option 'Blastclust_identity' in " + sectionName + " in your config file : '" + Blastclust_identity + "'. Must be >=0 or <=100."
            sys.exit(1)

        Blastclust_coverage = int(config.get(sectionName, "Blastclust_coverage"))
        if Blastclust_coverage < 0 or Blastclust_coverage > 100:
            print "ERROR: bad value for option 'Blastclust_coverage' in " + sectionName + " in your config file : '" + Blastclust_coverage + "'. Must be >=0 or <=100."
            sys.exit(1)

    elif step8ClustMethod == "MCL":
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "MCL_inflation")
        except NoOptionError:
            print "ERROR: the option 'MCL_inflation' must be defined in " + sectionName + " in your config file : " + configFileName
            sys.exit(1)
        try:
            CheckerUtils.checkOptionInSectionInConfigFile(config, sectionName, "MCL_coverage")
        except NoOptionError:
            print "ERROR: the option 'MCL_coverage' must be defined in " + sectionName + " in your config file : " + configFileName
            sys.exit(1)

        MCL_coverage = float(config.get(sectionName, "MCL_coverage"))
        if MCL_coverage < 0 or MCL_coverage > 1:
            print "ERROR: bad value for option 'MCL_coverage' in " + sectionName + " in your config file : '" + MCL_coverage + "'. Must be >=0 or <=1."
            sys.exit(1)

if __name__ == "__main__":
    main()
