#!/usr/bin/env python

import sys
import os
import getopt
import shutil
import re
import datetime

class TEdenovoToGnpSeq( object ):
    
    def __init__( self ):
        self._libTEs = ""
        self._configFile = ""
        self._classifFile = ""
        self._genera = ""
        self._species = ""
        self._genomeRelease = ""
        self._analysisName = ""
        self._relRepet = ""
        self._outputDir = ""
        self._verbose = 0
        self._projectCode = ""
        
        self._sequenceName = ""
        self._className = ""
        self._orderName = ""
        self._superfamilyName = ""
        self._classificationStatus = ""
        self._completeness = ""
        self._clusterName = ""
        self._sequence = ""
        
    def help( self ):
        print
        print "usage: %s [ options ]" % ( sys.argv[0].split("/")[-1] )
        print "options:"
        print "     -h: this help"
        print "     -i: file with TE consensus from TEdenovo (format=fasta)"
        print "         e.g. 'DmelChr4_Blaster_Piler_Map_uniq_TEclassifier.fa'"
        print "     -C: configuration file from TEdenovo"
        print "         e.g. 'DmelChr4_TEdenovo.cfg'"
        print "     -c: classification file from TEclassifier (format=classif)"
        print "         e.g. 'DmelChr4_Blaster_Piler_Map_uniq_TEclassifier.classif'"
        print "     -g: genera of the organism studied (e.g. 'Drosophila')"
        print "     -s: species of the organism studied (e.g. 'melanogaster')"
        print "     -r: genome release (e.g. 'Caf1')"
        print "     -a: analysis name (e.g. 'TEdenovo on DmelChr4')"
        print "     -d: date at which the analysis has been done (format=dd/mm/yyyy)"
        print "     -R: release of REPET used to make the analysis (e.g. '1.3.9')"
        print "     -o: output directory (e.g. 'TE_Submission')"
        print "     -v: verbosity level (default=0/1/2)"
        print
        
        
    def setAttributesFromCmdLine( self ):
        try:
            opts, args = getopt.getopt( sys.argv[1:], "hi:C:c:g:s:r:a:d:D:R:o:v:" )
        except getopt.GetoptError, err:
            sys.stderr.write( "%s\n" % ( str(err) ) )
            self.help()
            sys.exit(1)
        for o,a in opts:
            if o == "-h":
                self.help();
                sys.exit(0)
            elif o == "-i":
                self._libTEs = a
            elif o == "-C":
                self._configFile = a
            elif o == "-c":
                self._classifFile = a
            elif o == "-g":
                self._genera = a
            elif o == "-s":
                self._species = a
            elif o == "-r":
                self._genomeRelease = a
            elif o == "-R":
                self._relRepet = a
            elif o == "-v":
                self._verbose = int(a)
            elif o == "-a":
                self._analysisName = a
            elif o == "-o":
                self._outputDir = a

                
    def checkAttributes( self ):
        if self._libTEs == "":
            msg = "ERROR: missing TE library (-i)"
            sys.stderr.write( "%s\n" % ( msg ) )
            self.help()
            sys.exit(1)
        if not os.path.exists( self._libTEs ):
            msg = "ERROR: can't find file '%s'" % ( self._libTEs )
            sys.stderr.write( "%s\n" % ( msg ) )
            self.help()
            sys.exit(1)
        if self._configFile == "":
            msg = "ERROR: missing configuration file (-C)"
            sys.stderr.write( "%s\n" % ( msg ) )
            self.help()
            sys.exit(1)
        if not os.path.exists( self._configFile ):
            msg = "ERROR: can't find file '%s'" % ( self._configFile )
            sys.stderr.write( "%s\n" % ( msg ) )
            self.help()
            sys.exit(1)
        if self._classifFile != "" and not os.path.exists( self._classifFile ):
            msg = "ERROR: can't find file '%s'" % ( self._classifFile )
            sys.stderr.write( "%s\n" % ( msg ) )
            self.help()
            sys.exit(1)
        if self._genera == "":
            msg = "ERROR: missing the genera of the organism (-g)"
            sys.stderr.write( "%s\n" % ( msg ) )
            self.help()
            sys.exit(1)
        if self._species == "":
            msg = "ERROR: missing the species of the organism (-s)"
            sys.stderr.write( "%s\n" % ( msg ) )
            self.help()
            sys.exit(1)
        if self._genomeRelease == "":
            msg = "ERROR: missing the genome release (-g)"
            sys.stderr.write( "%s\n" % ( msg ) )
            self.help()
            sys.exit(1)
        if self._relRepet == "":
            msg = "ERROR: missing the release of REPET (-R)"
            sys.stderr.write( "%s\n" % ( msg ) )
            self.help()
            sys.exit(1)
        if self._analysisName == "":
            msg = "ERROR: missing the name of the analysis (-a)"
            sys.stderr.write( "%s\n" % ( msg ) )
            self.help()
            sys.exit(1)
        if self._outputDir == "":
            msg = "ERROR: missing the name of the output directory (-o)"
            sys.stderr.write( "%s\n" % ( msg ) )
            self.help()
            sys.exit(1)
        else:
            if not os.path.exists(self._outputDir):
                os.makedirs(self._outputDir)
            
        self.setProjectCode()
        
        
    def setProjectCode( self ):
        self._projectCode = "%s%s" % ( self._genera[0], self._species[0:3] )
       
        
    def getContactDataAsString( self ):
        data = ""
        data += "CONTACT_NUMBER"
        data += "\tCONTACT_NAME"
        data += "\tFIRST_NAME"
        data += "\tfax"
        data += "\ttelephone"
        data += "\tEMAIL"
        data += "\tlaboratory"
        data += "\tINSTITUTION"
        data += "\tADDRESS"
        data += "\tCITY"
        data += "\tstate"
        data += "\tZIP_CODE"
        data += "\tCOUNTRY"
        data += "\twebsite"
        data += "\n"
        
        data += "1"
        data += "\tQUESNEVILLE"
        data += "\tHadi"
        data += "\t"
        data += "\t"
        data += "\thadi.quesneville@versailles.inra.fr"
        data += "\tURGI"
        data += "\tINRA"
        data += "\tINRA, centre de Versailles-Grignon - RD 10 route de Saint Cyr"
        data += "\tVersailles"
        data += "\t"
        data += "\t78026"
        data += "\tFRANCE"
        data += "\thttp://urgi.versailles.inra.fr"
        data += "\n"
        
        data += "2"
        data += "\t<YOUR_FAMILY_NAME>"
        data += "\t<Your_first_name>"
        data += "\t"
        data += "\t"
        data += "\t<your_email>"
        data += "\t<your_lab>"
        data += "\t<your_institution>"
        data += "\t<your_lab_adress>"
        data += "\t<your_lab_city>"
        data += "\t"
        data += "\t<your_lab_ZIP_code>"
        data += "\t<your_country>"
        data += "\t<your_website>"
        data += "\n"
        
        return data
    
    
    def getProjectDataAsString( self ):
        data = ""
        data += "PROJECT_CODE"
        data += "\tPROJECT_NAME"
        data += "\tTITLE"
        data += "\tPROJECT_COORDINATOR"
        data += "\tBIOINFORMATICS_CONTACT"
        data += "\tpartner_contact"
        data += "\tbegining_date"
        data += "\tcomment"
        data += "\tAFFILIATION"
        data += "\tref_number"
        data += "\n"
        
        data += "%sTEannotation" % ( self._projectCode )
        data += "\tTE annotation of %s. %s genome" % ( self._genera[0], self._species )
        data += "\tTE annotation of %s %s genome sequences" % ( self._genera, self._species )
        data += "\t1"
        data += "\t2"
        data += "\t<partner_contacts>"
        data += "\t<begining_date>"
        data += "\t<comments>"
        data += "\tINRA"
        data += "\t<references>"
        data += "\n"
        
        return data
    
    
    def getReferenceDataAsString(self):
        data = ""
        data += "REF_NUMBER"
        data += "\tAUTHORS"
        data += "\tTITLE"
        data += "\tJOURNAL"
        data += "\tVOLUME"
        data += "\tsubvolume"
        data += "\tpages"
        data += "\tYEAR"
        data += "\tmedline_number"
        data += "\n"
        
        data += "<reference_number>"
        data += "\t<authors>"
        data += "\t<title>"
        data += "\t<journal>"
        data += "\t<volume>"
        data += "\t<subvolume>"
        data += "\t<pages>"
        data += "\t<year>"
        data += "\t<medline_number>"
        data += "\n"
        
        return data
    
    
    def getAnalysisDataAsString(self):
        data = ""
        data += "ANALYSIS_NAME"
        data += "\tANALYSIS_TYPE"
        data += "\tanalysis_date"
        data += "\tORGANISM"
        data += "\tGENOME_VERSION"
        data += "\tSOFTWARE_NUMBER"
        data += "\tCONFIGURATION_FILE"
        data += "\tcommand_line"
        data += "\tcomments"
        data += "\tCONTACT_NUMBER"
        data += "\tproject_code"
        data +=  "\n"
        
        data += "%s" % ( self._analysisName )
        data += "\tTE detection"
        data += "\t<analysis_date>"
        data += "\t%s %s" % ( self._genera, self._species )
        data += "\t%s" % ( self._genomeRelease )
        data += "\t1"
        data += "\t%s" % ( os.path.basename(self._configFile) )
        data += "\t<command_line>"
        data += "\t<comment>"
        data += "\t2"
        data += "\t%sTEannotation" % ( self._projectCode )
        data += "\n"
        
        return data   
    
    
    def getSoftwareDataAsString( self ):
        data = ""
        data += "SOFTWARE_NUMBER"
        data += "\tSOFTWARE_NAME"
        data += "\tVERSION"
        data += "\tTYPE"
        data += "\turl_get"
        data += "\turl_doc"
        data += "\tftp_site"
        data += "\tlocal"
        data += "\tGENOPLANTE"
        data += "\tFREE"
        data += "\tdescription"
        data += "\n"
        
        data += "1"
        data += "\tTEdenovo"
        data += "\t%s" % (self._relRepet)
        data += "\tREPET"
        data += "\thttp://urgi.versailles.inra.fr/index.php/urgi/Tools/REPET"
        data += "\t"
        data += "\thttp://urgi.versailles.inra.fr/download/repet/"
        data += "\tno"
        data += "\tno"
        data += "\tyes"
        data += "\t"
        data += "\n"
        
        return data        
    
    
    def getDataFromClassifFile( self ):
        dHeader2Classif = {}
        if self._classifFile == "":
            return dHeader2Classif
        if self._verbose > 0:
            print "parse classification file"
            sys.stdout.flush()
        classifFile = open( self._classifFile, "r" )
        while True:
            line = classifFile.readline()
            if line == "":
                break
            tokens = line[:-1].split("\t")
            seqname = tokens[0]
            seqlength = tokens[1]
            strand = tokens[2]
            confusedness = tokens[3]
            category = tokens[4].replace(" ","")
            order = tokens[5]
            superfam = "?"
            if len(tokens) == 8:
                completeness = tokens[6]
                comments=tokens[7]
            else:
                superfam = tokens[6]
                completeness = tokens[7]
                comments=tokens[8]
            if dHeader2Classif.has_key( seqname ):
                msg = "ERROR: two sequences have the same header"
                sys.stderr.write( "%s\n" % msg )
                sys.exit(1)
            classif = "%s bp" % ( seqlength )
            if confusedness == "confused":
                classif += ";confused"
            if category != "?":
                classif += ";%s" % ( category )
            if order != "?":
                classif += ";%s" % ( order )
            if superfam != "?":
                classif += ";%s" % ( superfam )
            if completeness == "comp":
                classif += ";complete"
            else:
                classif += ";incomplete"
            classif += ";%s" % ( comments )
            dHeader2Classif[ seqname ] = classif
        return dHeader2Classif
    
    
    def getTEconsensusDataAsString(self):
        stringTEConsensus = "%s" % ( self._sequenceName )
        stringTEConsensus += "\t"
        stringTEConsensus += "\t%s" % ( self._className )
        stringTEConsensus += "\t%s" % ( self._orderName )
        stringTEConsensus += "\t%s" % ( self._superfamilyName )
        stringTEConsensus += "\t%s" % ( self._completeness )
        stringTEConsensus += "\t%s" % ( self._classificationStatus )
        stringTEConsensus += "\tTEdenovo"
        stringTEConsensus += "\t"
        stringTEConsensus += "\t%s" % ( self._clusterName )
        stringTEConsensus += "\t"
        stringTEConsensus += "\t%s" % ( self._analysisName )
        stringTEConsensus += "\t%s" % ( self._sequence )
        stringTEConsensus += "\n"
        
        return stringTEConsensus
    
    
    def _getTEconsensusHeaderAsString(self):
        stringTEConsensus = "SEQUENCE_NAME"
        stringTEConsensus += "\tclassification_code"
        stringTEConsensus += "\tclass_name"
        stringTEConsensus += "\torder_name"
        stringTEConsensus += "\tsuperfamily_name"
        stringTEConsensus += "\tcompleteness"
        stringTEConsensus += "\tclassification_status"
        stringTEConsensus += "\tANNOTATION_STATUS"
        stringTEConsensus += "\tchimeric"
        stringTEConsensus += "\tcluster_name"
        stringTEConsensus += "\tdescription"
        stringTEConsensus += "\tANALYSIS_NAME"
        stringTEConsensus += "\tSEQUENCE"
        stringTEConsensus += "\n"
    
        return stringTEConsensus
        
        
    def getDataFromFastaFile( self ):
        if self._verbose > 0:
            print "parse fasta file"
            sys.stdout.flush()
        fastaFile = open( self._libTEs, "r" )
        classif = ""
        
        data = self._getTEconsensusHeaderAsString()
        
        while True:
            line = fastaFile.readline()
            if line == "":
                if classif != "":
                    data += self.getTEconsensusDataAsString()
                break
            
            if re.search('^>', line):
                if classif != "":
                    data += self.getTEconsensusDataAsString()
                    self._resetTEconsensusAttributs()
                    
                m = re.search('^>BlastclustCluster([0-9]+)Mb[0-9]+\s(([\w\.]+-)+Map[0-9]+_([\w-]+))$', line)
                self._clusterName = m.group(1)
                self._sequenceName = m.group(2)
                
                classif = m.group(4)
                lclassif = classif.split("-")
                if len(lclassif) == 1:
                    self._classificationStatus = lclassif[0]
                elif len(lclassif) == 2:
                    self._className = lclassif[0]
                    self._completeness = lclassif[1]
                elif len(lclassif) == 3:
                    self._className = lclassif[0]
                    self._orderName = lclassif[1]
                    self._completeness = lclassif[2]
                elif len(lclassif) == 4:
                    self._className = lclassif[0]
                    self._orderName = lclassif[1]
                    self._superfamilyName = lclassif[2]
                    self._completeness = lclassif[3]
            else:
                self._sequence = "%s%s" %(self._sequence, line.rstrip())
                                
        fastaFile.close()
        return data
    
    
    def _writeData( self, data, fileName ):
        if self._verbose > 0:
            msg = "save file '%s'" % ( fileName )
            sys.stdout.write( "%s\n" % msg )
            sys.stdout.flush()
        handler = open( fileName, "w" )
        handler.write( data )
        handler.close()
        
        
    def start( self ):
        self.checkAttributes()
        if self._verbose > 0:
            msg = "START %s" % ( sys.argv[0].split("/")[-1] )
            sys.stdout.write( "%s\n" % msg )
            sys.stdout.flush()
        msg = "This program generates submission files for GnpSeq but beforehand you will have to edit some of them."
        sys.stdout.write( "%s\n" % msg )
        sys.stdout.flush()
        
        
    def end( self ):
        msg = "\nSubmission files generated.\n"
        msg += "Complete project.csv, contact.csv and analysis.csv (reference.csv is optional).\n"
        msg += "Edit or remove entries in <>. Columns in upper case are mandatory.\n"
        sys.stdout.write( "%s\n" % msg )
        sys.stdout.flush()
        if self._verbose > 0:
            msg = "END %s" % ( sys.argv[0].split("/")[-1] )
            sys.stdout.write( "%s\n" % msg )
            sys.stdout.flush()
     
     
    def _resetTEconsensusAttributs(self) :
        self._sequenceName = ""
        self._className = ""
        self._orderName = ""
        self._superfamilyName = ""
        self._classificationStatus = ""
        self._completeness = ""
        self._clusterName = ""
        self._sequence = "" 
            
    def run( self ):
        self.start()
         
        if not (self._outputDir):
            os.makedirs(self._outputDir)
        
        contact = self.getContactDataAsString()
        self._writeData( contact, "%s/contact.csv" %(self._outputDir) )
        
        project = self.getProjectDataAsString()
        self._writeData( project, "%s/project.csv" %(self._outputDir) )
        
        reference = self.getReferenceDataAsString()
        self._writeData(reference, "%s/reference.csv" %(self._outputDir) )
        
        analysis = self.getAnalysisDataAsString()
        self._writeData(analysis, "%s/analysis.csv" %(self._outputDir) )
        
        software = self.getSoftwareDataAsString()
        self._writeData(software, "%s/software.csv" %(self._outputDir) )
        
        shutil.copy( self._configFile, "%s/" %(self._outputDir) )
        
        consensus = self.getDataFromFastaFile()
        self._writeData(consensus, "%s/TE_consensus.csv" %(self._outputDir) )

        self.end()
        
        
if __name__ == "__main__":
    i = TEdenovoToGnpSeq()
    i.setAttributesFromCmdLine()
    i.run()
