#!/usr/bin/env python

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import os
import time
import subprocess
from commons.core.LoggerFactory import LoggerFactory
from commons.core.checker.CheckerUtils import CheckerUtils
from commons.core.utils.FileUtils import FileUtils
from commons.core.utils.RepetOptionParser import RepetOptionParser

LOG_DEPTH = "commons.launcher"

##LaunchPindel
#
class LaunchPindel(object):
    
    def __init__(self, fastaFileName = "", outFileName = "", bamFileName = "", insertSize = 100, nbThreads = 1, referenceDate = "", doClean = False, verbosity = 3):
        self.fastaFileName = fastaFileName
        self.setOutFileName(outFileName)
        self.bamFileName = bamFileName
        self.insertSize = insertSize
        self.nbThreads = nbThreads
        self.setReferenceDate(referenceDate)
        self.setPindelConfigFile()
        self.doClean = doClean
        self.verbosity = verbosity
        self._log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, self.__class__.__name__), self.verbosity)
        
    def setAttributesFromCmdLine(self):
        usage = "LaunchPindel.py [options]"
        description = "LaunchPindel launches Pindel variant caller, and creates a VCF file containing retrieved insertions and deletions.\n"
        epilog = "Example 1: launch without verbosity and keep temporary files.\n"
        epilog += "\t$ LaunchPindel.py -i file.fa -b file.bam -s 100 -v 0\n"
        parser = RepetOptionParser(description = description, epilog = epilog, usage = usage)
        parser.add_option("-i", "--fasta",         dest = "fastaFileName", action = "store",       type = "string", help = "reference genome file name [compulsory] [format: fasta]", default = "")
        parser.add_option("-b", "--bamFileName",   dest = "bamFileName",   action = "store",       type = "string", help = "BAM file name (paired-end) [compulsory]", default = "")
        parser.add_option("-s", "--insertSize",    dest = "insertSize",    action = "store",       type = "int",    help = "insert size: the length of sequence between the paired-end adapters in paired-end sequence", default = 100)
        parser.add_option("-o", "--out",           dest = "outFileName",   action = "store",       type = "string", help = "output file name [default: <input_ref_genome>_D_LI_SI.vcf]", default = "")
        parser.add_option("-t", "--threads",       dest = "nbThreads",     action = "store",       type = "int",    help = "number of threads to use [optional] [default: 1]", default = 1)
        parser.add_option("-d", "--referenceDate", dest = "referenceDate", action = "store",       type = "string", help = "The date of the version of the reference genome [default: current year]", default = "")
        parser.add_option("-c", "--clean",         dest = "doClean",       action = "store_true",                   help = "clean temporary files [optional] [default: False]", default = False)
        parser.add_option("-v", "--verbosity",     dest = "verbosity",     action = "store",       type = "int",    help = "verbosity [optional] [default: 3]", default = 3)
        options = parser.parse_args()[0]
        self._setAttributesFromOptions(options)
        
    def _setAttributesFromOptions(self, options):
        self.fastaFileName = options.fastaFileName
        self.bamFileName = options.bamFileName
        self.insertSize = options.insertSize
        self.setOutFileName(options.outFileName)
        self.setReferenceDate(options.referenceDate)
        self.nbThreads = options.nbThreads
        self.setPindelConfigFile()
        self.doClean = options.doClean
        self.verbosity = options.verbosity
           
    def setOutFileName(self, outFileName):
        if outFileName == "":
            self._outPrefix = os.path.splitext(os.path.basename(self.fastaFileName))[0]
            self._outFileName = "%s_D_LI_SI.vcf" % self._outPrefix
        else:
            self._outFileName = outFileName
            self._outPrefix = os.path.splitext(os.path.basename(self._outFileName))[0]
        
    def setReferenceDate(self, referenceDate):
        if referenceDate == "":
            self._referenceDate = time.strftime("%Y-%m-%d")
        else:
            self._referenceDate = referenceDate
        
    def setPindelConfigFile(self):
        configFile = open('libs.txt','w')
        configFile.write('%s\t%s\t%s' % (os.path.basename(self.bamFileName), self.insertSize, self._outPrefix))
        configFile.close()
        self._pindelConfigFile = "libs.txt"
           
    def _checkOptions(self):
        if self.fastaFileName == "":
            self._logAndRaise("ERROR: Missing input fasta file name")
        else:
            if FileUtils.isRessourceExists(self.fastaFileName):
                self.fastaFileName = os.path.abspath(self.fastaFileName)
            else:
                self._logAndRaise("Input fasta file '%s' does not exist!" % self.fastaFileName)
        
        if self.bamFileName == "":
            self._logAndRaise("ERROR: Missing input BAM file name")
        else:
            if FileUtils.isRessourceExists(self.bamFileName):
                self.bamFileName = os.path.abspath(self.bamFileName)
            else:
                self._logAndRaise("Input BAM file '%s' does not exist!" % self.bamFileName)
                
        if self.insertSize < 1:
            self._logAndRaise("Insert size must be a strictly positive integer!" % self.bamFileName)
            
    def _logAndRaise(self, errorMsg):
        self._log.error(errorMsg)
        raise Exception(errorMsg)
    
    def getPindelCmd(self):
        cmd = "pindel " 
        cmd += "-f %s " % self.fastaFileName
        cmd += "-c ALL "
        cmd += "-i %s " % self._pindelConfigFile
        cmd += "-o %s " % self._outPrefix
        cmd += "-T %s " % self.nbThreads        
        cmd += "-x 5 -r false -t false"
        return cmd

    def getPindel2VCFCmd(self, outPutPindelFile):
        cmd = "pindel2vcf " 
        cmd += "-p %s " % outPutPindelFile
        cmd += "-r %s " % self.fastaFileName
        cmd += "-R %s " % os.path.splitext(self.fastaFileName)[0]
        cmd += "-d %s " % self._referenceDate
        cmd += "-v %s" % self._outFileName
        return cmd

    def generateIndexFiles(self):
        process = subprocess.Popen("samtools index %s" % self.bamFileName, shell = True)
        output = process.communicate()
        self._log.debug("Output:\n%s" % output[0])
        if process.returncode != 0:
            self._logAndRaise("ERROR '%s' indexing" % self.bamFileName)
            
        process = subprocess.Popen("samtools faidx %s" % self.fastaFileName, shell = True)
        output = process.communicate()
        self._log.debug("Output:\n%s" % output[0])
        if process.returncode != 0:
            self._logAndRaise("ERROR '%s' indexing" % self.fastaFileName)
    
    def removeTemporaryFiles(self):
        lFilesToRemove = ["%s_%s" % (self._outPrefix, suffix) for suffix in ["BP", "CloseEndMapped", "D", "LI", "SI", "TD", "INV", "INT_final", "RP"]]
        lFilesToRemove.append(self._pindelConfigFile)
        for f in lFilesToRemove:
            try:
                os.remove(f)
            except:
                pass

    def run(self):
        LoggerFactory.setLevel(self._log, self.verbosity)
        self._log.info("START LaunchPindel")
        
        if not CheckerUtils.isExecutableInUserPath("pindel"):
            self._logAndRaise("ERROR: 'pindel' must be in your path")
        if not CheckerUtils.isExecutableInUserPath("pindel2vcf"):
            self._logAndRaise("ERROR: 'pindel2vcf' must be in your path")
        if not CheckerUtils.isExecutableInUserPath("samtools"):
            self._logAndRaise("ERROR: 'samtools' must be in your path")
                              
        self._checkOptions()
        
        self._log.debug("Fasta file name: %s" % self.fastaFileName)
        self._log.debug("BAM file name: %s" % self.bamFileName)
        self._log.debug("Insert size: %d" % self.insertSize)
        self._log.debug("OutputFile: %s" % self._outFileName)
        
        self._log.debug("Generating indexes from Fasta and BAM files")
        self.generateIndexFiles()
        
        cmd = self.getPindelCmd()
        self._log.debug("Running Pindel with the following command: %s" % cmd)
        process = subprocess.Popen(cmd, shell = True)
        output = process.communicate()
        self._log.debug("Output:\n%s" % output[0])
        if process.returncode != 0:
            self._logAndRaise("ERROR when launching '%s'" % cmd)

        self._log.debug("Concatenating Pindel D, LI and SI output files")
        outPutPindelFileName = "%s_D_LI_SI" % self._outPrefix
        lPindelFiles = ["%s_D" % self._outPrefix, "%s_LI" % self._outPrefix, "%s_SI" % self._outPrefix]
        FileUtils.catFilesFromList(lPindelFiles, outPutPindelFileName, False, False)        

        self._log.debug("Converting concatenated Pindel output files to VCF format")
        cmd = self.getPindel2VCFCmd(outPutPindelFileName)
        self._log.debug("Running pindel2vcf with the following command: %s" % cmd)
        process = subprocess.Popen(cmd, shell = True)
        output = process.communicate()
        self._log.debug("Output:\n%s" % output[0])
        if process.returncode != 0:
            self._logAndRaise("ERROR when launching '%s'" % cmd)        

        if self.doClean:
            self._log.warning("Temporary files will be cleaned")
            self.removeTemporaryFiles()
        self._log.info("END LaunchPindel")

if __name__ == "__main__":
    iLaunchPindel = LaunchPindel()
    iLaunchPindel.setAttributesFromCmdLine()
    iLaunchPindel.run()        