#!/usr/bin/env python

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import os
import sys
import glob
import time
import shutil

if os.getenv("HOSTNAME") == "galaxy":
    try:
        sys.path.insert(0, "/home/galaxy-dev/tools/maphits2")
        import galaxyEnv
    except:
        pass

from commons.core.LoggerFactory             import LoggerFactory
from commons.core.utils.FileUtils           import FileUtils
from commons.core.utils.RepetOptionParser   import RepetOptionParser
from commons.core.checker.CheckerUtils      import CheckerUtils
from commons.core.checker.RepetException    import RepetException
from commons.core.stat.Stat import Stat
from commons.core.seq.BioseqDB import BioseqDB
from commons.core.launcher.Launcher import Launcher
from commons.core.sql.DbFactory import DbFactory
from commons.core.sql.TableJobAdaptatorFactory import TableJobAdaptatorFactory

LOG_DEPTH = "commons.tools"

## Estimates K-mer size best value for a guided assembly
#
class BenchVelvetKValue(object):
    
    def __init__(self, BAMFileName = "", refFileName = "", doClean = False, verbosity = 3):
        self._BAMFileName = BAMFileName
        self._refFileName = refFileName
        self._outFileName = "K.txt"
        self._genomeFractionFileName = "genomeFraction.fa"
        self._mergedBam = "merged.bam"
        self._jobsResources = ""
        self.setNbThreads(1)
        self._resources = ""
        self._tmpDir = ""
        self._doSecPass = False
        self._doClean = doClean
        self._verbosity = verbosity
        self._log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, self.__class__.__name__), self._verbosity)
        
    def setAttributesFromCmdLine(self):
        usage = "BenchVelvetKValue.py [options]"
        description = "Estimates K-mer size best value for a guided assembly.\n"
        epilog = "\nExample 1: launch a single pass, use 12 threads and clean temporary files.\n"
        epilog += "\t$ BenchVelvetKValue.py -i MyAnalysis.bam -r MyRefGenome.fa -t 12 -l 'pe multithread 12' -c -v 3"
        epilog += "\n\t"
        epilog += "\nExample 2: launch two passes, use a single thread and 100G RAM, keep temporary files and ask for no output.\n"
        epilog += "\t$ BenchVelvetKValue.py -i MyAnalysis.bam -r MyRefGenome.fa -l 'mem_free=100G' -s -v 0\n"
        epilog += "\n"
        epilog += "IMPORTANT NOTE ABOUT THREADS: to use several threads, Velvet must have been compiled with the 'OPENMP=1' flag.\n"
        epilog += "For an unknown reason, when multithreading is activated and only 1 thread is required, Velvet seems to run indefinitely.\n"
        epilog += "If you are in this case, you have to use at least 2 threads.\n"
        parser = RepetOptionParser(description = description, epilog = epilog, usage = usage)
        parser.add_option("-i", "--bam",        dest = "inputBAM",   action = "store",       type = "string", help = "Input BAM file name [compulsory] [format: BAM]", default = "")
        parser.add_option("-r", "--ref",        dest = "inputRef",   action = "store",       type = "string", help = "Input reference genome file name [compulsory] [format: fasta]", default = "")
        parser.add_option("-o", "--out",        dest = "outFile",    action = "store",       type = "string", help = "Output file name containing best K value found [optional] [default: K.txt]", default = "K.txt")
        parser.add_option("-t", "--thread",     dest = "nbThreads",  action = "store",       type = "int",    help = "Number of threads to use [optional] [default: 1]", default = 1)
        parser.add_option("-l", "--resources",  dest = "resources",  action = "store",       type = "string", help = "Resources to use in jobs. WARNING: ask slot number according to '-t' option [format: 'queue.q pe parallelEnv otherResources'][example: 'all.q pe multithread 24 mem_free=100G'] [default: None]", default = "")
        parser.add_option("-s", "--second",     dest = "doSecPass",  action = "store_true",                   help = "Refine K estimation with a second pass [optional] [default: False]", default = False)
        parser.add_option("-c", "--clean",      dest = "doClean",    action = "store_true",                   help = "Clean temporary files [optional] [default: False]", default = False)
        parser.add_option("-v", "--verbosity",  dest = "verbosity",  action = "store",       type = "int",    help = "Verbosity [optional] [default: 3]", default = 3)
        options = parser.parse_args()[0]
        self._setAttributesFromOptions(options)
        
    def _setAttributesFromOptions(self, options):
        self.setInputBAM(options.inputBAM)
        self.setInputRef(options.inputRef)
        self.setOutFileName(options.outFile)
        self.setNbThreads(options.nbThreads)
        self.setResources(options.resources)
        self.setDoSecPass(options.doSecPass)
        self.setDoClean(options.doClean)
        self.setVerbosity(options.verbosity)

    def setInputBAM(self, BAMFileName):
        self._BAMFileName = BAMFileName
        
    def setInputRef(self, refFileName):
        self._refFileName = refFileName

    def setOutFileName(self, outFile):
        self._outFileName = outFile
        
    def setNbThreads(self, nbThreads):
        self._nbThreads = int(nbThreads)
            
    def setResources(self, resources):
        self._jobsResources = resources
            
    def setDoSecPass(self, doSecPass):
        self._doSecPass = doSecPass
            
    def setDoClean(self, doClean):
        self._doClean = doClean
        
    def setVerbosity(self, verbosity):
        self._verbosity = verbosity
        
    def _checkOptions(self):
        if self._BAMFileName == "":
            self._logAndRaise("Missing input BAM")
        else:
            if FileUtils.isRessourceExists(self._BAMFileName):
                self._BAMFileName = os.path.abspath(self._BAMFileName)
            else:
                self._logAndRaise("Input BAM file '%s' does not exist!" % self._BAMFileName)
                
        if self._refFileName == "":
            self._logAndRaise("Missing input reference genome")
        else:
            if FileUtils.isRessourceExists(self._refFileName):
                self._refFileName = os.path.abspath(self._refFileName)
            else:
                self._logAndRaise("Input reference genome file '%s' does not exist!" % self._refFileName)
                
        if self._nbThreads < 1:
            self._logAndRaise("Bad number of threads '%i': must be a strictly positive integer" % self._nbThreads)
         
    def _extractGenomeFraction(self):
        iBSDB = BioseqDB(self._refFileName)
        iBSDB.sortByLength(reverse = True)
        lSizes = iBSDB.getListOfSequencesLength()
        refLength = iBSDB.getLength()
        
        nbSeq = 0
        cumLength = 0
        while cumLength < refLength * 0.1:
            cumLength += lSizes[nbSeq]
            nbSeq += 1
        
        genomeFractionBSDB = iBSDB.bestLength(nbSeq)
        genomeFractionBSDB.write(open(self._genomeFractionFileName, "w"))
        
        return genomeFractionBSDB.getHeaderList()
    
    def _selectReadsFromBAM(self, lHeaders):
        mappedBam = "mapped.bam"
        unmappedBam = "unmapped.bam"
        
        iDb = DbFactory.createInstance()
        jobdb = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
        cDir = os.getcwd()
        groupid = 'SelectReadFromBAM_%s' % time.strftime("%Y%m%d%H%M%S")
        acronym = 'SelectReadFromBAM'
        iLauncher = Launcher(jobdb, cDir, cDir, self._jobsResources, groupid, acronym)
        
        lCmds = []
        lCmds.append(iLauncher.getSystemCommand("samtools sort -@ %d -f %s %s" % (self._nbThreads, self._BAMFileName, self._BAMFileName), []))
        lCmds.append(iLauncher.getSystemCommand("samtools index %s" % self._BAMFileName, []))
        lCmds.append(iLauncher.getSystemCommand("samtools view -@ %d -b %s %s > %s" % (self._nbThreads, self._BAMFileName, " ".join(lHeaders), mappedBam), []))
        #-f 12: both reads of a pair are unmapped
        lCmds.append(iLauncher.getSystemCommand("samtools view -@ %d -f 12 -b %s > %s" % (self._nbThreads, self._BAMFileName, unmappedBam), []))
        lCmds.append(iLauncher.getSystemCommand("samtools sort -@ %d -f %s %s" % (self._nbThreads, mappedBam, mappedBam), []))
        lCmds.append(iLauncher.getSystemCommand("samtools sort -@ %d -f %s %s" % (self._nbThreads, unmappedBam, unmappedBam), []))
        lCmds.append(iLauncher.getSystemCommand("samtools merge -@ %d %s %s %s" % (self._nbThreads, self._mergedBam, mappedBam, unmappedBam), []))
        lCmds.append(iLauncher.getSystemCommand("samtools index %s" % self._mergedBam, []))
        lCmdStart = []
        lCmdFinish = []
        lCmdFinish.append("shutil.move(\"%s\", \"%s/%s\")" % (self._mergedBam, cDir, self._mergedBam))
        lCmdFinish.append("shutil.move(\"%s.bai\", \"%s/%s.bai\")" % (self._mergedBam, cDir, self._mergedBam))
        if not self._doClean:
            lCmdFinish.append("shutil.move(\"%s\", \"%s/%s\")" % (mappedBam, cDir, mappedBam))
            lCmdFinish.append("shutil.move(\"%s\", \"%s/%s\")" % (unmappedBam, cDir, unmappedBam))
        
        lCmdsTuples = [iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart, lCmdFinish)]
        iLauncher.runLauncherForMultipleJobs(acronym, lCmdsTuples, cleanMustBeDone = self._doClean)
        
    def _estimateReadSize(self, header):
        os.system("samtools view %s %s:1-100000 | cut -f 10 | wc -L > maxReadSize.txt" % (self._mergedBam, header))
        with open("maxReadSize.txt") as fH:
            size = int(fH.readlines()[0].strip())
        return size
    
    def _createVelvetHCommand(self, iLauncher, start, end, step):
        advancedOptions = ""
        cmd = "export OMP_THREAD_LIMIT=%s; export OMP_NUM_THREADS=%s; velveth " % (self._nbThreads, self._nbThreads)
        cmd += "%s/%s " % (os.getcwd(), self._assemblyFolderPrefix)
        cmd += "%i,%i,%i " % (start, end, step) 
        cmd += "-reference %s " % self._genomeFractionFileName
        cmd += "-shortPaired "
        cmd += "-bam %s " % self._mergedBam
        cmd += "%s" % advancedOptions
        return iLauncher.getSystemCommand(cmd, [])
    
    def _createVelvetGCommand(self, iLauncher, folder):
        advancedOptions = ""
        cmd = "export OMP_THREAD_LIMIT=%s; export OMP_NUM_THREADS=%s; velvetg " % (self._nbThreads, self._nbThreads)
        cmd += "%s " % folder
        cmd += "-ins_length auto "
        cmd += "-scaffolding no "
        cmd += "-min_contig_lgth 100 "
        cmd += "-exp_cov auto "
        cmd += "%s" % advancedOptions
        return iLauncher.getSystemCommand(cmd, [])
        
    def _indexGenome(self, start = 21, end = 101, step = 10):
        iDb = DbFactory.createInstance()
        jobdb = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
        cDir = os.getcwd()
        groupid = 'Velveth_%s' % time.strftime("%Y%m%d%H%M%S")
        acronym = 'Velveth'
        iLauncher = Launcher(jobdb, cDir, cDir, self._jobsResources, groupid, acronym)
        
        lCmds = []
        lCmds.append(self._createVelvetHCommand(iLauncher, start, end, step))
        lCmdStart = []
        lCmdStart.append("os.symlink('%s/%s', '%s')" % (cDir, self._genomeFractionFileName, os.path.basename(self._genomeFractionFileName)))
        lCmdStart.append("os.symlink('%s/%s', '%s')" % (cDir, self._mergedBam, self._mergedBam))
        lCmdFinish = []
        lCmdsTuples = [iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart, lCmdFinish)]
        iLauncher.runLauncherForMultipleJobs(acronym, lCmdsTuples, cleanMustBeDone = self._doClean)
        
    def _generateAssembly(self):
        iDb = DbFactory.createInstance()
        jobdb = TableJobAdaptatorFactory.createInstance(iDb, "jobs")
        cDir = os.getcwd()
        groupid = 'Velvetg_%s' % time.strftime("%Y%m%d%H%M%S")
        acronym = 'Velvetg'
        iLauncher = Launcher(jobdb, cDir, cDir, self._jobsResources, groupid, acronym)
        
        lCmdsTuples = []
        for folder in glob.glob("%s_*" % self._assemblyFolderPrefix):
            lCmds = []
            lCmds.append(self._createVelvetGCommand(iLauncher, os.path.abspath(folder)))
            lCmdStart = []
            lCmdFinish = []
            lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart, lCmdFinish))
        iLauncher.runLauncherForMultipleJobs(acronym, lCmdsTuples, cleanMustBeDone = self._doClean)
        
    def _getBestKValue(self):
        bestK = 0
        dKToN50 = {}
        for assemblyFile in glob.glob("%s_*/contigs.fa" % self._assemblyFolderPrefix):
            iBSDB = BioseqDB(assemblyFile)
            iStat = Stat(iBSDB.getListOfSequencesLength())
            N50 = iStat.N50()
            k = int(assemblyFile.split("/")[0].split("_")[-1])
            dKToN50[k] = N50
            
        maxN50 = max(dKToN50.values())
        for k, N50 in dKToN50.items():
            if N50 == maxN50:
                bestK = max(k, bestK)
        
        return bestK
    
    def _logAndRaise(self, errorMsg):
        self._log.error(errorMsg)
        raise RepetException(errorMsg)
                    
    def run(self):
        LoggerFactory.setLevel(self._log, self._verbosity)
        self._log.info("START BenchVelvetKValue")
        if not CheckerUtils.isExecutableInUserPath("samtools"):
            self._logAndRaise("ERROR: 'samtools' must be in your path")
        if not CheckerUtils.isExecutableInUserPath("velveth"):
            self._logAndRaise("ERROR: 'velveth' must be in your path")
        if not CheckerUtils.isExecutableInUserPath("velvetg"):
            self._logAndRaise("ERROR: 'velvetg' must be in your path")
        self._checkOptions()
        self._log.debug("BAM file name: %s" % self._BAMFileName)
        self._log.debug("Reference genome file name: %s" % self._refFileName)
        
        originalDir = os.getcwd()
        workingDir = os.path.join(os.getcwd(), "BenchVelvetKValue_%s" % time.strftime("%Y%m%d%H%M%S"))
        os.mkdir(workingDir)
        os.chdir(workingDir)
        
        self._log.debug("Extracting sample from reference genome")
        lHeaders = self._extractGenomeFraction()
        
        self._log.debug("Extracting sample from BAM file")
        self._selectReadsFromBAM(lHeaders)

        readSize = self._estimateReadSize(lHeaders[0])
        if readSize == 0:
            self._logAndRaise("ERROR: can't estimate read size, '0' found")
        self._log.debug("Observed read size: %d" % readSize)
        
        self._log.debug("Running Velveth Pass 1: k from %d to %d step %d" % (21, min(readSize, 101), 10))
        self._assemblyFolderPrefix = "velvetPass1_k"
        self._indexGenome(start = 21, end = min(readSize, 101), step = 10)

        self._log.debug("Running Velvetg Pass 1")
        self._generateAssembly()
        
        bestK = self._getBestKValue()
        self._log.info("Best K value for pass 1: %d" % bestK)
        
        if self._doSecPass:
            self._log.debug("Running Velveth Pass 2: k from %d to %d step %d" % (max(21, bestK - 8), min(readSize, 101, bestK + 10), 2))
            self._assemblyFolderPrefix = "velvetPass2_k"
            self._indexGenome(start = max(21, bestK - 8), end = min(readSize, 101, bestK + 10), step = 2)
            
            self._log.debug("Running Velvetg Pass 2")
            self._generateAssembly()
            
            bestK = self._getBestKValue()
            self._log.info("Best K value for pass 2: %d" % bestK)

        os.chdir(originalDir)

        with open(self._outFileName, "w") as fH:
            fH.write("K-value maximizing assembly quality: %d" % bestK)
            
        self._log.info("K-value maximizing assembly quality: %d" % bestK)
        
        if self._doClean:
            self._log.warning("Temporary files will be cleaned")
            shutil.rmtree(workingDir)
        
        self._log.info("END BenchVelvetKValue")
        
        return bestK

if __name__ == "__main__":
    iBenchVelvetKValue = BenchVelvetKValue()
    iBenchVelvetKValue.setAttributesFromCmdLine()
    iBenchVelvetKValue.run()      