#!/usr/bin/env python

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import os
import re
import glob
import time
import itertools
from commons.core.LoggerFactory import LoggerFactory
from commons.core.utils.RepetOptionParser import RepetOptionParser
from commons.core.utils.ClassifUtils import ClassifUtils
from commons.core.checker.ConfigChecker import ConfigChecker, ConfigRules
from commons.core.launcher.Launcher import Launcher
from commons.core.sql.DbFactory import DbFactory
from commons.core.sql.TableJobAdaptatorFactory import TableJobAdaptatorFactory
from commons.core.sql.TableSeqAdaptator import TableSeqAdaptator

LOG_DEPTH = "repet.commons.tools"

class CreateGFF3sForClassifFeatures(object):
    
    def __init__(self, configFileName = "", fastaFilePath = "", doClean = True, verbosity = 1):
        self.setConfigFileName(configFileName)
        self.setFastaFilePath(fastaFilePath)
        self.setDoClean(doClean)
        self.setVerbosity(verbosity)
        
        self._repetEnvSectionName = "repet_env"
        self._projectSectionName = "project"
        self._gff3DenovoOptionsSectionName = "gff3_TEdenovo_options"
        self._gff3AnnotOptionsSectionName = "gff3_TEannot_options"
        self._otherSectionName = "other"
        self._iDb = None
        self._cgcfDir = "Visualization_Files"
        
        # dFeatToInfos store all informations linked to one of the possible features type.
        # Here, pre-known informations are added : Data format, color for the track, color for features used in the classification
        ## Multicolor palette : Light / Flashy
        self._dFeatToInfos = {"polyA": {"format": "set", "color": "169,173,228", "usedColor": "#111764"},
                         "rDNA_BLRn": {"format": "path", "color": "156,198,210", "usedColor": "#015d6b"},
                         "tRNA": {"format": "map", "color": "150,203,252", "usedColor": "48,116,161"},
                         "HG_BLRn": {"format": "path", "color": "161,210,197", "usedColor": "#017451"},
                         "ORF": {"format": "map", "color": "164,233,211", "usedColor": "#02aa72"},
                         "Profiles": {"format": "path", "color": "225,243,195", "usedColor": "#89c926"},
                         "SSR": {"format": "set", "color": "247,231,186", "usedColor": "#f9ac0f"},
                         # SSRtrf for backward compatibility with classif files generated with REPET pre-2.4
                         "SSRtrf": {"format": "set", "color": "247,231,186", "usedColor": "#f9ac0f"},
                         "TE_BLRn": {"format": "path", "color": "252,193,176", "usedColor": "#f65612"},
                         "TE_BLRtx": {"format": "path", "color": "254,173,193", "usedColor": "#d2023a"},
                         "TE_BLRx": {"format": "path", "color": "241,161,196", "usedColor": "#a41855"},
                         "TR": {"format": "set", "color": "228,174,222", "usedColor": "#710162"},
                         "annotated_copies": {"format": "path", "color": "0,0,0"}}
        self._colorOther = "#8c8c8c" # Gray
        # Then, other informations will be added during the run for selected feature types ("yes" in the config file) :
        # name of the table, name of the text file sent to GFF3Maker, name of the output gff3 file
        
        self._log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, self.__class__.__name__), self._verbosity)
    
    def setConfigFileName(self, configFileName):
        self._configFileName = configFileName
        
    def setFastaFilePath(self, fastaFilePath):
        self._fastaFilePath = fastaFilePath
        self._fastaFileName = os.path.basename(fastaFilePath)
    
    def setDoClean(self, doClean):
        self._doClean = doClean
        
    def setVerbosity(self, verbosity):
        self._verbosity = verbosity

    def _setAttributesFromCmdLine(self):
        usage = "CreateGFF3sForClassifFeatures.py [options]"
        description = "Create gff3 files to load in a genome browser for the visualisation of features on consensus.\n"
        description += "See the README for more informations.\n"
        epilog = "\n"
        parser = RepetOptionParser(description = description, epilog = epilog, usage = usage)
        parser.add_option("-C", "--config",    dest = "configFileName", action = "store", type = "string", help = "Config file name [compulsory]", default = "")
        parser.add_option("-f", "--fasta",     dest = "fastaFilePath",  action = "store", type = "string", help = "path to consensus file (format='fasta') [compulsory]", default = "")
        parser.add_option("-v", "--verbosity", dest = "verbosity",      action = "store", type = "int",    help = "verbosity [default: 1]", default = 1)
#        parser.add_option("", "--noClean", dest = "doClean",      action = "store_false",    help = "disable cleaning of temporary files [warning: debug option]", default = True)
        options = parser.parse_args()[0]
        self._setAttributesFromOptions(options)
    
    def _setAttributesFromOptions(self, options):
        self.setConfigFileName(options.configFileName)
        self.setFastaFilePath(options.fastaFilePath)
#        self.setDoClean(options.doClean)
        self.setVerbosity(options.verbosity)
        
    def _checkOptions(self):
        if self._configFileName == "":
            self._logAndRaise("Missing config file")
        elif not os.path.exists(self._configFileName):
            self._logAndRaise("Config file does not exist : %s" % self._configFileName)
        
        if self._fastaFileName == "":
            self._logAndRaise("Missing input fasta file")
        elif not os.path.exists(self._fastaFilePath):
            self._logAndRaise("Fasta file does not exist : %s" % self._fastaFilePath)
        
    def _logAndRaise(self, errorMsg):
        self._log.error(errorMsg)
        raise SystemExit(1)
    
    def _checkConfig(self):       
        iConfigRules = ConfigRules()
        iConfigRules.addRuleOption(section=self._repetEnvSectionName, option ="repet_host", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._repetEnvSectionName, option ="repet_user", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._repetEnvSectionName, option ="repet_pw", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._repetEnvSectionName, option ="repet_db", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._repetEnvSectionName, option ="repet_port", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._projectSectionName, option ="project_name", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._projectSectionName, option ="project_dir", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._gff3DenovoOptionsSectionName, option ="add_classif_infos", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._gff3DenovoOptionsSectionName, option ="TR", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._gff3DenovoOptionsSectionName, option ="polyA", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._gff3DenovoOptionsSectionName, option ="ORF", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._gff3DenovoOptionsSectionName, option ="TE_BLRn", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._gff3DenovoOptionsSectionName, option ="TE_BLRtx", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._gff3DenovoOptionsSectionName, option ="TE_BLRx", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._gff3DenovoOptionsSectionName, option ="HG_BLRn", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._gff3DenovoOptionsSectionName, option ="rDNA_BLRn", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._gff3DenovoOptionsSectionName, option ="tRNA", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._gff3DenovoOptionsSectionName, option ="Profiles", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._gff3DenovoOptionsSectionName, option ="SSR", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._gff3AnnotOptionsSectionName, option ="project_name_teannot", mandatory=False, type="string")
        iConfigRules.addRuleOption(section=self._gff3AnnotOptionsSectionName, option ="annotated_copies", mandatory=True, type="string")
        iConfigRules.addRuleOption(section=self._otherSectionName, option ="original_HSP", mandatory=True, type="string")

        iConfigChecker = ConfigChecker(self._configFileName, iConfigRules)
        self._iConfig = iConfigChecker.getConfig()
        self._setAttributesFromConfig()
        
    def _setAttributesFromConfig(self):
        self._repetHost = self._iConfig.get(self._repetEnvSectionName, "repet_host")
        self._repetUser = self._iConfig.get(self._repetEnvSectionName, "repet_user")
        self._repetPw = self._iConfig.get(self._repetEnvSectionName, "repet_pw")
        self._repetDb = self._iConfig.get(self._repetEnvSectionName, "repet_db")
        self._repetPort = self._iConfig.get(self._repetEnvSectionName, "repet_port")
        self._projectName = self._iConfig.get(self._projectSectionName, "project_name")
        self._projectDir = self._iConfig.get(self._projectSectionName, "project_dir")
        self._addClassifInfos = self._iConfig.get(self._gff3DenovoOptionsSectionName, "add_classif_infos") == "yes"
        self._lFeatureType = [featType for featType in self._dFeatToInfos.keys() 
                           if self._iConfig.get(self._gff3DenovoOptionsSectionName, featType) == "yes"]
        if self._iConfig.get(self._gff3AnnotOptionsSectionName, "annotated_copies") == "yes":
            self._lFeatureType.append("annotated_copies")
            if self._iConfig.get(self._gff3AnnotOptionsSectionName, "project_name_teannot") == "<your_project_name>":
                msg = "Please fill the project name of your TEannot if you want to add annotated copies"
                self._logAndRaise(msg)
            else:
                self._TEannotProjectName = self._iConfig.get(self._gff3AnnotOptionsSectionName, "project_name_teannot")
        self._addOriginalCopies = self._iConfig.get(self._otherSectionName, "original_HSP") == "yes"
    
    ## Create the seq table for the consensus library and the mapping dictionary between wicker and pre-wicker headers
    #
    def _createConsensusTableAndHeadersMapping(self):
        self._log.info("Creating consensus table...")
        self._seqTable = "%s_ConsensusLib_seq" % self._projectName
        self._iDb.createTable(self._seqTable, "seq", self._fastaFileName)
        
        iTSa = TableSeqAdaptator(self._iDb, self._seqTable)
        lWickerHeaders = iTSa.getAccessionsList()
        lLongHeaders = [self._transformHeader(header) for header in lWickerHeaders]
        self._dHeadersMap = dict(itertools.izip(lLongHeaders, lWickerHeaders))

    ## Converts a Wicker-tagged header (TEdenovo step 5+ outputs) to its previous form (step 4 output)
    #
    # @param header Header in "Wicker" format
    # @return Header in "Long name" format
    #
    def _transformHeader(self, header):
        if len(header.split("_")) < 2 or len(header.split("-")) < 4:
            msg = "Unknown header format : %s. Headers should be in TEdenovo's Wicker format (fasta \
output of step 6 or 7): [WickerTag]_[ProjectName]-[B|L]-[G|R|P|B|M + number]-[Map + number]" % header
            self._logAndRaise(msg)
        isReversed = header.endswith("_reversed")
        if not isReversed:
            coreHeader = "_".join(header.split("_")[1:])
        else:
            coreHeader = "_".join(header.split("_")[1:-1])
        splitCore = coreHeader.split("-")
        mapInfo = splitCore.pop()
        mapInfo, mapNumber = mapInfo[0:3], mapInfo[3:]
        clusterInfo = splitCore.pop()
        clusterInfo, clusterNumber = clusterInfo[0], clusterInfo[1:]
        selfAlignInfo = splitCore.pop()
        projectInfo = "-".join(splitCore)
        
        if selfAlignInfo == "B":
            selfAlignInfo = "Blaster"
        elif selfAlignInfo == "L":
            selfAlignInfo = "LTRharvest"
        else:
            msg = "Unknown letter %s in header %s : must be B(laster) or L(TRharvest)" % (selfAlignInfo, header)
            self._log.error(msg)
        
        if clusterInfo == "G":
            clusterInfo = "Grouper"
        elif clusterInfo == "R":
            clusterInfo = "Recon"
        elif clusterInfo == "P":
            clusterInfo = "Piler"
        elif clusterInfo == "B":
            clusterInfo = "Blastclust"
        elif clusterInfo == "M":
            clusterInfo = "MCL"
        else:
            msg = "Unknown letter %s in header %s : must be G(rouper) or R(econ) or P(iler) or B(lastclust) or M(CL)" % (clusterInfo, header)
            self._log.error(msg)
        
        newHeader = "_".join((projectInfo, selfAlignInfo, clusterInfo, clusterNumber, mapInfo, mapNumber))
        return newHeader
        
    ## Create temporary tables containing only the features for the consensus in the fasta file.
    ## Also modify the headers in the tables and create the txt files to be used by GFF3Maker.
    #
    def _createFeatureTablesAndInfoFiles(self):
        self._log.info("Creating features tables...")
        for featType in self._lFeatureType:
            tableFormat = self._dFeatToInfos[featType]["format"]
            tableName = self._dFeatToInfos[featType]["table"]
            newTableName = "%s_tmpForCGCF" % tableName
            if featType == "annotated_copies":
                self._copyAnnotatedCopiesTable(tableName, newTableName)
            else:
                self._copyFeatureTablesWithEditedHeaders(tableName, newTableName, tableFormat)
                if tableFormat == "map":
                    tableFormat = "set"
                    oldNewTableName = newTableName
                    newTableName = "%s_%s_%s_tmpForCGCF" % (self._projectName, featType, tableFormat)
                    self._iDb.convertMapTableIntoSetTable(oldNewTableName, newTableName)
            infosTablesFileName = "InfosTables_%s.txt" % featType
            self._dFeatToInfos[featType]["info_file"] = infosTablesFileName
            pjName = self._projectName if featType != "annotated_copies" else self._TEannotProjectName
            string = "%s_%s\t%s\t%s\n" % (pjName, featType, tableFormat, newTableName)
            with open(infosTablesFileName, "w") as f:
                f.write(string)

    def _copyFeatureTablesWithEditedHeaders(self, tableName, newTableName, tableFormat):
        dFormatToField = {"set": "chr",
                      "map": "chr",
                      "path": "query_name"}
        fieldName = dFormatToField[tableFormat]
        sqlParams = "where %s in ('%s')" % (fieldName, "','".join(self._dHeadersMap.keys()))
        self._iDb.copyTable(tableName, newTableName, sqlParams)
        sqlCmd = "select %s from %s" % (fieldName, newTableName)
        headersToChange = self._iDb.getStringListWithSQLCmd(sqlCmd)
        for header in headersToChange:
            sqlCmd = "update %s set %s='%s' where %s='%s'" \
                % (newTableName, fieldName, self._dHeadersMap[header], fieldName, header)
            self._iDb.execute(sqlCmd)
    
    def _copyAnnotatedCopiesTable(self, tableName, newTableName):
        sqlCmd = "create table %s like %s" % (newTableName, tableName)
        self._iDb.execute(sqlCmd)
        sqlCmd = "insert into %s select path, subject_name, subject_start, subject_end, \
query_name, query_start, query_end, E_value, score, identity from %s \
where subject_name in ('%s') and subject_start <= subject_end" \
% (newTableName, tableName, "','".join(self._dHeadersMap.values()))
        self._iDb.execute(sqlCmd)
        sqlCmd = "insert into %s select path, subject_name, subject_end, subject_start, \
query_name, query_start, query_end, E_value, score, identity from %s \
where subject_name in ('%s') and subject_end < subject_start" \
% (newTableName, tableName, "','".join(self._dHeadersMap.values()))
        self._iDb.execute(sqlCmd)

    ## Create the GFF3 files using GFF3Maker
    #
    def _createGFF3s(self):
        self._log.info("Creating GFF3 files...")
        cDir = os.getcwd()
        tmpDir = cDir
        groupid = "%s_CGCFgff3" % (self._projectName)
        acronym = "GFF3Maker"
        jobdb = TableJobAdaptatorFactory.createInstance(self._iDb, "jobs")
        iLauncher = Launcher(jobdb, cDir, tmpDir, "", groupid, acronym)
        lCmdsTuples = []
        for featType in self._lFeatureType:
            infoFile = self._dFeatToInfos[featType]["info_file"]
            lCmdStart = []
            lCmdStart.append("os.symlink(\"%s/%s\", \"%s\")" % (cDir, infoFile, infoFile))
            lCmds = []
            lCmds.append(self._createGFF3MakerCommand(infoFile, iLauncher))
            lCmdFinish = []
            pjName = self._projectName if featType != "annotated_copies" else self._TEannotProjectName
            outGFF3 = "%s_%s.gff3" % (pjName, featType)
            lCmdFinish.append("os.system(\"cat *.gff3 > %s\")" % outGFF3)
            lCmdFinish.append("os.system(\"mv %s %s/.\")" % (outGFF3, cDir))
            lCmdsTuples.append(iLauncher.prepareCommands_withoutIndentation(lCmds, lCmdStart, lCmdFinish))
            self._dFeatToInfos[featType]["gff3"] = outGFF3
        iLauncher.runLauncherForMultipleJobs(acronym, lCmdsTuples, self._doClean)
        # Reopen iDb because runLauncherForMultipleJobs closes it
        self._iDb = DbFactory.createInstance()
    
    ## Create soft links to the fa_aln files in the _Map folders
    #
    def _createOriginalCopiesLinks(self):
        cgcfDirPath = os.getcwd()
        dirName = "Original_HSP_fastaAlignment"
        if not os.path.exists(dirName):
            os.mkdir(dirName)
        os.chdir(dirName)
        errorList = []
        for longHeader, wickerHeader in self._dHeadersMap.iteritems():
            clusterTool, clusterNumber = self._getLongHeaderClusterInfos(longHeader)
            alnDir = "%s_%s_Map" % (self._projectName, clusterTool)
            if not os.path.exists("%s/%s" % (self._projectDir, alnDir)):
                if alnDir not in errorList:
                    errorList.append(alnDir)
            else:
                alnFileName = "seqCluster%s.fa.fa_aln" % clusterNumber
                os.symlink("%s/%s/%s" % (self._projectDir, alnDir, alnFileName), "%s.fa_aln" % wickerHeader)
        os.chdir(cgcfDirPath)
        if errorList:
            msg = "Couldn't find directory : %s" % "\n".join(errorList)
            self._log.error(msg)
    
    ## Get the infos for a "Long Name" header (clustering tool, number)
    #
    # @param longHeader Header in pre-Wicker format
    # @return Self-align_Clustering tools and cluster number
    #
    def _getLongHeaderClusterInfos(self, longHeader):
        coreInfos = longHeader[len(self._projectName) + 1:].rsplit("_Map", 1)[0]
        clusterTool, clusterNumber = coreInfos.rsplit("_", 1)
        return clusterTool, clusterNumber
        
    def _createGFF3MakerCommand(self, infoFile, iLauncher):
        lArgs = []
        lArgs.append("-t %s" % infoFile)
        lArgs.append("-f %s" % self._seqTable)
        # The -p option makes heavier gff3 (all lines are duplicated) but allows 
        # "squished" representation in IGV. Squished rules !
        lArgs.append("-p")
        lArgs.append("-v %i" % (self._verbosity - 1))
        return iLauncher.getSystemCommand("GFF3Maker.py", lArgs)
    
    ## Edit GFF3 files to format some fields and add color tags for used features.
    #
    def _editGFF3s(self):
        self._log.info("Editing GFF3 files...")
        if self._addClassifInfos:
            # Load classif infos
            dClassifInfo = ClassifUtils.getClassifInfosAsDict(self._classifFileName)
        for featType in self._lFeatureType:
            gff3File = self._dFeatToInfos[featType]["gff3"]
            tmpGff3File = "%s_tmp" % gff3File
            # Copy content of the gff3 in a temporary file, remove residual headers from the earlier
            # concatenation and replace "match" and "match-part" by keywords for genome browsers.
            cmd = "grep -v '##' %s > %s" % (gff3File, tmpGff3File)
            os.system(cmd)
            cmd = "sed -i 's/match\t/%s\t/' %s" % (featType, tmpGff3File)
            os.system(cmd)
            cmd = "sed -i 's/match_part\t/exon\t/' %s" % (tmpGff3File)
            os.system(cmd)
            # Prepare new header for the gff3
            gff3Header = "##gff-version 3\n"
            gff3Header += "#track name=\"%s\" color=%s\n" % (featType, self._dFeatToInfos[featType]["color"])
            with open(gff3File, "w") as wf:
                # Write the header in the new gff3 file
                wf.write(gff3Header)
                # Rewrite content of the temporary gff3 in the new gff3
                with open(tmpGff3File, "r") as rf:
                    # If classif infos aren't considered or if the feature doesn't have classif info, just rewrite
                    if not self._addClassifInfos or featType == "annotated_copies":
                        for line in rf:
                            wf.write(line)
                    else:
#TODO: Remove the "if" dedicated to TR when classif and tables have the same values for the length
# Explanation : TRsearch.cpp compute the length (len) of a TermRepeats during the detectFeatures step and write
# this info in the feature's name (e.g. "Rep2LTR|len=861|id=0.997677|lenTE=6349" in the table).
# During PASTEC, TRSearch_Agent.py compute another value for the length, using the mean length of the 2 parts,
# and write it in the classif file (e.g. "TermRepeats: non-termLTR: 3005").
# We can't compare these values because they're not computed the same way. This whole "if" here is meant to 
# re-compute the length so they're comparable.
                        if featType == "TR":
                            trLines = rf.readlines()
                            if len(trLines)%3 != 0:
                                msg = "TR file's line counter isn't modulo 3"
                                self._logAndRaise(msg)
                            i = 0
                            while i < len(trLines):
                                trLine = trLines[i]
                                exon1Line = trLines[i+1]
                                exon2Line = trLines[i+2]
                                i += 3
                                if self._dealWithTRLine(trLine, exon1Line, exon2Line, dClassifInfo, "struct"):
                                    trLine = "%s;color=%s\n" % (trLine.strip(), self._dFeatToInfos["TR"]["usedColor"])
                                elif self._dealWithTRLine(trLine, exon1Line, exon2Line, dClassifInfo, "other"):
                                    trLine = "%s;color=%s\n" % (trLine.strip(), self._colorOther)
                                wf.write(trLine)
                                wf.write(exon1Line)
                                wf.write(exon2Line)
                        else:
                            for line in rf:
                                # "exon" lines doesn't need a color tag
                                if "\texon\t" not in line:
                                    # Check if the feature is found in coding, struct or other to add the color tag
                                    if (self._isFeatureUsed(line, "coding", featType, dClassifInfo) or
                                        self._isFeatureUsed(line, "struct", featType, dClassifInfo)):
                                        line = "%s;color=%s\n" % (line.strip(), self._dFeatToInfos[featType]["usedColor"])
                                    elif self._isFeatureUsed(line, "other", featType, dClassifInfo):
                                        line = "%s;color=%s\n" % (line.strip(), self._colorOther)
                                wf.write(line)

            os.remove(tmpGff3File)
    
    #TODO: Remove when possible (see TODO above)
    ## Check whether a TR feature is used in the classification or not
    #
    # @param trLine Line for a TR feature from the gff3
    # @param exon1Line Line for first part of the TR
    # @param exon2Line Line for second part of the TR
    # @param dClassifInfo dictionary with the classification infos
    # @param category Classification category to check ("struct" or "other")
    # @return Boolean
    #
    def _dealWithTRLine(self, trLine, exon1Line, exon2Line, dClassifInfo, category):
        used = False
        exon1Line = exon1Line.split()
        exon2Line = exon2Line.split()
        e1Start = int(exon1Line[3])
        e1End = int(exon1Line[4])
        e1Len = int(abs(e1Start-e1End)) + 1
        e2Start = int(exon2Line[3])
        e2End = int(exon2Line[4])
        e2Len = int(abs(e2Start-e2End)) + 1
        trueLen = (e1Len + e2Len) / 2
        
        consensusName = trLine.split()[0]
        featureName = trLine.split("Target=")[1].split()[0]
        usedFeatures = dict()
        try:
            usedFeatures = dClassifInfo[consensusName][category]["TermRepeats"]
        except KeyError:
            pass

        for k, v in usedFeatures.iteritems():
            # non-term LTR or TIR
            if "non" in k:
                if (int(v) == int(trueLen) and 
                    (re.match("Rep\d+LTR.*", featureName) or re.match("Rep\d+non-termTIR.*", featureName))):
                    used = True
                    break
            # term LTR or TIR
            else:
                if (int(v) == int(trueLen) and 
                    (re.match("Rep\d+termLTR.*", featureName) or re.match("Rep\d+termTIR.*", featureName))):
                    used = True
                    break
        return used
    
    ## Check whether a feature is used in the classification or not
    #
    # @param gff3Line Line for a feature from the gff3
    # @param category Classification category to check ("coding", "struct" or "other")
    # @param featType Feature type
    # @param dClassifInfo dictionary with the classification infos
    # @return Boolean
    #
    def _isFeatureUsed(self, gff3Line, category, featType, dClassifInfo):
        used = False
        consensusName = gff3Line.split()[0]
        # The "Target" field in the GFF3 ends with 2 values to delete.
        featureName = gff3Line.split("Target=")[1].rsplit(" ", 2)[0]
        featStart = int(gff3Line.split()[3])
        featEnd = int(gff3Line.split()[4])
        if featStart > featEnd:
            featStart, featEnd = featEnd, featStart
        featLength = featEnd - featStart + 1
        usedFeatures = []
        try:
            res = []
            if featType == "Profiles":
                kErr = 0
                try:
                    tmp = dClassifInfo[consensusName][category]["profiles"]
                except KeyError:
                    kErr += 1
                else:
                    res.extend(tmp.keys())
                try:
                    tmp = dClassifInfo[consensusName][category]["Other_profiles"]
                except KeyError:
                    kErr +=1
                else:
                    res.extend(tmp.keys())
                if kErr == 2:
                    raise KeyError
            elif featType == "TR": # Unused as long as _dealWithTRLine is on
                res = dClassifInfo[consensusName][category]["TermRepeats"]
            elif featType in ["ORF", "SSR", "SSRtrf"]:
                res = dClassifInfo[consensusName][category][featType]
            elif featType == "polyA":
                if "polyAtail" in dClassifInfo[consensusName][category].keys():
                    used = True
            elif featType == "TE_BLRn":
                res = dClassifInfo[consensusName][category]["helitronExtremities"].keys()
            elif featType in ["rDNA_BLRn", "HG_BLRn"]:
                res = [dClassifInfo[consensusName][category][featType]["name"]]
            else:
                res = dClassifInfo[consensusName][category][featType].keys()
        except KeyError:
            pass
        else:
            usedFeatures = res
        
        if len(usedFeatures) > 0:
            if featType == "TR": # Unused as long as _dealWithTRLine is on
                trLength = gff3Line.split("Target=")[1].split("|")[1].split("len=")[1]
                for k, v in usedFeatures.iteritems():
#TODO: in classif, LTR and TIR can either be "term" or "non-term", but in the tables, the possible
# names are "termLTR", "LTR", "termTIR", "non-termTIR".
# Maybe an error to fix for the LTR ?
                    # non-term LTR or TIR
                    if "non" in k:
                        if (int(v) == int(trLength) and 
                            (re.match("Rep\d+LTR.*", featureName) or re.match("Rep\d+non-termTIR.*", featureName))):
                            used = True
                            break
                    # term LTR or TIR
                    else:
                        if (int(v) == int(trLength) and 
                            (re.match("Rep\d+termLTR.*", featureName) or re.match("Rep\d+termTIR.*", featureName))):
                            used = True
                            break
            elif featType == "ORF":
                usedORFLengthsText = [e.split()[0] for e in usedFeatures]
                usedORFLengthsValues = [int(e[1:-3]) for e in usedORFLengthsText]
                thresholdLength = min(usedORFLengthsValues)
                if featLength > thresholdLength:
                    used = True
            elif featType in ["SSR","SSRtrf"]:
                consEnd = dClassifInfo[consensusName]["length"]
                usedSSRNames = [e.replace("(","").replace(")","") for e in usedFeatures]
                for e in usedSSRNames:
                    ssrName, suffixe = e.split("_")[0], e.split("_")[-1]
                    if featureName == ssrName and ((suffixe == "start" and featStart == 1) or (suffixe == "end" and featEnd == consEnd)):
                        used = True
                        break
            else:
                used = bool(featureName in usedFeatures)
        return used
    
    ## Checks the existence of the features tables
    #
    def _checkFeatureTables(self):
        # Loop over a copy of the list (with [:]) so we can modify the original while iterating
        for featType in self._lFeatureType[:]:
            featTableName = self._searchTableForFeature(featType)
            if featTableName:
                # Add table name information to the infos dict
                self._dFeatToInfos[featType]["table"] = featTableName
            else:
                # Backward compatibility for classif files produced with repet < 2.4 : retry with SSRtrf
                #TODO: Might be removed if backward compatibility isn't needed anymore
                if featType == "SSR":
                    self._log.info("Checking alternative for SSR features : SSRtrf")
                    altFeatType = "SSRtrf"
                    altFeatTableName = self._searchTableForFeature(altFeatType)
                    if featTableName:
                        self._lFeatureType.append(altFeatType)
                        self._dFeatToInfos[altFeatType]["table"] = altFeatTableName

                # No table found, remove feature from list
                self._lFeatureType.remove(featType)
        if self._lFeatureType == []:
            msg = "No GFF3 can be created"
            self._logAndRaise(msg)
    
    def _searchTableForFeature(self, featType):
        tableName = ""
        if featType == "annotated_copies":
            if self._iDb.doesTableExist("%s_chr_allTEs_nr_noSSR_join_path" % self._TEannotProjectName):
                tableName = "%s_chr_allTEs_nr_noSSR_join_path" % self._TEannotProjectName
            elif self._iDb.doesTableExist("%s_chr_allTEs_nr_join_path" % self._TEannotProjectName):
                tableName = "%s_chr_allTEs_nr_join_path" % self._TEannotProjectName
            elif self._iDb.doesTableExist("%s_chr_allTEs_nr_noSSR_path" % self._TEannotProjectName):
                tableName = "%s_chr_allTEs_nr_noSSR_path" % self._TEannotProjectName
            elif self._iDb.doesTableExist("%s_chr_allTEs_nr_path" % self._TEannotProjectName):
                tableName = "%s_chr_allTEs_nr_path" % self._TEannotProjectName
            else:
                msg = "Table for annotated copies couldn't be found, make sure you ran TEannot"
                self._log.error(msg)
        else:
            suffix = "%s_%s" % (featType, self._dFeatToInfos[featType]["format"])
            if self._iDb.doesTableExist("%s_%s" % (self._projectName, suffix)):
                tableName = "%s_%s" % (self._projectName, suffix)
            elif self._iDb.doesTableExist("%s_sim_%s" % (self._projectName, suffix)):
                tableName = "%s_sim_%s" % (self._projectName, suffix)
            elif self._iDb.doesTableExist("%s_struct_%s" % (self._projectName, suffix)):
                tableName = "%s_struct_%s" % (self._projectName, suffix)
            else:
                msg = "No table found for %s (table name should be like %s_[...]_%s)" % (featType, self._projectName, suffix)
                self._log.error(msg)
        
        return tableName
    
    ## Checks the existence of the classif file
    #
    def _checkClassifFile(self):
        classifConsensusFolder = glob.glob("%s*_Map_TEclassif" % self._projectName)
        if len(classifConsensusFolder) > 0:
            classifConsensusFolder = "%s/classifConsensus" % classifConsensusFolder[0]
            if not os.path.exists(classifConsensusFolder):
                msg = "Missing classifConsensus directory : %s/classifConsensus" % classifConsensusFolder
                self._logAndRaise(msg)
        else:
            msg = "Missing Map_TEclassif directory : %s_[...]_Map_TEclassif" % self._projectName
            self._logAndRaise(msg)
        self._classifFileName = glob.glob("%s/%s*_WickerH.classif" % (classifConsensusFolder, self._projectName))
        if len(self._classifFileName) > 0:
            self._classifFileName = "../%s" % self._classifFileName[0]
        else:
            msg = "Couldn't find classif file in folder %s" % classifConsensusFolder
            self._logAndRaise(msg)
    
    def _start(self):
        self._checkOptions()
        self._checkConfig()
        self._iDb = DbFactory.createInstance()
        if self._addClassifInfos:
            self._checkClassifFile()
        self._checkFeatureTables()
        if not os.path.exists(self._cgcfDir):
            os.mkdir(self._cgcfDir)
        else:
            msg = "%s directory already exists." % self._cgcfDir
            self._log.warning(msg)
        os.chdir(self._cgcfDir)
        os.symlink("../%s" % self._fastaFilePath, self._fastaFileName)
        
    def _clean(self):
        self._iDb.dropTable(self._seqTable)
        os.system("ListAndDropTables.py -d _tmpForCGCF")
        for featType in self._lFeatureType:
            try:
                os.remove(self._dFeatToInfos[featType]["info_file"])
            except Exception:
                pass
        os.chdir(self._projectDir)
         
    def run(self):
        LoggerFactory.setLevel(self._log, self._verbosity)
        self._log.info("START CreateGFF3sForClassifFeatures (%s)" % time.strftime("%Y-%m-%d %H:%M:%S"))
        
        self._start()
        try:
            self._createConsensusTableAndHeadersMapping()
            self._createFeatureTablesAndInfoFiles()
            self._createGFF3s()
            self._editGFF3s()
            if self._addOriginalCopies:
                self._createOriginalCopiesLinks()
        finally:
            if self._doClean:
                self._clean()
        
        self._log.info("END CreateGFF3sForClassifFeatures (%s)" % time.strftime("%Y-%m-%d %H:%M:%S"))
        
        
if __name__ == "__main__":
    iCGCF = CreateGFF3sForClassifFeatures()
    iCGCF._setAttributesFromCmdLine()
    iCGCF.run()
    