#!/usr/bin/env python

import os
from commons.core.parsing.FastaParser import FastaParser
from commons.core.utils.FileUtils import FileUtils
from commons.core.utils.RepetOptionParser import RepetOptionParser

class ReverseComplementAccordingToClassif(object):
    
    def __init__(self, fastaFile = None, outFileName = None, classifFileName = None, isOutClassif = False):
        self._fastaFile = fastaFile
        self._outFileName = outFileName
        self._classifFileName = classifFileName
        self._isOutClassif = isOutClassif
        
    def setFastaFile(self, fastaFile):
        self._fastaFile = fastaFile
        
    def setOutFileName(self, fileName):
        self._outFileName = fileName

    def setClassifFile(self, classifFile):
        self._classifFileName = classifFile
        
    def setIsOutClassif(self, isOutClassif):
        self._isOutClassif = isOutClassif
    
    def setAttributesFromCmdLine(self):
        usage = "ReverseComplementAccordingToClassif.py [options]"
        descr = "According to classification file from PASTEC, the sequence header is tagged with '_reversed' if it is on negative strand\n"
        descr += "with -i option, The sequence header is also tagged in classif file\n"
        epilog = "Example: add tag '_reversed' in seqFileName_negStrandReversed.fa output file and seqFileName_negStrandReversed.classif\n"
        epilog += "\t ReverseComplementAccordingToClassif.py -f seqFileName.fa -i fileName.classif -n True\n"
        parser = RepetOptionParser(description = descr, usage = usage)
        parser.add_option('-f', '--fasta', dest='fastaFile', help='fasta file [Mandatory]', default= None)
        parser.add_option("-o", "--output",dest = "outputFileName", help = "output fasta file name [default: <fasta file name>_negStrandReversed.fa])", default = None)
        parser.add_option('-i', '--input', dest='classifFileName', help='classification file from PASTEC [Mandatory]', default= None)
        parser.add_option('-n', '--new',   dest='isOutClassif', help='demand output classification file with reversed headers', default= False, action="store_true")
        (options, args) = parser.parse_args()
        self._setAttributesFromOptions(options)
        
    def _setAttributesFromOptions(self, options):
        self.setFastaFile(options.fastaFile)
        self.setOutFileName(options.outputFileName)
        self.setClassifFile(options.classifFileName)
        self.setIsOutClassif(options.isOutClassif)
        
    def checkOptions(self):        
        if self._fastaFile == None:
            raise Exception("ERROR: No fasta file specified in option!")
        elif not FileUtils.isRessourceExists(self._fastaFile):
            raise Exception("ERROR: Can't open fasta file at %s" % self._fastaFile)
        
        if self._classifFileName == None:
            raise Exception("ERROR: No classification file specified in option!")
        elif not FileUtils.isRessourceExists(self._classifFileName):
            raise Exception("ERROR: Can't open classification file at %s" % self._classifFileName)

        if self._outFileName == None:
            prefixe = os.path.basename(self._fastaFile)
            prefixe = os.path.splitext(prefixe)
            self._outFileName = "%s_negStrandReversed%s" % (prefixe[0],prefixe[1])
    
    def _writeOutputClassif(self, dNameToNegStrand):
#        the new fasta and classif files have the same name
        outClassifFileName = "%s.classif" % os.path.splitext(os.path.basename(self._outFileName))[0]

        outClassifFile = open(outClassifFileName, 'a')
        classifFile = open(self._classifFileName, "r")
        line = classifFile.readline()
        while line:
            newline = line
            if line.split('\t')[0] in dNameToNegStrand.keys():
                newName = ["%s_reversed" % line.split('\t')[0]]
                newName.extend(line.split('\t')[1:])
                newline = '\t'.join(newName)
            outClassifFile.write(newline)
            line = classifFile.readline()
        classifFile.close()
        outClassifFile.close()
    
    def run(self):
        self.checkOptions()
        classifFile = open(self._classifFileName, "r")
        dNameToNegStrand = {}
        line = classifFile.readline()
        while line:
            if line.split('\t')[2] == '-':
                dNameToNegStrand[line.split('\t')[0]] = '-'
            line = classifFile.readline()
        classifFile.close()
        
        resultFile = open(self._outFileName, "w")
        iFastaParser = FastaParser(self._fastaFile)
        for iSequence in iFastaParser.getIterator():
            if iSequence.getName() in dNameToNegStrand.keys():
                iSequence.reverseComplement()
                iSequence.setName("%s_reversed" % iSequence.getName())
            resultFile.write(iSequence.printFasta())
        resultFile.close()
        
        if self._isOutClassif:
            self._writeOutputClassif(dNameToNegStrand)
        
if __name__ == '__main__':
    iReverseComplementAccordingToClassif = ReverseComplementAccordingToClassif()
    iReverseComplementAccordingToClassif.setAttributesFromCmdLine()
    iReverseComplementAccordingToClassif.run()