#!/usr/bin/env python

# Copyright INRA (Institut National de la Recherche Agronomique)
# http://www.inra.fr
# http://urgi.versailles.inra.fr
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use, 
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info". 
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability. 
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or 
# data to be ensured and,  more generally, to use and operate it in the 
# same conditions as regards security. 
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import os
from commons.core.LoggerFactory import LoggerFactory
from commons.core.utils.FileUtils import FileUtils
from commons.core.utils.RepetOptionParser import RepetOptionParser
from PASTEC.ClassifHierarchyContainer import ClassifHierarchyContainer

LOG_DEPTH = "commons.tools"

class TELibFormatter(object):
    
    def __init__(self, fastaFileName = "", rulesFileName = "", outFileName = "", separators = "", verbosity = 3):
        self._fastaFileName = fastaFileName
        self._rulesFileName = rulesFileName
        self._separators = separators
        self.setOutFileName(outFileName)
        self._verbosity = verbosity
        self._log = LoggerFactory.createLogger("%s.%s" % (LOG_DEPTH, self.__class__.__name__), self._verbosity)
        
    def setAttributesFromCmdLine(self):
        usage = "TELibFormatter.py.py [options]"
        description = "Format TEs library, in fasta format, for REPET using a formatting rules file of headers and field separator(s)\n"
        description += "formatting rules file of headers describes the fields corresponding with REPET format fields and separator(s).\n"
        epilog = "\nExample: launch with |(pipe) and \(backslash) as header fields separators (TAIR10 format).\n"
        epilog += "\t using tairTELib.fa as input TE library \n"
        epilog += "\t using header formating rules specified in headerRules.txt file.\n"
        epilog += "\t This file contains for example 'name|strand|other|other|Family|order/superFamily|other'\n"
        epilog += "\t 'name', 'Family' and 'order/superFamily' fields will be used to write the REPET header 'name:class:order:superFamily:family'\n"
        epilog += "\t In this case the class field is no explicit but it is inferred from 'order/superFamily'\n"
        epilog += "\t Command line\n"
        epilog += "\t $ TELibFormatter.py -i tairTELib.fa -r headerRules.txt -s '|\' -v 3\n"
        epilog += "\t the result file is input fasta file with new headers, used as a bank by REPET.\n"
        parser = RepetOptionParser(description = description, epilog = epilog, usage = usage)
        parser.add_option("-i", "--fasta",      dest = "fastaFileName", action = "store",       type = "string", help = "input fasta file name [compulsory] [format: fasta]", default = "")
        parser.add_option("-r", "--rules",      dest = "rulesFileName", action = "store",       type = "string", help = "input header formating rules file name [compulsory]", default = "")
        parser.add_option("-s", "--separators", dest = "separators",    action = "store",       type = "string", help = "list of separators (one char length: for example: | (1 seperator) or |: (2 seperators) or \|; etc [compulsory]", default = "")
        parser.add_option("-o", "--out",        dest = "outFileName",   action = "store",       type = "string", help = "output file name [default: <input>.newH]", default = "")
        parser.add_option("-v", "--verbosity",  dest = "verbosity",     action = "store",       type = "int",    help = "verbosity [optional] [default: 3]", default = 3)
        options = parser.parse_args()[0]
        self._setAttributesFromOptions(options)
        
    def _setAttributesFromOptions(self, options):
        self._fastaFileName = options.fastaFileName
        self._rulesFileName = options.rulesFileName
        self.setOutFileName(options.outFileName)
        self._separators = options.separators
        self._verbosity = options.verbosity

    def setOutFileName(self, outFileName):
        if outFileName == "":
            self._outFileName = "%s.newH" % self._fastaFileName
        else:
            self._outFileName = outFileName
        
    def _checkOptions(self):
        if self._fastaFileName == "":
            self._logAndRaise("ERROR: Missing input fasta")
        else:
            if not FileUtils.isRessourceExists(self._fastaFileName):
                self._logAndRaise("Input fasta file '%s' does not exist!" % self._fastaFileName)
        
        if self._rulesFileName == "":
            self._logAndRaise("ERROR: Missing rules file")
        else:
            if not FileUtils.isRessourceExists(self._rulesFileName):
                self._logAndRaise("Rules file '%s' does not exist!" % self._rulesFileName)
    
        if self._separators == "":
            self._logAndRaise("At least one separator is required")
            
    def _logAndRaise(self, errorMsg):
        self._log.error(errorMsg)
        raise Exception(errorMsg)
    
    def _generateHeader(self, line, mainSep, subSeps, dFieldPositions, iClassifHierarchyContainer):
        outputLine = line
        for subSep in subSeps:
            outputLine = outputLine.replace(subSep, mainSep)
        lFields = outputLine.split(mainSep) 
        
        def getFieldValue(name):
            index = dFieldPositions.get(name, None)
            if index is not None:
                if index != -1:
                    return lFields[index]
            return ""
        
        name = getFieldValue("name")
        klass = getFieldValue("class")
        order = getFieldValue("order")
        sFamily = getFieldValue("superfamily")
        family = getFieldValue("family")
        
        ignoreCase = True
        if klass is "":
            if order != "":
                klass = iClassifHierarchyContainer.getClassFromOrder(order)
            if sFamily != "" and klass is "":
                klass = iClassifHierarchyContainer.getClassAndOrderFromSuperFamily(sFamily, ignoreCase)[0]
            if family != "" and klass is "":
                klass = iClassifHierarchyContainer.getClassOrderAndSuperFamilyFromFamily(family, ignoreCase)[0]
                
        if order is "":
            if sFamily != "" and order is "":
                order = iClassifHierarchyContainer.getClassAndOrderFromSuperFamily(sFamily, ignoreCase)[1]
            if family != "" and order is "":
                order = iClassifHierarchyContainer.getClassOrderAndSuperFamilyFromFamily(family, ignoreCase)[1]
                
        elif sFamily is "":
            if family != "":
                sFamily = iClassifHierarchyContainer.getClassOrderAndSuperFamilyFromFamily(family, ignoreCase)[2]
        
        newHeader = [name, klass, order, sFamily, family]
        if "" in newHeader:
            self._logAndRaise("Badly formatted header or missing fields")
        newHeader = ":".join(newHeader) + "\n"
        return newHeader
    
    def run(self):
        LoggerFactory.setLevel(self._log, self._verbosity)
        self._checkOptions()
        self._log.info("START TELibFormatter")
        self._log.debug("Fasta file name: %s" % self._fastaFileName)
        
        with open(self._rulesFileName) as rulesFileHandler:
            rules = rulesFileHandler.readlines()
            if len(rules) != 1:
                self._logAndRaise("Only one line header rules are allowed")

        rules = rules[0].lower()
        
        mainSep = self._separators[0]
        subSeps = self._separators[1:]
        for subSep in subSeps:
            rules = rules.replace(subSep,mainSep)
            
        lKeyWords = rules.split(mainSep)
        dFieldPositions = {"name": -1, "class": -1, "order": -1, "superfamily": -1, "family": -1}
        
        for index, inkeyWord in enumerate(lKeyWords):
            if inkeyWord in dFieldPositions.keys():
                dFieldPositions[inkeyWord] = index
        
        classifFileName = "%s/PASTEC/WickerClassifHierarchy.txt" % os.environ["REPET_PATH"]
        iClassifHierarchyContainer = ClassifHierarchyContainer(classifFileName)
        iClassifHierarchyContainer.parse()

        fastaFileHandler = open(self._fastaFileName)
        outputFileHandler = open(self._outFileName,'w')
        line = fastaFileHandler.readline()
        while line:
            if ">" in line:
                newHeader = self._generateHeader(line, mainSep, subSeps, dFieldPositions, iClassifHierarchyContainer)
                outputFileHandler.write(newHeader)
            else:
                outputFileHandler.write(line)
            line = fastaFileHandler.readline()
        fastaFileHandler.close()
        outputFileHandler.close()
        
        self._log.info("END TELibFormatter")

if __name__ == "__main__":
    iTELibFormatter = TELibFormatter()
    iTELibFormatter.setAttributesFromCmdLine()
    iTELibFormatter.run()