
import user, os, re, sys, time, logging
from exceptions import *

from pyRepet.sql.RepetDBFile import RepetDB

#------------------------------------------------------------------------------

class Job:

    def __init__( self, tablename, jobid, groupid, queue, command, launcherFile, node="?" ):

        self.tablename = tablename
        self.jobid = jobid
        self.groupid = groupid
        self.queue = queue
        self.command = command
        self.launcher = launcherFile
        self.node = node

#------------------------------------------------------------------------------

class RepetJob( RepetDB ):

    """
    connector to manage jobs launched in parallel, their features being recorded
    in a MySQL database
    """

    #--------------------------------------------------------------------------

    def create_job_table( self, tablename ):

        """
        create a job table
        
        @param tablename: new table name
        @type tablename: string
        """

        self.remove_if_exist( tablename )
        os.mkdir( self.dbname + "/" + tablename )
        print "table created !!"

    #--------------------------------------------------------------------------

    def record_job( self, job ):

        """
        record a job
        
        @param job: a L{Job<Job>} instance with the job informations
        @type job: L{Job<Job>} instance
        """

        if not os.path.exists( self.dbname + "/" + job.tablename + "/" + job.groupid ):
            os.mkdir( self.dbname + "/" + job.tablename + "/" + job.groupid )
        else:
            self.remove_job( job )

        jobFile = open( self.dbname + "/" + job.tablename + "/" + job.groupid + "/" + job.jobid, "w" )
        now = time.localtime()
        now = str(now[0]) + "-" + str(now[1]) + "-" + str(now[2]) + "-" + str(now[3]) + ":" + str(now[4]) + ":" + str(now[5])
        jobFile.write( job.jobid + "\t" + job.groupid + "\tcommand\t" + job.launcher + "\t" + job.queue + "\twaiting\t" + now + "\t?\n" )
        jobFile.close()

    #--------------------------------------------------------------------------

    def remove_job( self, job ):

        """
        remove a job from the job table

        @param tablename: table name to record the jobs
        @type tablename: string
    
        @param jobid: a job identifier
        @type jobid: string

        @param groupid: a group identifier to record related job series 
        @type groupid: string

        @param queue: queue name of the batch job manager
        @type queue: string

        @param command: command launched
        @type command: string
        """

        if os.path.exists( self.dbname + "/" + job.tablename + "/" + job.groupid + "/" + job.jobid ):
            os.system( "rm -f " + self.dbname + "/" + job.tablename + "/" + job.groupid + "/" + job.jobid )

    #--------------------------------------------------------------------------

    def change_job_status( self, job, status, method ):

        """
        change a job status
        
        @param job: a L{Job<Job>} instance with the job informations
        @type job: L{Job<Job>} instance

        @param status: the new status (waiting,running,finished,error)
        @type status: string
        """

        jobFile = open( self.dbname + "/" + job.tablename + "/" + job.groupid + "/" + job.jobid, "a" )
        now = time.localtime()
        now = str(now[0]) + "-" + str(now[1]) + "-" + str(now[2]) + "-" + str(now[3]) + ":" + str(now[4]) + ":" + str(now[5])
        jobFile.write( job.jobid + "\t" + job.groupid + "\tcommand\t" + job.launcher + "\t" + job.queue + "\t" + status + "\t" + now + "\t?\n" )
        jobFile.close()

    #--------------------------------------------------------------------------

    def get_job_status( self, job ):

        """
        get a job status
        
        @param job: a L{Job<Job>} instance with the job informations
        @type job: L{Job<Job>} instance
        """

        if not os.path.exists( self.dbname + "/" + job.tablename + "/" + job.groupid + "/" + job.jobid ):
            return

        jobFile = open( self.dbname + "/" + job.tablename + "/" + job.groupid + "/" + job.jobid, "r" )
        lines = jobFile.readlines()
        last = lines[-1].split()
        jobFile.close()
        return last[5]

    #--------------------------------------------------------------------------

    def clean_job_group( self, tablename, groupid ):

        """
        clean all job from a job group.

        @param tablename: table name to record the jobs
        @type tablename: string
    
        @param groupid: a group identifier to record related job series 
        @type groupid: string
        """

        if os.path.exists( self.dbname + "/" + tablename + "/" + groupid ):
            os.system( "rm -rf " + self.dbname + "/" + tablename + "/" + groupid )

    #--------------------------------------------------------------------------

    def has_unfinished_job( self, tablename, groupid ):

        """
        check if there is unfinished job from a job group

        @param tablename: table name to record the jobs
        @type tablename: string
    
        @param groupid: a group identifier to record related job series 
        @type groupid: string
        """

        if not self.exist( tablename ):
            return False

        if not os.path.exists( self.dbname + "/" + tablename + "/" + groupid ):
            return False

        files = os.listdir( self.dbname + "/" + tablename + "/" + groupid )
        for f in files:
            jobFile = open( self.dbname + "/" + tablename + "/" + groupid + "/" + f, "r" )
            lines = jobFile.readlines()
            last = lines[-1].split()
            jobFile.close()
            if last[5] != "finished":
                return False

    #--------------------------------------------------------------------------

    def wait_job_group( self, tablename, groupid ):

        """
        wait job finished status from a job group.

        @note: job are re-launched if error

        @param tablename: table name to record the jobs
        @type tablename: string
    
        @param groupid: a group identifier to record related job series 
        @type groupid: string
        """

        error_count = {}
        timeout_count = {}

        files = os.listdir( self.dbname + "/" + tablename + "/" + groupid )
        k = 0

        while k < len(files):
            k = 0
            time.sleep(60)

            for f in files:
                jobFile = open( self.dbname + "/" + tablename + "/" + groupid + "/" + f, "r" )
                lines = jobFile.readlines()
                last = lines[-1].split()
                jobFile.close()

                if last[5] == "finished":
                    k += 1

                if last[5] == "error":
                    ecount = error_count.get( last[0], 0 )
                    if ecount == 0:
                        error_count[last[0]] = 1
                    elif ecount > 2:
                        print( "%s permanent error (%d times)!" %(last[0],ecount) )
                    else:
                        error_count[last[0]] += 1
                        print( last[0] + " error, re-submiting (" + str(ecount+1) + " times)!" )
                        job = Job( tablename, last[0], groupid, last[4], last[2], last[3] )
                        self.submit_job( job )

##                now = time.localtime()
##                if str(last[6].split("-")[2]) != str(now[2]):
##                    tocount = timeout_count.get(last[0],0)
##                    if tocount == 0:
##                        timeout_count[last[0]] = 1
##                    elif tocount > 2:
##                        print("%s permanent timeout (%d times)!"%(last[0],tocount))
##                    else:
##                        timeout_count[last[0]] += 1
##                        print( "%s timeout, re-submiting (%d times)!"%(last[0],tocount+1) )
##                        job = Job( tablename, last[0], groupid, last[4], last[2], last[3] )
##                        self.submit_job( job )

    #--------------------------------------------------------------------------

    def submit_job( self, job ):

        """
        submit a job to a queue and record it in job table
        
        @param job: a L{Job<Job>} instance with the job informations
        @type job: L{Job<Job>} instance
        """

        if not self.exist(job.tablename):
            self.create_job_table(job.tablename)

        if self.get_job_status(job)=="finished" \
               or self.get_job_status(job)=="waiting" \
               or self.get_job_status(job)=="running":
            logging.error( "*** Warning: job " + job.jobid +\
                          " already submitted" )
            return

        self.record_job( job )

        if os.environ.get( "REPET_QUEUE" ) == "PBS":
            launchCmd = "qsub -q " + job.queue +\
                         " -j oe -l \"mem=500Mb\" -V -N " + job.jobid +\
                         " <<EOF\n" + job.launcher + "\nEOF\n"
            os.system( launchCmd )

        #TODO: change queue management
        elif os.environ.get( "REPET_QUEUE" ) == "SGE":
            launchCmd = "qsub -q " + job.queue + " -j y -cwd -V -N \"" +\
                         job.jobid + "\" <<EOF\n" + job.launcher + "\nEOF\n"
            os.system( launchCmd )

        else:
            print "*** Error: queue manager " +\
                  os.environ.get( "REPET_QUEUE" ) +\
                  " not supported"
            pass
